# PYTHON_ARGCOMPLETE_OK

# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

import argparse
import argcomplete
import logging
import six
import sys
import traceback

from pvmcli.command import cluster
from pvmcli.command import io_slot
from pvmcli.command import logical_unit
from pvmcli.command import logical_partition
from pvmcli.command import logical_volume
from pvmcli.command import managed_system
from pvmcli.command import network_bridge
from pvmcli.command import physical_volume
from pvmcli.command import shared_processor_pool
from pvmcli.command import shared_storage_pool
from pvmcli.command import sriov_adapter
from pvmcli.command import sriov_logical_port
from pvmcli.command import virtual_ethernet_adapter
from pvmcli.command import virtual_fc_mapping
from pvmcli.command import virtual_io_server
from pvmcli.command import virtual_media_repository
from pvmcli.command import virtual_network
from pvmcli.command import virtual_optical_media
from pvmcli.command import virtual_scsi_mapping
from pvmcli.command import virtual_switch
from pvmcli.command import vnic
from pvmcli.command import volume_group
from pvmcli.command import pmem
from pvmcli.command import virtual_serial_number as vsn
from pvmcli.command import health_monitor

import pvmcli.help.help_pvmctl as help_pvmctl
import pvmcli.util.exceptions as exc
import pvmcli.util.help as help
from pvmcli.util.i18n import _
import pvmcli.util.offline as offline
import pvmcli.util.util as util

import pypowervm.exceptions as pyexc


OK_RETURN_CODE = 0


ARG_AUTH_USER = 'user'
ARG_AUTH_PASS = 'password'
ARG_AUTH_HOST = 'host'
ARG_AUTH_PORT = 'port'


OBJ_TYPE_NAME_TO_CLASS = {
    'managedsystem': managed_system.ManagedSystem,
    'logicalpartition': logical_partition.LogicalPartition,
    'virtualethernetadapter': virtual_ethernet_adapter.VirtualEthernetAdapter,
    'virtualswitch': virtual_switch.VirtualSwitch,
    'virtualnetwork': virtual_network.VirtualNetwork,
    'networkbridge': network_bridge.NetworkBridge,
    'virtualioserver': virtual_io_server.VirtualIOServer,
    'logicalunit': logical_unit.LogicalUnit,
    'physicalvolume': physical_volume.PhysicalVolume,
    'sharedstoragepool': shared_storage_pool.SharedStoragePool,
    'sharedprocessorpool': shared_processor_pool.SharedProcessorPool,
    'logicalvolume': logical_volume.LogicalVolume,
    'virtualmediarepository': virtual_media_repository.VirtualMediaRepository,
    'virtualopticalmedia': virtual_optical_media.VirtualOpticalMedia,
    'volumegroup': volume_group.VolumeGroup,
    'ioslot': io_slot.IOSlot,
    'virtualscsimapping': virtual_scsi_mapping.SCSIStorageMapping,
    'virtualfibrechannelmapping': virtual_fc_mapping.VFCStorageMapping,
    'cluster': cluster.Cluster,
    'sriovadapter': sriov_adapter.SRIOVAdapter,
    'sriovethernetlogicalport': sriov_logical_port.SRIOVLogicalPort,
    'virtualnicdedicated': vnic.VNIC,
    'persistentmemoryvolume': pmem.PMEMVolume,
    'persistentmemorydevice': pmem.PMEMDevice,
    'virtualserialnumber': vsn.VirtualSN,
    'healthmonitor': health_monitor.HealthMon
}

OBJ_TYPE_ALIASES = {
    'sys': 'managedsystem',
    'lpar': 'logicalpartition',
    'vm': 'logicalpartition',
    'vea': 'virtualethernetadapter',
    'eth': 'virtualethernetadapter',
    'vswitch': 'virtualswitch',
    'vsw': 'virtualswitch',
    'vnet': 'virtualnetwork',
    'net': 'virtualnetwork',
    'nbr': 'networkbridge',
    'bridge': 'networkbridge',
    'sharedethernetadapter': 'networkbridge',
    'sea': 'networkbridge',
    'vios': 'virtualioserver',
    'lu': 'logicalunit',
    'pv': 'physicalvolume',
    'ssp': 'sharedstoragepool',
    'sharedprocpool': 'sharedprocessorpool',
    'spp': 'sharedprocessorpool',
    'lv': 'logicalvolume',
    'repo': 'virtualmediarepository',
    'vmr': 'virtualmediarepository',
    'media': 'virtualopticalmedia',
    'vom': 'virtualopticalmedia',
    'vg': 'volumegroup',
    'io': 'ioslot',
    'scsi': 'virtualscsimapping',
    'scsimapping': 'virtualscsimapping',
    'vfc': 'virtualfibrechannelmapping',
    'vfcmapping': 'virtualfibrechannelmapping',
    'sriov': 'sriovadapter',
    'ethlogicalport': 'sriovethernetlogicalport',
    'ethlp': 'sriovethernetlogicalport',
    'vnic': 'virtualnicdedicated',
    'pmemv': 'persistentmemoryvolume',
    'pmemd': 'persistentmemorydevice',
    'vsn': 'virtualserialnumber',
    'hms': 'healthmonitor'
}


class CLIParser(object):
    ''' Entry point parser for pvmctl.

    Handles help, authentication, version, and object type construction based
    on user provided arguments.

    :param args: A list of argument strings, generally sys.argv[1:]
    :param help_out: The output file for help text output.
    '''
    def __init__(self, args, help_out=sys.stdout):
        self.args = args
        self.help_out = help_out
        self.obj_type = None
        self.parser = self._create_parser()

        self.is_help_text_requested = False
        self.is_version_requested = False
        self.user_arg = None
        self.pass_arg = None
        self.host_arg = None
        self.port_arg = None
        self.parsed_args = self._parse_args()

    def _create_parser(self):
        # Construct argument parser
        parser_usage = util.utf8_str(' '.join([
            util.utf8_str(help.PROG_NAME), util.utf8_str(help.HELP_ARGS),
            util.utf8_str(help.VERSION_ARGS),
            util.utf8_str(help.OBJ_TYPE_USG)]))

        parser = argparse.ArgumentParser(
            prog=util.utf8_str(help.PROG_NAME),
            description=util.utf8_str(help.PROG_DESC),
            usage=parser_usage, add_help=False)

        # Add all arguments to the pvmctl base command
        self._add_outer_command_arguments(parser)

        # Enable bash auto complete
        argcomplete.autocomplete(parser)

        return parser

    def _version(self):
        try:
            with open('/opt/pvm/version/pvm-cli.version', 'r') as v_f:
                return v_f.read()
        except Exception:
            return _("pvm-cli version file is missing.")

    def _add_outer_command_arguments(self, parser):
        ''' Build up all outer command arguments (version, auth, help).'''
        # Version information
        parser.add_argument('--version', action='store_true',
                            help=util.utf8_str(_("show version of pvmctl")))

        # Hidden debug option
        parser.add_argument('--debug', action='store_true',
                            help=argparse.SUPPRESS)

        # Hidden demo file option
        offline_grp = _('offline arguments')
        og = parser.add_argument_group(offline_grp)
        og.add_argument('--demo', action='store_true', help=argparse.SUPPRESS)
        og.add_argument('--demo-errors', action='store_true',
                        help=argparse.SUPPRESS)

        # Login information
        auth_grp = _('authentication arguments')
        ag = parser.add_argument_group(auth_grp)
        ag.add_argument('--%s' % ARG_AUTH_USER, type=str,
                        help=argparse.SUPPRESS)
        ag.add_argument('--%s' % ARG_AUTH_PASS, type=str,
                        help=argparse.SUPPRESS)
        ag.add_argument('--%s' % ARG_AUTH_HOST, type=str,
                        help=argparse.SUPPRESS)
        ag.add_argument('--%s' % ARG_AUTH_PORT, type=int,
                        help=argparse.SUPPRESS)

        # Help argument
        # Note: This has to stay a store_true action, since setting it
        # to a help action causes problems with sending the help request
        # to the inner command
        help_group = parser.add_argument_group(
            util.utf8_str(_('Help arguments')))
        help_group.add_argument('help',
                                help=util.utf8_str(help.HELP_DESC),
                                action="store_true")

        help_group.add_argument('--help', action="store_true")

        # Add object type argument and choices
        # Cast to list for python 3 extend support
        obj_type_choices = list(OBJ_TYPE_NAME_TO_CLASS.keys())
        obj_type_choices.extend(OBJ_TYPE_ALIASES.keys())

        parser.add_argument(
            'ObjectType', nargs='?', type=str.lower, choices=obj_type_choices)

        # Inner command arguments
        parser.add_argument(
            'inner_args', metavar='ObjectType arguments',
            nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    def _parse_args(self):
        ''' Parse the user provided arguments through argparse.

        :returns: The argparse parsed args for execution.
        '''
        args_to_parse = []
        for arg_to_check in self.args:
            if arg_to_check in ('--help', 'help'):
                self.is_help_text_requested = True
            elif arg_to_check == '--version':
                self.is_version_requested = True
            else:
                args_to_parse.append(arg_to_check)

        # Parse args
        parsed_args = self.parser.parse_args(args_to_parse)

        if parsed_args.ObjectType:
            obj = parsed_args.ObjectType
            self.obj_type = (OBJ_TYPE_NAME_TO_CLASS[OBJ_TYPE_ALIASES[obj]]()
                             if obj in OBJ_TYPE_ALIASES.keys()
                             else OBJ_TYPE_NAME_TO_CLASS[obj]())

        if ARG_AUTH_USER in dir(parsed_args):
            self.user_arg = parsed_args.user

        if ARG_AUTH_PASS in dir(parsed_args):
            self.pass_arg = parsed_args.password

        if ARG_AUTH_HOST in dir(parsed_args):
            self.host_arg = parsed_args.host

        if ARG_AUTH_PORT in dir(parsed_args):
            self.port_arg = parsed_args.port

        return parsed_args

    def _log_exception(self, prog, args, err_str):
        log_err_str = ('Command "%(prog)s %(args)s" failed with %(err)s' %
                       {'prog': prog, 'args': args,
                        'err': util.utf8_str(err_str)})
        self.log.exception(util.utf8_str(log_err_str))

    def process(self, auto_exec=True, log=logging.getLogger('pvmctl')):
        ''' Process the parsed arguments and execute the ObjectType parser.

        :param auto_exec: True to automatically begin execution on the
                          ObjectType parser if necessary/possible.
        :returns: Execution return code.
        '''
        self.log = log
        self.log.info(('Running command "%(prog)s %(args)s"'),
                      {'prog': util.utf8_str(self.parser.prog),
                       'args': ' '.join(self.args)})

        if self.parsed_args.demo_errors:
            offline.print_all_errors()
            return OK_RETURN_CODE

        if self.is_version_requested:
            util.print_(self._version())
            return OK_RETURN_CODE

        inner_args = self.parsed_args.inner_args
        if not self.obj_type:
            alias_help = help_pvmctl.gen_alias_str(
                OBJ_TYPE_NAME_TO_CLASS.keys(), OBJ_TYPE_ALIASES)
            util.print_(help_pvmctl.help_text % {'ot_aliases': alias_help},
                        outfile=self.help_out)
            return OK_RETURN_CODE
        else:
            if self.is_help_text_requested or not inner_args:
                inner_args.append('--help')
            try:
                needs_exec = self.obj_type.build(
                    self.parsed_args.inner_args, self.user_arg, self.pass_arg,
                    self.host_arg, self.port_arg, help_out=self.help_out)
                if self.parsed_args.demo:
                    offline.execute(self.obj_type)
                elif needs_exec and auto_exec:
                    self.obj_type.execute()
            except exc.CLIAbstractMsgFmtError as cli_exc:
                util.print_('%s\n' % six.text_type(cli_exc),
                            outfile=sys.stderr)
                if self.parsed_args.debug:
                    traceback.print_exc(file=sys.stderr)
                self._log_exception(self.parser.prog,
                                    ' '.join(self.args),
                                    six.text_type(cli_exc))
                return cli_exc.ret_code
            except pyexc.HttpError as http_err:
                if http_err.her_wrap and http_err.her_wrap.message:
                    util.print_('%s\n' % six.text_type(
                        http_err.her_wrap.message), outfile=sys.stderr)
                    self._log_exception(self.parser.prog,
                                        ' '.join(self.args),
                                        six.text_type(http_err))
                else:
                    util.print_(
                        '%s\n' % six.text_type(http_err), outfile=sys.stderr)
                    self._log_exception(self.parser.prog,
                                        ' '.join(self.args),
                                        six.text_type(http_err))
                if self.parsed_args.debug:
                    traceback.print_exc(file=sys.stderr)
                self._log_exception(self.parser.prog,
                                    ' '.join(self.args),
                                    six.text_type(http_err))
                return exc.http_err_ret_code(http_err)
            except pyexc.ConnectionError:
                conn_err_msg = _("Unable to connect to the NovaLink services. "
                                 "Check to make sure the pvm-core and "
                                 "pvm-rest services are running (systemctl "
                                 "status), and verify that this partition has "
                                 "PowerVM management authority. If you are "
                                 "currently updating pvm-novalink, "
                                 "you may see this error. Wait for "
                                 "the update to finish, then try again.")
                util.print_('%s\n' % conn_err_msg, outfile=sys.stderr)
                if self.parsed_args.debug:
                    traceback.print_exc(file=sys.stderr)
                self._log_exception(self.parser.prog,
                                    ' '.join(self.args),
                                    six.text_type(conn_err_msg))
            except Exception as e:
                # If the exception is of the type AbstractMsgFmtError,
                # we will get a properly formed translated string back.
                # What happens otherwise depends on the exception.
                # TODO (clbush): split this up, do something more
                #                interesting with non-amfe errors.
                util.print_('%s\n' % six.text_type(e), outfile=sys.stderr)
                if self.parsed_args.debug:
                    traceback.print_exc(file=sys.stderr)
                self._log_exception(self.parser.prog,
                                    ' '.join(self.args), six.text_type(e))

                # TODO More interesting return codes based on the error
                return exc.GENERAL_ERROR_RETURN_CODE
            # TODO REMOVE THIS once the __del__ session issues are fixed
            # TODO SERIOUSLY, remove this
            # TODO Don't forget.
            finally:
                try:
                    self.obj_type.ec.sm._session_adapter.session = None
                except Exception:
                    pass

        return OK_RETURN_CODE
