# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


class Field(object):
    ''' Represents atomic information when performing an operation.

    For example, display fields can be specified to limit output.
    '''
    def __init__(self, field_name, label=None, help_text=None,
                 field_type=str):
        '''
        :param field_name: The name of this field.
                           Used to access fields through their Arguments.
        :param label: The label of this field, used for output.
                            Defaults to field_name if not specified.
        :param help_text: The help text for this field.
        :param field_type: The type of this field (int, element, str, etc)
        '''
        self._field_name = field_name
        self._label = label if label else field_name
        self._help_text = help_text
        self._field_type = field_type

    @property
    def field_name(self):
        ''' The name of this field. '''
        return self._field_name

    @property
    def label(self):
        ''' The label of this field, used for output. '''
        return self._label

    @label.setter
    def label(self, value):
        ''' Set the label of this field. '''
        self._label = value

    @property
    def help_text(self):
        ''' The help text for this field. '''
        return self._help_text

    @property
    def field_type(self):
        ''' The type of this field. '''
        return self._field_type

    @field_type.setter
    def field_type(self, value):
        ''' Set the type of this field.

        :param value: A valid type defined by the schema.
        '''
        self._field_type = value


class ModelField(Field):
    ''' Special Field for fields with a model associated with it. '''
    def __init__(self, wrp_func_name, model, schema_name, help_text=None,
                 field_type=str):
        '''
        :param wrp_func_name: Name of the pypowervm function this field calls.
        :param model: The pypowervm wrapper object the wrapper function is on
        :param schema_name: The name of a schema. Normally matches the
                            schema_type of the model, but can be overridden
                            in cases like ClientNetworkAdapter to
                            VirtualEthernetAdapter.
        :param help_text: Help text to display for this field.
        :param field_type: The type of this field.
                           Default: str
        '''
        field_name = '%(model)s.%(fn)s' % {'model': schema_name,
                                           'fn': wrp_func_name}
        super(ModelField, self).__init__(
            field_name=field_name, label=wrp_func_name, help_text=help_text)
        self._wrp_func_name = wrp_func_name
        # If model is None, we will assume it is the model of the associated
        # object type.
        self._model = model if model else self.model
        self._schema_name = schema_name

    @property
    def wrapper_func_name(self):
        ''' The name of the pypowervm function this display field calls. '''
        return self._wrp_func_name

    @property
    def model(self):
        ''' The pypowervm wrapper this field maps to. '''
        return self._model

    @property
    def schema_name(self):
        ''' The CLI reference for a schema_type. '''
        return self._schema_name


class DisplayField(ModelField):
    ''' Special Field for --display-fields argument. '''
    def __init__(self, wrp_func_name, model, schema_name, out_func=str,
                 help_text=None, field_type=str):
        '''
        :param wrp_func_name: Name of the pypowervm function this field calls.
        :param model: The pypowervm wrapper object the wrapper function is on
        :param schema_name: The name of a schema. Normally matches the
                            schema_type of the model, but can be overridden
                            in cases like ClientNetworkAdapter to
                            VirtualEthernetAdapter.
        :param out_func: The function that builds the string to output.
                         Default: str
        :param help_text: Help text to display for this field.
        :param field_type: The type of this field.
                           Default: str
        '''
        super(DisplayField, self).__init__(
            wrp_func_name=wrp_func_name, model=model, schema_name=schema_name,
            help_text=help_text, field_type=field_type)
        self._out_func = out_func

    @property
    def out_func(self):
        ''' The output function for this display. For atomic properties, this
            will generally stay as str. However, for more complicated
            properties, this function will generally build a string based on
            the wrapper_func_name.

            The provided function should take 1 argument (the object to
            process and convert into a string to output) and return that
            string.
        '''
        return self._out_func

    @out_func.setter
    def out_func(self, value):
        ''' Set the output function of this field. '''
        self._out_func = value
