# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.util.op_generator as o_g
import pvmcli.help.help_health_monitor as help_hms
import pypowervm.adapter as adp
from pypowervm.wrappers import managed_system as ms_w
from pypowervm.wrappers import logical_partition as lpar_w
import subprocess

PRI_DISP_FIELD_LABELS_VSN = [("MEMORY CONSUMPTION"),
                             ("FILES OPENED"),
                             ("SSH FILES OPENED"),
                             ("PIPE FILES OPENED"), ("REST STATUS")]


class HealthMon(ObjectType):
    def _command(self):
        memtotal = subprocess.run("free | awk '/Mem:/ {print $2}'",
                                  shell=True, capture_output=True, text=True)
        self.memtotal = int(memtotal.stdout)
        swaptotal = subprocess.run("free | awk '/Swap:/ {print $2}'",
                                   shell=True, capture_output=True, text=True)
        self.swaptotal = int(swaptotal.stdout)
        memfree = subprocess.run("free | awk '/Mem:/ {print $4}'",
                                 shell=True, capture_output=True, text=True)
        self.memfree = int(memfree.stdout)
        swapfree = subprocess.run("free | awk '/Swap:/ {print $4}'",
                                  shell=True, capture_output=True, text=True)
        self.swapfree = int(swapfree.stdout)
        bufferCacheMem = subprocess.run("free | awk '/Mem:/ {print $6}'",
                                        shell=True, capture_output=True,
                                        text=True)
        self.bufferCacheMem = int(bufferCacheMem.stdout)
        totalOpenedFiles = subprocess.run(
            "sudo find /proc/*/fd/ |"
            "grep -v "'/$'" |"
            "wc -l |"
            "tail -n 1",
            shell=True, capture_output=True, text=True)
        self.totalOpenedFiles = int(totalOpenedFiles.stdout)
        totalLimitOfFiles = subprocess.run("ulimit -n",
                                           shell=True, capture_output=True,
                                           text=True)
        self.totalLimitOfFiles = int(totalLimitOfFiles.stdout)
        totalOpenedSSHFiles = subprocess.run(
            "sudo find /proc/*/fd/ |"
            "grep -v "'/$'" |"
            "grep "'ssh'" |"
            "wc -l | tail -n 1",
            shell=True, capture_output=True, text=True)
        self.totalOpenedSSHFiles = int(totalOpenedSSHFiles.stdout)
        totalOpenedPIPEFiles = subprocess.run(
            "sudo find /proc/*/fd/ |"
            "grep -v "'/$'" |"
            "grep "'pipe'" |"
            "wc -l | tail -n 1",
            shell=True, capture_output=True, text=True)
        self.totalOpenedPIPEFiles = int(totalOpenedPIPEFiles.stdout)
        self.totalmemInUse = (self.memtotal + self.swaptotal) - (
            self.memfree + self.swapfree + self.bufferCacheMem)
        memInfo = subprocess.run("free",
                                 shell=True, capture_output=True, text=True)
        self.memInfo = memInfo.stdout
        diskInfo = subprocess.run("df -h",
                                  shell=True, capture_output=True, text=True)
        self.diskInfo = diskInfo.stdout

    def __init__(self):
        super(HealthMon, self).__init__()
        self._add_display_operation()
        self.setup_arguments()

    @property
    def model(self):
        return lpar_w.LPAR

    def _percentage_cal(self, mem, tmem):
        try:
            return (mem / tmem) * 100
        except ZeroDivisionError:
            return 0

    def _rest(self):
        adap = adp.Adapter()
        resp = adap.read_by_path(path='/rest/api/web/HeartBeat')
        if resp.status == 204:
            return "Passed"
        else:
            return "Failed"

    def _display(self):
        verbose = self.ec.parsed_args.verbose

        if (verbose):
            self.verbose()
        else:
            self.display()

    def _add_display_operation(self):

        crt_args = o_g.default_args(self,
                                    parent_models=[ms_w.System])

        crt_args.extend([op.Argument('--verbose',
                                     action='store_true',
                                     required=False)])

        self.add_operation(op.Operation(op.LIST_NAME, self._display,
                                        op_args=crt_args,
                                        op_help=help_hms.help_op,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def out_rows(self):
        self._command()
        o_rows = []
        Memory_Consumption = self._percentage_cal(self.totalmemInUse,
                                                  self.memtotal)
        Files_Opened = self._percentage_cal(self.totalOpenedFiles,
                                            self.totalLimitOfFiles)
        SSH_Files_Opened = self._percentage_cal(self.totalOpenedSSHFiles,
                                                self.totalOpenedFiles)
        PIPE_Files_Opened = self._percentage_cal(self.totalOpenedPIPEFiles,
                                                 self.totalOpenedFiles)
        REST_Status = self._rest()

        o_rows.append([str(Memory_Consumption),
                       str(Files_Opened),
                       str(SSH_Files_Opened),
                       str(PIPE_Files_Opened),
                       str(REST_Status)])
        return o_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS_VSN,
                                             self.out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': ("Health Monitor"),
            'n_tbl': po_table.get_string()}

    def verbose(self):
        self._command()
        print("Total Memory Information:\n", self.memInfo)
        print("Total Limit of files Information:", self.totalLimitOfFiles)
        print("Total Number of Opened Files:", self.totalOpenedFiles)
        print("Total Number of Opened SSH Files:", self.totalOpenedSSHFiles)
        print("Total Number of Opened PIPE Files:", self.totalOpenedPIPEFiles)
        print("Total Disk Information:\n", self.diskInfo)

    @property
    def help(self):
        return help_hms.help_op
