# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.help.help_logical_unit as help_lu
import pvmcli.util.exceptions as exc
from pvmcli.util import help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.util as util

from pypowervm.wrappers import storage as s
from pypowervm.tasks import storage as stor_task


PRI_DISP_FIELD_LABELS = [_("SSP"), _("Name"), _("Cap (GB)"), _("Type"),
                         _("Thin"), _("Clone"), _("In use")]

LU_TYPE_PREFIX = 'VirtualIO_'


class LogicalUnit(ObjectType):
    def __init__(self):
        super(LogicalUnit, self).__init__()

        self.add_display_operation(parent_models=[s.SSP],
                                   elem_wrp_funcs=[s.SSP.logical_units])
        self._add_upload_operation()
        self.add_delete_operation(parent_models=[s.SSP],
                                  elem_wrp_funcs=[s.SSP.logical_units])

        self.setup_arguments()

    @property
    def model(self):
        return s.LU

    def upload(self):
        data_arg, data_size_arg, vios_uuid = util.upload_setup(self)

        if self.ec.parsed_args.ssp:
            ssp_wrp = util.find_ssp_by_name(self.ec.sm.adapter,
                                            self.ec.parsed_args.ssp)
        else:
            ssp_wrps = s.SSP.get(self.ec.sm.adapter)
            if not ssp_wrps:
                raise exc.NoEntriesFoundError(object_type=s.SSP.schema_type)
            if len(ssp_wrps) > 1:
                raise exc.MultipleEntriesError(operation='upload')
            ssp_wrp = ssp_wrps[0]
            if s.SSP.schema_type not in util.name_cache:
                util.name_cache[s.SSP.schema_type] = {}
            if ssp_wrp.name not in util.name_cache[s.SSP.schema_type]:
                util.name_cache[s.SSP.schema_type][ssp_wrp.name] = ssp_wrp
        if not ssp_wrp:
            raise exc.NoEntriesFoundError(object_type=s.SSP.schema_type)

        with open(data_arg) as upstream:
            stor_task.upload_new_lu(
                vios_uuid, ssp_wrp, upstream, self.ec.parsed_args.name,
                data_size_arg, d_size=self.ec.parsed_args.size,
                sha_chksum=self.ec.parsed_args.checksum)

    def _add_upload_operation(self):
        args = o_g.default_args(self, include_output=False)
        args.extend([op.Argument('--vios', arg_help=help.UPLOAD_VIOS_HELP,
                                 arg_type=str, required=False),
                     op.Argument('--data', arg_help=help.UPLOAD_DATA_HELP,
                                 arg_type=str, required=False),
                     op.Argument('--data-size',
                                 arg_help=help.UPLOAD_DATA_SIZE_HELP,
                                 arg_type=int, required=False),
                     op.Argument('--name', arg_help=help.UPLOAD_NAME_HELP,
                                 arg_type=str, required=True),
                     op.Argument('--size', arg_help=help.UPLOAD_SIZE_HELP,
                                 arg_type=int, required=False),
                     op.Argument('--checksum',
                                 arg_help=help.UPLOAD_CHKSUM_HELP,
                                 arg_type=str, required=False),
                     op.Argument('--ssp', arg_help=help.UPLOAD_SSP_HELP,
                                 arg_type=str, required=False)])
        self.add_operation(op.Operation('upload', self.upload, op_args=args,
                                        object_model=self.model,
                                        query_wrps=False))

    def delete(self):
        parent_ssps = self.ec.elem_p_wrappers
        self.validate_wrps_single_elem(parent_ssps, self.name, op.DELETE_NAME)
        self.validate_wrps_single_elem(self.ec.wrappers,
                                       self.name, op.DELETE_NAME)

        ssp_lus = []
        for lu in parent_ssps[0].logical_units:
            if lu.udid != self.ec.wrappers[0].udid:
                ssp_lus.append(lu)

        parent_ssps[0].logical_units = ssp_lus
        parent_ssps[0].update()

    def lu_out_rows(self):
        out_rows = []

        name_to_ssp = {}
        for ssp_wrp in self.ec.elem_p_wrappers:
            name_to_ssp[ssp_wrp.name] = ssp_wrp

        udid_to_lu = {}
        for wrp in self.ec.wrappers:
            udid_to_lu[wrp.udid] = wrp

        for ssp_name in sorted(name_to_ssp):
            out_row = [ssp_name]
            ssp_wrp = name_to_ssp[ssp_name]

            name_to_lu = {}
            for wrp in ssp_wrp.logical_units:
                name_to_lu[wrp.name] = wrp

            for wrp_name in sorted(name_to_lu):
                wrp = name_to_lu[wrp_name]

                lu_type = self.dfield_val(wrp, "LogicalUnit.lu_type")
                if lu_type.startswith(LU_TYPE_PREFIX):
                    lu_type = lu_type[len(LU_TYPE_PREFIX):]

                clone_name = ""
                if wrp.udid in udid_to_lu:
                    clone_name = udid_to_lu[wrp.udid].name

                out_row.extend([wrp_name,
                                self.dfield_val(wrp, "LogicalUnit.capacity"),
                                lu_type,
                                self.dfield_val(wrp, "LogicalUnit.is_thin"),
                                clone_name,
                                self.dfield_val(wrp, "LogicalUnit.in_use")])
                out_rows.append(out_row)
                out_row = [""]
        return out_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.lu_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Logical Units"), 'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_lu.help_text
