# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.command.volume_group as vg
import pvmcli.help.help_logical_volume as help_lv
import pvmcli.util.exceptions as exc
from pvmcli.util import help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.util as util
import re
from pypowervm.tasks import storage as stor_task
from pypowervm.wrappers import storage as s
import pypowervm.wrappers.virtual_io_server as vios_w


class LogicalVolume(ObjectType):
    def __init__(self):
        super(LogicalVolume, self).__init__()

        self.remap_schema(s.VDisk.schema_type, vg.LV_SCHEMA)

        self.add_display_operation(parent_models=[vios_w.VIOS],
                                   child_models=[s.VG],
                                   elem_wrp_funcs=[s.VG.virtual_disks])
        self._add_create_operation()
        self._add_upload_operation()
        self.add_update_operation(parent_models=[vios_w.VIOS],
                                  child_models=[s.VG],
                                  elem_wrp_funcs=[s.VG.virtual_disks])
        self.add_delete_operation(parent_models=[vios_w.VIOS],
                                  child_models=[s.VG],
                                  elem_wrp_funcs=[s.VG.virtual_disks],
                                  child_id_req=True)

        self.setup_arguments()

    def delete(self):
        self.validate_wrps_single_elem(self.ec.wrappers, s.VDisk.schema_type,
                                       op.DELETE_NAME)
        self.validate_wrps_single_elem(self.ec.elem_c_wrappers,
                                       s.VG.schema_type, op.DELETE_NAME)
        self.ec.elem_c_wrappers[0].virtual_disks.remove(self.ec.wrappers[0])
        self.ec.elem_c_wrappers[0].update()

    def _add_create_operation(self):
        name_help = _("The name of this logical volume. Required.")
        size_help = _("A float representing the size, in GB, of this volume. "
                      "Required.")
        crt_args = o_g.default_args(self, parent_models=[vios_w.VIOS],
                                    child_models=[s.VG], include_output=False)
        crt_args.extend([op.Argument('--name', arg_help=name_help,
                                     required=True, arg_type=str),
                         op.Argument('--size', arg_help=size_help,
                                     required=True, arg_type=float)])

        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[vios_w.VIOS],
                                        child_models=[s.VG],
                                        elem_wrp_funcs=[s.VG.virtual_disks]))

    def create(self):
        ''' Create operation processing function. '''
        self.validate_wrps_single_elem(self.ec.elem_c_wrappers,
                                       s.VG.schema_type,
                                       op.CREATE_NAME)
        sizeV = float(self.ec.parsed_args.size)
        if sizeV <= 0:
            raise exc.InvalidArgumentValue(value=sizeV, arg="--size")

        # Validation to Check the Special Characters on LogicalVolume
        validate_name = re.compile('[`,.;=@!#$%^&*()<>?/|}{~:\\\\]')
        if validate_name.search(self.ec.parsed_args.name) is None:

            self.ec.elem_c_wrappers[0].virtual_disks.append(
                s.VDisk.bld(adapter=self.ec.sm.adapter,
                            name=self.ec.parsed_args.name,
                            capacity=self.ec.parsed_args.size))

            self.ec.elem_c_wrappers[0].update()

        else:
            raise exc.InvalidLVArgumentValue()

    @property
    def model(self):
        return s.VDisk

    @property
    def name(self):
        return vg.LV_SCHEMA

    def upload(self):
        data_arg, data_size_arg, vios_uuid = util.upload_setup(self)

        vg_wrp = util.find_vg_by_name(
            self.ec.sm.adapter, self.ec.parsed_args.vg, vios_uuid)
        if not vg_wrp:
            raise exc.NoEntriesFoundError(object_type=s.VG.schema_type)

        if self.ec.parsed_args.size is not None:
            sizeV = float(self.ec.parsed_args.size)
            if sizeV <= 0:
                raise exc.InvalidArgumentValue(value=sizeV, arg="--size")

        with open(data_arg, 'rb') as upstream:
            stor_task.upload_new_vdisk(
                self.ec.sm.adapter, vios_uuid, vg_wrp.uuid, upstream,
                self.ec.parsed_args.name, data_size_arg,
                d_size=self.ec.parsed_args.size,
                sha_chksum=self.ec.parsed_args.checksum)

    def _add_upload_operation(self):
        args = o_g.default_args(self, include_output=False)
        args.extend([op.Argument('--vios', arg_help=help.UPLOAD_VIOS_HELP,
                                 arg_type=str, required=False),
                     op.Argument('--data', arg_help=help.UPLOAD_DATA_HELP,
                                 arg_type=str, required=False),
                     op.Argument('--data-size',
                                 arg_help=help.UPLOAD_DATA_SIZE_HELP,
                                 arg_type=int, required=False),
                     op.Argument('--name', arg_help=help.UPLOAD_NAME_HELP,
                                 arg_type=str, required=True),
                     op.Argument('--size', arg_help=help.UPLOAD_SIZE_HELP,
                                 arg_type=int, required=False),
                     op.Argument('--checksum',
                                 arg_help=help.UPLOAD_CHKSUM_HELP,
                                 arg_type=str, required=False),
                     op.Argument('--vg', arg_help=help.UPLOAD_VG_HELP,
                                 arg_type=str, required=True)])
        self.add_operation(op.Operation('upload', self.upload, op_args=args,
                                        object_model=self.model,
                                        query_wrps=False))

    def pri_output(self):
        po_table = self.resize_pri_out_table(
            vg.PRI_DISP_FIELD_LABELS, vg.vg_out_rows(self,
                                                     self.ec.elem_c_wrappers))

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Logical Volumes"), 'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_lv.help_text
