# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import argparse
from pvmcli.util.i18n import _
import pvmcli.util.util as util


# Common Operation Names
CREATE_NAME = 'create'
REMAP_NAME = 'remap'
LIST_NAME = 'list'
UPDATE_NAME = 'update'
DELETE_NAME = 'delete'
POWER_ON_NAME = 'power-on'
POWER_OFF_NAME = 'power-off'
RESTART_NAME = 'restart'
NET_INSTALL = 'network-install'

# Common Argument Names
DF_NAME = '--display-fields'
SF_NAME = '--set-fields'
ID_NAME = '--object-id'
PARENT_ID_NAME = '--parent-id'
CHILD_ID_NAME = '--child-id'

# Common Argument Group Names
REQUIRED_ARGUMENTS_GROUP = _('required arguments')


class Operation(object):
    def __init__(self, op_name, op_func, op_args, object_model,
                 parent_models=None, child_models=None, elem_wrp_funcs=None,
                 op_help='', suppress_help=False, query_wrps=True):
        ''' Operation class to allow an operation to be easily constructed.
            This is used by ObjectType to handle operation processing as
            generically as possible.
        '''
        self._op_name = op_name
        self._op_func = op_func
        self._object_model = object_model
        self._parent_models = parent_models
        self._child_models = child_models
        self._elem_wrp_funcs = elem_wrp_funcs
        self._op_args = {}
        self._op_help = op_help
        self._suppress_help = suppress_help
        self._query_wrps = query_wrps
        self.subparser = None

        # Create a dict from an argument name to its argument object.
        for op_arg in op_args:
            self._op_args[op_arg.arg_name] = op_arg

    @property
    def suppress_help(self):
        ''' Hide (True) or show (False) this operation in the parent's help.'''
        return self._suppress_help

    @property
    def op_name(self):
        ''' Name of the operation (e.g. create, update, list, delete) '''
        return self._op_name

    @property
    def op_func(self):
        ''' Function to call to process this operation '''
        return self._op_func

    @property
    def op_help(self):
        ''' Help text string describing an operation in ObjectType help. '''
        return self._op_help

    def add_arg(self, op_arg):
        ''' Add an argument to the arg_name -> Argument dictionary for this
            operation.
        '''
        self._op_args[op_arg.arg_name] = op_arg

    @property
    def op_args(self):
        ''' Returns a dictionary of Argument.arg_name -> Argument object for
            all arguments associated with this operation.
        '''
        return self._op_args

    def get_disp_field(self, model, field_name):
        ''' Return a display field based on the provided model and field_name.

        :param model: The model of the wrapper with which the DisplayField
                      is associated.
        :param field_name: The name of the field on the wrapper.
        '''
        return self.op_args[DF_NAME].fields['%(m)s.%(n)s'
                                            % {'m': model.schema_type,
                                               'n': field_name}]

    def set_d_field_label(self, model, field_name, label):
        ''' Set label on the DisplayField with provided model and name.

        :param model: The model class of the display field.
        :param field_name: The name of the display field.
        :param label: The new label to set.
        '''
        self.get_disp_field(model, field_name).label = label

    def set_d_field_out_func(self, model, field_name, out_func):
        ''' Set the output function for the DisplayField.

        :param model: The model class of the display field.
        :param field_name: The name of the display field.
        :param out_func: The new output function to set.
        '''
        self.get_disp_field(model, field_name).out_func = out_func

    @property
    def id_fields(self):
        ''' ID fields to search against for this operation. '''
        return self.op_args[ID_NAME].fields

    @property
    def parent_id_fields(self):
        ''' Parent ID fields to search against for this operation. '''
        return self.op_args[PARENT_ID_NAME].fields

    @property
    def child_id_fields(self):
        ''' Child ID fields to search against for this operation. '''
        return self.op_args[CHILD_ID_NAME].fields

    @property
    def object_model(self):
        ''' pypowervm wrapper model of this operations object type. Used for
            session adapter processing and object type wrapper cache retrieval
        '''
        return self._object_model

    @property
    def obj_schema(self):
        ''' The schema type of this operation's object model. '''
        return self._object_model.schema_type

    @property
    def parent_models(self):
        ''' pypowervm wrapper models of this operation's parent type. Used for
            session adapter processing and object type wrapper cache retrieval
        '''
        return self._parent_models

    @property
    def parent_schemas(self):
        ''' The schema types of this operation's parent models. '''
        return [p_m.schema_type for p_m in self.parent_models]

    @property
    def child_models(self):
        ''' pypowervm wrapper models of this operation's child types. Used for
            session adapter processing and object type wrapper cache retrieval
        '''
        return self._child_models

    @property
    def child_schemas(self):
        ''' The schema types of this operation's child models. '''
        return [c_m.schema_type for c_m in self.child_models]

    @property
    def elem_wrp_funcs(self):
        ''' Element functions to call that will return a WrapperElemList object
            that represents all elements for this operation.

            This is a list of functions to be called, in order, to retrieve
            the elements that are needed. This is necessary if the object type
            represents an element that is nested within other elements.
        '''
        return self._elem_wrp_funcs

    @property
    def query_wrps(self):
        ''' Specifies if wrappers are eagerly queried in object type execution.
        '''
        return self._query_wrps


class Argument(object):
    def __init__(self, arg_name, short_name=None, required=None,
                 arg_help=None, arg_type=None, metavar=None,
                 nargs=None, choices=None, action=None,
                 group=None, default=None, fields=[]):
        ''' Wrapper for argparse package arguments. This is necessary
            to allow us to build operation/argument combinations
            generically. See argparse documentation for more details.
        '''
        self._arg_name = arg_name
        self._arg_short_name = short_name
        self._arg_required = required
        if arg_help != argparse.SUPPRESS:
            arg_help = util.utf8_str(arg_help)
        self._arg_help = arg_help
        self._arg_type = arg_type
        self._arg_metavar = util.utf8_str(metavar)
        self._arg_nargs = nargs
        self._arg_choices = choices
        self._arg_action = action
        # If an argument has no group and is required,
        # add to Required arguments group.
        if group is None and required:
            self._arg_group = util.utf8_str(REQUIRED_ARGUMENTS_GROUP)
        else:
            self._arg_group = group
        self._arg_default = default
        self._fields = {}
        self.add_fields(fields)

    def add_field(self, field):
        ''' Add a constructed Field object to this Argument.

        :param field: The constructed Field object to add.
        '''
        self._fields[field.field_name] = field

    def add_fields(self, fields):
        ''' Add a list of Field objects to this Argument.

        :param fields: A list of constructed Field objects.
        '''
        for f in fields:
            self._fields[f.field_name] = f

    @property
    def fields(self):
        ''' The Fields associated with this argument. '''
        return self._fields

    @property
    def arg_name(self):
        ''' Name of this argument, generally of the form --arg '''
        return self._arg_name

    @property
    def short_name(self):
        ''' Short representation of this arg, generally of the form -a '''
        return self._arg_short_name

    @property
    def required(self):
        ''' Returns a boolean representing if this arg is required. '''
        return self._arg_required

    @property
    def arg_help(self):
        ''' Returns a string representing this argument's help text. '''
        return self._arg_help

    @property
    def arg_type(self):
        ''' Returns a type representing the type of this argument '''
        return self._arg_type

    @property
    def metavar(self):
        ''' Returns a string for argparse's metavar attribute for this arg '''
        return self._arg_metavar

    @property
    def nargs(self):
        ''' Returns a number representing the number of valid values for this
            argument. See argparse documentation for more details.
        '''
        return self._arg_nargs

    @property
    def choices(self):
        ''' Returns a list of valid choices for this argument. None if any
            choice is valid.
        '''
        return self._arg_choices

    @property
    def action(self):
        ''' Returns a string representing the action to take on this argument.
            See argparse documentation for more details.
        '''
        return self._arg_action

    @property
    def group(self):
        ''' Returns a string argument group title. argparse will automatically
            place arguments in appropriate groups if this is not specified.
        '''
        return self._arg_group

    @property
    def default(self):
        ''' Returns a default value for this argument. '''
        return self._arg_default
