# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.util.exceptions as ex
import pvmcli.util.help as h
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
from pypowervm.wrappers import pmem as p
from pypowervm.wrappers import managed_system as ms_w
import pvmcli.help.help_pmemv as help_pmemv
import pvmcli.help.help_pmemd as help_pmemd
PRI_DISP_FIELD_LABELS_VOL = [_("NAME"), _("ID"), _("SIZE"), _("CUR_SIZE"),
                             _("LPAR"), _("LPAR_ID"), _("UUID"),
                             _("AFFINITY"), _("DEVICE")]
PRI_DISP_FIELD_LABELS_DEV = [_("DRC/Device Index"), _("TYPE"), _("STATUS"),
                             _("BLOCK_SIZE"), _("TOTAL_SIZE"), _("FREE_SIZE"),
                             _("MAX_VOL_LIMITE"), _("CUR_VOL"), _("PHYS_LOC"),
                             _("SER_NUM")]


class PMEMVolume(ObjectType):
    def __init__(self):
        super(PMEMVolume, self).__init__()

        self.add_display_operation(parent_models=[ms_w.System])
        self.add_create_operation()
        self.add_delete_operation(parent_models=[ms_w.System])
        self.setup_arguments()

    @property
    def model(self):
        return p.VirtualPMEMVolume

    def out_rows(self):
        o_rows = []
        for wrp in self.ec.wrappers:
            name = "VirtualPersistentMemoryVolume.name"
            volume_id = "VirtualPersistentMemoryVolume.volume_id"
            size = "VirtualPersistentMemoryVolume.size"
            cur_size = "VirtualPersistentMemoryVolume.cur_size"
            assoc_part_n = "VirtualPersistentMemoryVolume.assoc_partition_name"
            assoc_part_id = "VirtualPersistentMemoryVolume.assoc_partition_id"
            pmem_vol_uuid = "VirtualPersistentMemoryVolume.pmem_vol_uuid"
            affinity = "VirtualPersistentMemoryVolume.affinity"
            o_rows.append([self.dfield_val(wrp, name),
                           self.dfield_val(wrp, volume_id),
                           self.dfield_val(wrp, size),
                           self.dfield_val(wrp, cur_size),
                           self.dfield_val(wrp, assoc_part_n),
                           self.dfield_val(wrp, assoc_part_id),
                           self.dfield_val(wrp, pmem_vol_uuid),
                           self.dfield_val(wrp, affinity),
                           'dram'])

        return o_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS_VOL,
                                             self.out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Virtual Persistent Memory Volume"),
            'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_pmemv.help_text

    def add_create_operation(self):
        name_help = _("Required. The name of the Volume.")
        size_help = _("Required. The size of Volume.")
        lpar_help = _("Required. The associated lpar id.")
        affinity_help = _("Required. The affinity of the Volume.")
        crt_args = o_g.default_args(self, parent_models=[ms_w.System])
        crt_args.extend([op.Argument('--name', arg_type=str,
                                     required=True,
                                     arg_help=h.wrap_help(name_help)),
                         op.Argument('--size', arg_type=int,
                                     required=True,
                                     arg_help=h.wrap_help(size_help)),
                         op.Argument('--lpar_id', arg_type=int,
                                     required=True,
                                     arg_help=h.wrap_help(lpar_help)),
                         op.Argument('--affinity', arg_type=str,
                                     required=False,
                                     choices=['True', 'False'],
                                     arg_help=h.wrap_help(affinity_help))])

        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def create(self):
        vv_name = self.ec.parsed_args.name
        vv_size = self.ec.parsed_args.size
        if vv_size <= 0:
            raise ex.InvalidArgumentValue(value=vv_size, arg="--size")

        vv_lpar = self.ec.parsed_args.lpar_id
        vv_affinity = 'False'
        if self.ec.parsed_args.affinity:
            vv_affinity = self.ec.parsed_args.affinity

        new_vol = p.VirtualPMEMVolume.bld(adapter=self.ec.sm.adapter,
                                          lpar_id=vv_lpar, name=vv_name,
                                          size=vv_size, affinity=vv_affinity)

        c_pv = new_vol.create(parent_type=self.ec.exec_op.parent_schemas[0],
                              parent_uuid=self.ec.msys.uuid)

        self.ec.wrappers = [c_pv]
        # Only display if a user specifies specific display fields
        if (self.ec.parsed_args.display_fields
                or self.ec.parsed_args.all_out):
            self.display()


class PMEMDevice(ObjectType):
    def __init__(self):
        super(PMEMDevice, self).__init__()

        self.add_display_operation(parent_models=[ms_w.System])
        self.setup_arguments()

    @property
    def model(self):
        return p.PersistentMemoryDevice

    def out_rows(self):
        out_rows = []
        for wrp in self.ec.wrappers:
            drc_index = "PersistentMemoryDevice.drc_index"
            dtype = "PersistentMemoryDevice.type"
            status = "PersistentMemoryDevice.status"
            blocksize = "PersistentMemoryDevice.blocksize"
            totalsize = "PersistentMemoryDevice.totalsize"
            freesize = "PersistentMemoryDevice.freesize"
            max_num_volumes = "PersistentMemoryDevice.max_num_volumes"
            cur_num_volumes = "PersistentMemoryDevice.cur_num_volumes"
            pys_loc = "PersistentMemoryDevice.pys_loc"
            serial_number = "PersistentMemoryDevice.serial_number"

            out_rows.append([self.dfield_val(wrp, drc_index),
                             self.dfield_val(wrp, dtype),
                             self.dfield_val(wrp, status),
                             self.dfield_val(wrp, blocksize),
                             self.dfield_val(wrp, totalsize),
                             self.dfield_val(wrp, freesize),
                             self.dfield_val(wrp, max_num_volumes),
                             self.dfield_val(wrp, cur_num_volumes),
                             self.dfield_val(wrp, pys_loc),
                             self.dfield_val(wrp, serial_number)])

        return out_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS_DEV,
                                             self.out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Persistent Memory Device"),
            'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_pmemd.help_text
