# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

import pvmcli.util.exceptions as e
from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
from pypowervm.wrappers import virtual_serial_number as v
from pypowervm.tasks import virtual_serial_number as v_t
from pypowervm.wrappers import managed_system as ms_w
import pvmcli.help.help_vsn as help_vsn

PRI_DISP_FIELD_LABELS_VSN = [_("VSN"), _("LPAR ID"), _("AUTO ASSIGN")]


class VirtualSN(ObjectType):
    def __init__(self):
        super(VirtualSN, self).__init__()

        self.add_display_operation(parent_models=[ms_w.System])
        self._add_delete_operation()
        self._add_transfer_operation()
        self.setup_arguments()

    @property
    def model(self):
        return v.VirtualSerialNumber

    def out_rows(self):
        o_rows = []
        for wrp in self.ec.wrappers:
            vsn = "VirtualSerialNumberInformation.vsn"
            lpar_id = "VirtualSerialNumberInformation.assoc_partition_id"
            auto_assign = "VirtualSerialNumberInformation.auto_assign"

            o_rows.append([self.dfield_val(wrp, vsn),
                           self.dfield_val(wrp, lpar_id),
                           self.dfield_val(wrp, auto_assign)])

        return o_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS_VSN,
                                             self.out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Virtual Serial Numbers"),
            'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_vsn.help_text

    def delete(self):
        reusevsn = self.ec.parsed_args.reusevsn
        v_t.delete_vsn(vsn=self.ec.parsed_args.object_id.split("=")[1],
                       reusevsn=reusevsn)

    def _add_delete_operation(self):
        reusevsn_help = _("Reuse the VSN after deleting the VSN.")
        args = o_g.default_args(self, include_output=False)
        args.extend([op.Argument('--reusevsn', required=False,
                                 arg_type=bool, arg_help=reusevsn_help,
                                 action="store_true")])
        self.add_operation(op.Operation('delete', self.delete, op_args=args,
                                        object_model=self.model,
                                        query_wrps=False))

    def transfer(self):
        mgmt_usr = self.ec.parsed_args.mgmt_usr
        mgmt_svr = self.ec.parsed_args.mgmt_svr
        sys = self.ec.parsed_args.sys
        lpar_id = self.ec.parsed_args.lpar_id
        self.validate_transfer(self.ec.parsed_args.object_id,
                               "id=<value> | name=<value>")
        v_t.transfer_vsn(vsn=self.ec.parsed_args.object_id.split('=')[1],
                         mgmt_usr=mgmt_usr, mgmt_svr=mgmt_svr, sys=sys,
                         lpar_id=lpar_id)

    def validate_transfer(self, val, valid_fields):
        split_val = val.split("=", 1)
        args = ['id', 'name']
        if split_val[0] not in args or (split_val[-1] == ''):
            raise e.InvalidIDArgumentError(id_arg='-i',
                                           valid_fields=valid_fields)

    def _add_transfer_operation(self):
        args = o_g.default_args(self, include_output=False)
        tgt_mgmt_srv_help = _("Required. The IP of the PowerVM management "
                              "platform managing the target host.")
        tgt_mgmt_usr_help = _("The user ID to use on the target "
                              "PowerVM management platform.")
        tgt_mgmt_sys_help = _("The name of target managed system to"
                              " transfer VNS")
        tgt_mgmt_lparid_help = _("The ID of Target Lpar given to"
                                 "transfer the VSN")

        args.extend([op.Argument('--mgmt-usr', arg_help=tgt_mgmt_usr_help,
                                 arg_type=str, required=False),
                     op.Argument('--mgmt-svr', arg_help=tgt_mgmt_srv_help,
                                 arg_type=str, required=True),
                     op.Argument('--sys', arg_help=tgt_mgmt_sys_help,
                                 arg_type=str, required=False),
                     op.Argument('--lpar-id', arg_help=tgt_mgmt_lparid_help,
                                 arg_type=str, required=False)])

        self.add_operation(op.Operation('transfer', self.transfer,
                                        op_args=args,
                                        object_model=self.model,
                                        query_wrps=False))
