import sys
import subprocess
from pvmcli.net_conf import utils as utils
from pvmcli.net_conf import messageCatalog as MSGCATALOG

MGMT_VSWITCH_VLAN_ID = "4094"
SUBNET_MASK = "255.255.128.0"
VIOS_CMD = "viosvrcmd"
SET_PRIVATE_IP_SCRIPT_PATH = "/usr/sbin/setprivatenetwork.sh"


class VIOShandler:

    # Configures devices in the Virtual I/O Server
    def configure_device(self, ):
        command = 'cfgdev'
        return command

    # Fetch hostname of VIOS
    def get_hostname(self,):
        command = 'hostname'
        return command

    # Fetch empty interface for setting private IP
    def get_empty_interface(self,):
        command = '"lstcpip -interfaces" | grep -i "detach" | grep -i "en"'
        return command

    # Fetch inetrafce with exitsing private IPs
    def get_assigned_interface(self,):
        command = "'lstcpip -interfaces' | grep -i 'up' | \
grep -i '192.168.128.' | awk '{print $2, $5}'"
        return command

    # Filter ethernet adapters with VLAN ID 4094
    def get_mgmt_switch_interface_by_vlanId(self, vEth, vlan_id):
        command = '"entstat -all "'+str(
            vEth)+' | grep -i "Port VLAN" | grep -i "'+vlan_id+'"'
        return command

    # Filter management switch by switch name
    def get_mgmt_switch_interface_by_switch(self, vEth, switchname):
        command = '"entstat -all "'+str(
            vEth)+' | grep -i "Switch" | grep -i "'+switchname+'"'
        return command

    # Add IP address to interface
    def add_network(self, hostname, ip_add, netmask, interface):
        command = '"mktcpip -hostname '+hostname+' -inetaddr '+str(
            ip_add)+' -interface '+interface+' -netmask '+netmask+' -start"'
        return command


class MGMT_SWITCH:

    def __init__(self, log_path, MGMT_VSWITCH_NAME, NL_PRIVATE_IP):
        self.log_path = log_path
        self.switchname = MGMT_VSWITCH_NAME
        self.nl_pvt_ip = NL_PRIVATE_IP
        self.pvid = MGMT_VSWITCH_VLAN_ID
        self.viosHandler = VIOShandler()

    def log_cmd(self, info):
        with open(self.log_path, 'w+') as out_file:
            out_file.write(str(info)+'\n')

    # Execute all CLI commands
    def execute_cmd(self, cmd):
        self.log_cmd(cmd)
        res = subprocess.Popen(
            [cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out, err = res.communicate()
        out = out.decode("utf-8")
        err = err.decode("utf-8")
        self.log_cmd(out)
        self.log_cmd(err)
        returncode = res.returncode
        return returncode, out, err

    # List of VIOS
    def get_vios_list(self, ):
        cmd = "pvmctl vios list | grep -i running"
        returncode, out, err = self.execute_cmd(cmd)
        res = str(out).split("\n")
        vios_info = []
        for info in res:
            data = info.split("|")
            if len(data) > 1:
                vios_info.append({'vios_name': data[1], 'id': data[2]})
        return vios_info

    # Create management vswitch for VIOS and Novalink on CEC
    def create_mgmt_vswitch(self, ):
        cmd = 'pvmctl virtualswitch create --name ' + str(self.switchname) + \
            ' --mode vepa'
        code, out, err = self.execute_cmd(cmd)
        return code, out, err

    # Create ethernet adapters and attached to management vswitch
    # with port VLAN ID 4094
    def create_eth(self, vm_id):
        vm_id = str(vm_id).strip()
        cmd = 'pvmctl eth create --vswitch '+self.switchname+' --pvid '+str(
            self.pvid)+' -p id='+str(vm_id)
        code, out, err = self.execute_cmd(cmd)
        return code, out, err

    # apply changes after creating ethernet adapters using cfgdev
    def configure_device_vios(self, vios_id):
        viosCmd = self.viosHandler.configure_device()
        cmd = VIOS_CMD+' --id '+str(vios_id)+' -c '+viosCmd
        code, out, err = self.execute_cmd(cmd)

    def get_vios_hostname(self, vios_id):
        viosCmd = self.viosHandler.get_hostname()
        cmd = VIOS_CMD+' --id '+str(vios_id)+' -c '+viosCmd
        code, out, err = self.execute_cmd(cmd)
        hostname = False
        if code == 0:
            hostname = out.strip()
        else:
            print(out)
            print(err)
        return hostname

    # Fetch ineterface with VLAN ID 4094
    # Fetch vswitch for management vswitch
    # Validate the interface must be connected to same management vswitch
    def filter_mgmt_switch_adapter(self, vios_id, interface):
        pvid = self.pvid
        viosCmd = self.viosHandler.get_mgmt_switch_interface_by_vlanId(
            interface, pvid)
        cmd = VIOS_CMD+' --id '+vios_id+' -c '+viosCmd
        code, out, err = self.execute_cmd(cmd)
        if code == 0:
            vlans = out.split(':')
            vLanId = vlans[1].replace("\n", "").strip()
            if int(vLanId) == int(pvid):
                switchname = self.switchname
                viosCmd = self.viosHandler.get_mgmt_switch_interface_by_switch(
                    interface, switchname)
                cmd = VIOS_CMD+' --id '+vios_id+' -c '+viosCmd
                code, out, err = self.execute_cmd(cmd)
                if code == 0:
                    vlans = out.split(':')
                    mgmt_vsw = vlans[1].replace("\n", "").strip()
                    if mgmt_vsw == switchname:
                        return True
            else:
                return False
        else:
            print(out)
            return False

    # Fetch empty interface
    # Filter interface which is connect to management switch
    def get_empty_interface_vios(self, vios_id):
        viosCmd = self.viosHandler.get_empty_interface()
        cmd = VIOS_CMD+' --id '+vios_id+' -c '+viosCmd
        code, out, err = self.execute_cmd(cmd)
        interface = False
        if code == 0:
            out = out.split("\n")
            for line in out:
                adpt = line.split(' ', 1)
                interface = adpt[0].strip()
                res = self.filter_mgmt_switch_adapter(vios_id, interface)
                if res:
                    print(MSGCATALOG.VIOS_EMPTY_INTERFACE_FOUND.format(
                        interface))
                    return interface
        else:
            for line in out:
                print(line)
        return interface

    # Add private IP on interface
    def add_network_on_vios(self, vios_id, hostname, ip_add, nmask, interface):
        viosCmd = self.viosHandler.add_network(
            hostname, ip_add, nmask, interface)
        cmd = VIOS_CMD+' --id '+vios_id+' -c '+viosCmd
        code, out, err = self.execute_cmd(cmd)
        if code == 0:
            print(MSGCATALOG.VIOS_ADD_NETWORK_SUCCESS.format(
                  ip_add, vios_id, interface))
            return True
        else:
            print(out)
        return False

    # Get all ethernet list filtered by MAC and management switch
    def get_eth_list(self, mac):
        cmd = "pvmctl eth list | grep -i "+mac+" | grep "+str(
            self.switchname)+" | awk '{print $4}'"
        code, out, err = self.execute_cmd(cmd)
        if code == 0:
            if out:
                return out.strip()
        return False

    # Validate the existing VIOS private network
    # If private IP already setup on VIOS, skip setting private IP again.
    def validate_vios_nw(self, vios_id, pvt_ip_list):
        viosCmd = self.viosHandler.get_assigned_interface()
        cmd = VIOS_CMD+' --id '+vios_id+' -c '+viosCmd
        code, out, err = self.execute_cmd(cmd)
        if code == 0:
            res = out.split("\n")
            for line in res:
                info = line.split(" ")
                if len(info) > 1:
                    ip = info[0]
                    mac = info[1]
                    if ip in pvt_ip_list:
                        mac = mac.replace(":", "")
                        res = self.get_eth_list(mac)
                        if res:
                            print(MSGCATALOG.VIOS_IP_ALREADY_EXIST.format(
                                ip, self.switchname, vios_id))
                            return False
            return True
        return False

    # Main function for setting Private IP on VIOS
    # Validating exiting VIOS network
    # If private IP not setup, create new ethernet adapter
    # Set private IP on newly created adapter.
    def configure_network_vios(self, vios_id, ip_add, pvt_ip_list):
        print(MSGCATALOG.VIOS_VALIDATE_EXISTING_NETWORK)
        res = self.validate_vios_nw(vios_id, pvt_ip_list)
        if res:
            print(MSGCATALOG.VIOS_CREATE_ETH.format(vios_id))
            code, out, err = self.create_eth(vios_id)
            if code == 0:
                print(MSGCATALOG.VIOS_ETHERNET_CREATED_SUCCESS.format(
                    vios_id))
                print(MSGCATALOG.VIOS_CONFIGURE_DEVICE)
                self.configure_device_vios(vios_id)
                hostname = self.get_vios_hostname(vios_id)
                if hostname:
                    interface = self.get_empty_interface_vios(vios_id)
                    if interface:
                        nmask = SUBNET_MASK
                        print(MSGCATALOG.VIOS_ADD_NETWORK.format(
                            ip_add, interface))
                        res = self.add_network_on_vios(
                            vios_id, hostname, ip_add, nmask, interface)
                        return res
                else:
                    print(MSGCATALOG.VIOS_HOSTNAME_NOT_FOUND)
                    return False
            else:
                print(out)
                return False
        else:
            return False

    # For getting Novalink ID
    def get_nl_id(self, ):
        cmd = "pvmctl vm list --display-fields \
LogicalPartition.id LogicalPartition.is_mgmt_partition | grep -i true"
        code, out, err = self.execute_cmd(cmd)
        res = str(out).split(",")[0]
        nl_id = str(res).split("=")[1]
        return nl_id

    # Filter MAC address of Novalink ethernet adapter connect to
    # management switch with Port VLAN ID 4094
    def get_nl_eth_mac(self, ):
        cmd = "pvmctl vea list | grep -i "+self.switchname+" | grep -i "+str(
            self.pvid)+" | awk '{print $6, $8}'"
        code, out, err = self.execute_cmd(cmd)
        res = str(out).split("\n")
        eth_info = []
        for info in res:
            data = info.split(" ")
            if len(data) > 1:
                eth_info.append({'nl_id': data[0], 'mac_add': data[1].lower()})
        return eth_info

    # Get interface list of Novalink using ifconfig
    # Filter interface with Novalink Private IP
    # Return MAC address of interface
    def get_mac_ifconfig(self, ):
        cmd = "ifconfig -a"
        code, out, err = self.execute_cmd(cmd)
        res = str(out).split("\n")
        eth_info = []
        eth_no = False
        mac = False
        pvt_ip = False
        for info in res:
            if "flags" in info:
                eth_no = info.split(":")[0]
                pvt_ip = False
                mac = False
            nl_pvt_ip = self.nl_pvt_ip
            if nl_pvt_ip in info:
                pvt_ip = True
            if not pvt_ip:
                if "ether" in info:
                    info = info.strip()
                    mac = info.split(" ")[1]
                if eth_no and mac:
                    mac = mac.replace(":", "")
                    eth_info.append({'eth_no': eth_no, 'mac_add': mac})
                    eth_no = False
        return eth_info

    def get_os(self, ):
        cmd = "cat /etc/os-release"
        code, out, err = self.execute_cmd(cmd)
        res = str(out)
        return res

    # Setup private IP on novalink
    # execute script for setting IP based on OS name (rhel, ubuntu)
    def add_network_novalink(self, mac_add):
        os_name = self.get_os().lower()
        if 'rhel' in os_name or 'red hat' in os_name:
            os_name = "rhel"
        if 'ubuntu' in os_name:
            os_name = "ubuntu"
        # res = subprocess.call([SET_PRIVATE_IP_SCRIPT_PATH,
        # os_name, mac_add, self.nl_pvt_ip])

    # Validate Novalink private network
    def validate_nl_nw(self, nl_pvt_ip):
        cmd = "ifconfig -a"
        code, out, err = self.execute_cmd(cmd)
        res = str(out).split("\n")
        mac_info = []
        pvt_ip = False
        for info in res:
            if nl_pvt_ip in info:
                pvt_ip = True
            if pvt_ip:
                if "ether" in info:
                    info = info.strip()
                    mac = info.split(" ")[1]
                    mac_info.append(mac)
                    pvt_ip = False
        return mac_info

    # Main function for Configuring Novalink private network
    # Validate exisiting private network for Novalink
    # If private IP is not setup, create new ethernet adapter
    # Get MAC address of newly created interface
    # Add private network on Novalink
    def configure_network_nl(self, ):
        print(MSGCATALOG.NOVALINK_VALIDATE_EXISTING_NETWORK)
        mac_list = self.validate_nl_nw(self.nl_pvt_ip)
        for mac in mac_list:
            res = self.get_eth_list(mac.replace(":", ""))
            if res:
                print(MSGCATALOG.NOVALINK_IP_ALREADY_EXIST.format(
                    self.nl_pvt_ip, self.switchname))
                return False
        nl_id = self.get_nl_id()
        code, out, err = self.create_eth(nl_id)
        if code == 0:
            print(MSGCATALOG.NOVALINK_ETHERNET_CREATED_SUCCESS.format(nl_id))
            mac_list = self.get_nl_eth_mac()
            ifconfig_eth = self.get_mac_ifconfig()
            for info in mac_list:
                if int(nl_id) == int(info['nl_id']):
                    mac = info['mac_add']
                    for ifconfig_info in ifconfig_eth:
                        if mac == ifconfig_info['mac_add']:
                            eth_no = ifconfig_info['eth_no']
                            print(MSGCATALOG.NOVALINK_ETH_FOUND.format(eth_no))
                            print(MSGCATALOG.NOVALINK_ETH_MAC.format(mac))
                            self.add_network_novalink(eth_no)
                            print(MSGCATALOG.NOVALINK_IP_ADDED.format(eth_no))
                            return True


# Main Function
def main(MGMT_VSWITCH_NAME, VIOS_PRIVATE_IP_START, NL_PRIVATE_IP):
    log_path = utils.LOG_FILE_PATH
    mgmt_obj = MGMT_SWITCH(log_path, MGMT_VSWITCH_NAME, NL_PRIVATE_IP)
    # Create management switch
    print(MSGCATALOG.CREATE_MGMT_SWITCH.format(MGMT_VSWITCH_NAME))
    code, out, err = mgmt_obj.create_mgmt_vswitch()
    if code != 0:
        print(err)
    else:
        print(MSGCATALOG.MGMT_SWITCH_CREATE_SUCCESS)

    # Fetch VIOS list
    # Prepare a list of private IPs for VIOS based on numeber of VIOS
    print(MSGCATALOG.VIOS_GET_VIOS_LIST)
    res = mgmt_obj.get_vios_list()
    vios_pvt_ip = VIOS_PRIVATE_IP_START
    ip_count = 0
    pvt_ip_list = []
    for vios_info in res:
        ip_add_split = vios_pvt_ip.split(".")
        next_last_ip = int(ip_add_split[3])+ip_count
        ip_add = ""
        for info in ip_add_split[:-1]:
            ip_add = ip_add+info+'.'
        ip_add = ip_add+str(next_last_ip)
        pvt_ip_list.append(ip_add)
        ip_count = ip_count+1
    # For each VIOS, validate and configure private network
    x = 0
    for vios_info in res:
        vios_id = vios_info['id']
        print(MSGCATALOG.VIOS_ADD_NETWORK_START.format(vios_id))
        ip_add = pvt_ip_list[x]
        mgmt_obj.configure_network_vios(vios_id, ip_add, pvt_ip_list)
        x = x+1

    # Configure Novalink private netmask
    print(MSGCATALOG.NOVALINK_ADD_NETWORK_START)
    res = mgmt_obj.configure_network_nl()
    if res:
        print(MSGCATALOG.NOVALINK_REBOOT)
        # mgmt_obj.execute_cmd(cmd)
    print(utils.LOG_FILE_MSG)


# Main Function
if __name__ == "__main__":
    sys.exit(main())
