# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import logging
import mock
import six
import testtools

import pvmcli.cli_parser as cli_parser
import pvmcli.util.exceptions as exc

import pypowervm.exceptions as pyexc


OBJECT_TYPES = (
    ['VirtualEthernetAdapter', 'vea', 'eth'],
    ['IOSlot', 'io'],
    ['LogicalPartition', 'lpar', 'vm'],
    ['LogicalUnit', 'lu'],
    ['ManagedSystem', 'sys'],
    ['NetworkBridge', 'nbr', 'bridge', 'sea', 'sharedethernetadapter'],
    ['PhysicalVolume', 'pv'],
    ['SharedStoragePool', 'ssp'],
    ['VirtualIOServer', 'vios'],
    ['VirtualMediaRepository', 'repo', 'vmr'],
    ['VirtualNetwork', 'vnet', 'net'],
    ['VirtualOpticalMedia', 'media'],
    ['VirtualSwitch', 'vswitch', 'vsw'],
    ['VolumeGroup', 'vg'],
    ['VirtualSCSIMapping', 'scsimapping', 'scsi'])


TEST_LOG = logging.getLogger('pvmctl')
EXP_HELP = '''\
usage: pvmctl [--version] [help] <object type> [help] <op> [help] [<args>]

Power Virtualization NovaLink CLI

pvmctl is a CLI to manage Power Systems through the NovaLink REST API.

All CLI commands are driven by operations on object types (listed below).
The available operations are available in the help text for each object
type. For example, `pvmctl vm help` will list all operations available to
manage your logical partitions.

Operations also have their own help text. For example, `pvmctl vm create help`
will give the help text associated with creating a new logical partition.

Object type options (aliases):
    cluster
    healthmonitor (hms)
    ioslot (io)
    logicalpartition (lpar, vm)
    logicalunit (lu)
    logicalvolume (lv)
    managedsystem (sys)
    networkbridge (bridge, nbr, sea, sharedethernetadapter)
    persistentmemorydevice (pmemd)
    persistentmemoryvolume (pmemv)
    physicalvolume (pv)
    sharedprocessorpool (sharedprocpool, spp)
    sharedstoragepool (ssp)
    sriovadapter (sriov)
    sriovethernetlogicalport (ethlogicalport, ethlp)
    virtualethernetadapter (eth, vea)
    virtualfibrechannelmapping (vfc, vfcmapping)
    virtualioserver (vios)
    virtualmediarepository (repo, vmr)
    virtualnetwork (net, vnet)
    virtualnicdedicated (vnic)
    virtualopticalmedia (media, vom)
    virtualscsimapping (scsi, scsimapping)
    virtualserialnumber (vsn)
    virtualswitch (vsw, vswitch)
    volumegroup (vg)

Optional arguments:
    --version             show version of pvmctl
    --help | help         show this help message and exit

'''


class CLIParserTest(testtools.TestCase):
    ''' Test framework for non-ObjectType commands. '''
    def test_help(self):
        ''' Test help arguments for non-object type parser. '''
        # Test no arguments print help text
        # pvmctl
        no_arg_out = six.StringIO()
        cp = cli_parser.CLIParser([], help_out=no_arg_out)
        self.assertEqual(cp.process(), exc.OK_RETURN_CODE)
        self.assertEqual(no_arg_out.getvalue(), EXP_HELP)

        # Test help argument
        # pvmctl help
        help_out = six.StringIO()
        cp = cli_parser.CLIParser(['help'], help_out=help_out)
        self.assertEqual(cp.process(), exc.OK_RETURN_CODE)
        self.assertEqual(help_out.getvalue(), EXP_HELP)

        # Test --help argument
        # pvmctl --help
        dash_help_out = six.StringIO()
        cp = cli_parser.CLIParser(['--help'], help_out=dash_help_out)
        self.assertEqual(cp.process(), exc.OK_RETURN_CODE)
        self.assertEqual(dash_help_out.getvalue(), EXP_HELP)

        self.assertEqual(no_arg_out.getvalue(), help_out.getvalue())
        self.assertEqual(help_out.getvalue(), dash_help_out.getvalue())

    @mock.patch('pvmcli.cli_parser.CLIParser._version',
                return_value='VERSIONNUM')
    def test_version(self, version_mock):
        cli_parser.CLIParser(['--version']).process()
        version_mock.assert_called_once_with()

    def test_obj_types(self):
        for obj_type_list in OBJECT_TYPES:
            help_out_list = []
            for obj_type in obj_type_list:
                proc_out = six.StringIO()
                cp = cli_parser.CLIParser([obj_type], help_out=proc_out)
                self.assertEqual(cp.process(), exc.OK_RETURN_CODE)
                self.assertIsNotNone(proc_out)
                self.assertIsNotNone(proc_out.getvalue())
                help_out_list.append(proc_out.getvalue())

                proc_out = six.StringIO()
                cp = cli_parser.CLIParser([obj_type.upper()],
                                          help_out=proc_out)
                self.assertEqual(cp.process(), exc.OK_RETURN_CODE)
                self.assertIsNotNone(proc_out)
                self.assertIsNotNone(proc_out.getvalue())
                help_out_list.append(proc_out.getvalue())

                proc_out = six.StringIO()
                cp = cli_parser.CLIParser([obj_type.lower()],
                                          help_out=proc_out)
                self.assertEqual(cp.process(), exc.OK_RETURN_CODE)
                self.assertIsNotNone(proc_out)
                self.assertIsNotNone(proc_out.getvalue())
                help_out_list.append(proc_out.getvalue())

            self.assertTrue(all(help_out == help_out_list[0]
                                for help_out in help_out_list))

    def test_invalid_arg(self):
        ''' Test invalid argument handling in non-object type parser. '''
        # Doesn't matter what we put in, if we pass in an invalid argument
        # we will fail. argparse then exits the program with an error return
        # code of 2.
        # pvmctl notanarg
        se_exc = self.assertRaises(SystemExit, cli_parser.CLIParser,
                                   ['notanarg'])
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

        # pvmctl --help notanarg
        se_exc = self.assertRaises(SystemExit, cli_parser.CLIParser,
                                   ['--help', 'notanarg'])
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

        # pvmctl notanarg help
        se_exc = self.assertRaises(SystemExit, cli_parser.CLIParser,
                                   ['notanarg', 'help'])
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

    @mock.patch('pvmcli.command.object_type.ObjectType.execute')
    def test_short_display_fields(self, mock_exec):
        def assert_ok(cmd):
            """Valid display field spec (single, unambiguous match)."""
            mock_exec.reset_mock()
            cp = cli_parser.CLIParser(cmd.split())
            self.assertEqual(exc.OK_RETURN_CODE, cp.process())
            self.assertEqual(1, mock_exec.call_count)

        def assert_ambiguous(cmd):
            """Ambiguous display field spec (more than one possible match)."""
            mock_exec.reset_mock()
            cp = cli_parser.CLIParser(cmd.split())
            self.assertEqual(exc.ARGPARSE_ERR_RET_CODE, cp.process())
            self.assertEqual(0, mock_exec.call_count)

        def assert_invalid(cmd):
            """Invalid display field spec (no matches)."""
            mock_exec.reset_mock()
            cp = cli_parser.CLIParser(cmd.split())
            se_exc = self.assertRaises(SystemExit, cp.process)
            self.assertEqual(exc.ARGPARSE_ERR_RET_CODE, se_exc.code)
            self.assertEqual(0, mock_exec.call_count)

        # Fully-qualified
        assert_ok('LogicalPartition list -d LogicalPartition.name')
        # Case-insensitive
        assert_ok('LogicalPartition list -d LOGICALPARTITION.NAME')
        # Whole label, defaults back to parent type (LogicalPartition.name,
        # PhysicalFibreChannelPort.name)
        assert_ok('LogicalPartition list -d name')
        # Partial model to disambiguate label
        assert_ok('LogicalPartition list -d Lo.name')
        # Ambiguous partial label (PartitionMemoryConfiguration.shared_enabled,
        # PartitionProcessorConfiguration.shared_proc_cfg)
        assert_ambiguous('LogicalPartition list -d shared')
        # Unambiguous partial label (SharedProcessorConfiguration.pool_id)
        assert_ok('LogicalPartition list -d pool')
        # Bogus label
        assert_invalid('LogicalPartition list -d aouename')
        # Valid model, bogus label
        assert_invalid('LogicalPartition list -d LogicalPartition.foo')
        # Valid label, bogus model
        assert_invalid('LogicalPartition list -d Foo.name')
        # Empty model makes in invalid, not ambiguous
        assert_invalid('LogicalPartition list -d .name')
        # Empty model makes in invalid, not matching
        assert_invalid('LogicalPartition list -d .pool_id')
        # Empty label makes it invalid, not ambiguous
        assert_invalid('LogicalPartition list -d LogicalPartition.')
        # Exact label match ignores substring possibilities
        assert_ok('LogicalPartition list -d Lo.name')
        # ...whereas non-exact match is ambiguous, but defaults back to
        # parent wrapper (LogicalPartition.name,
        # ProfileIOSlot.pci_subsys_drc_name,
        # AssociatedIOSlot.pci_subsys_drc_name)
        assert_ok('LogicalPartition list -d Lo.nam')

    def _raise_exc(self):
        raise Exception("testerr")

    def _raise_conn_err(self):
        raise pyexc.ConnectionError("msg")

    def test_logger(self):
        cp = cli_parser.CLIParser(['sys', 'help'])
        TEST_LOG.info = mock.Mock()
        TEST_LOG.exception = mock.Mock()
        cp.process(auto_exec=False, log=TEST_LOG)

        cp.log.info.assert_called_once_with(
            'Running command "%(prog)s %(args)s"',
            {'prog': 'pvmctl', 'args': 'sys help'})
        cp.log.info.reset_mock()

        cp = cli_parser.CLIParser(['sys', 'list'])
        cp.obj_type.execute = mock.Mock(side_effect=self._raise_exc)
        cp.process()
        cp.log.info.assert_called_once_with(
            'Running command "%(prog)s %(args)s"',
            {'prog': 'pvmctl', 'args': 'sys list'})
        cp.log.exception.assert_called_once_with(
            'Command "pvmctl sys list" failed with testerr')
        cp.log.info.reset_mock()
        cp.log.exception.reset_mock()

        cp = cli_parser.CLIParser(['sys', 'list'])
        cp.obj_type.execute = mock.Mock(side_effect=self._raise_conn_err)
        cp.process()
        cp.log.info.assert_called_once_with(
            'Running command "%(prog)s %(args)s"',
            {'prog': 'pvmctl', 'args': 'sys list'})
        cp.log.exception.assert_called_once_with(
            "Command \"pvmctl sys list\" failed with Unable to connect to the "
            "NovaLink services. Check to make sure the pvm-core and pvm-rest "
            "services are running (systemctl status), and verify that this "
            "partition has PowerVM management authority. If you are currently "
            "updating pvm-novalink, you may see this error. Wait for the "
            "update to finish, then try again.")
