# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
import six

import pvmcli.command.network_bridge as nb_c
import pvmcli.help.help_network_bridge as help_nbr
import pvmcli.tests.util.test_util as test_util

import pypowervm.tests.test_fixtures as fx
import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.network as n
from pypowervm.wrappers import managed_system as ms_w
import pypowervm.wrappers.virtual_io_server as v


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] NetworkBridge list "
                    "{--object-id | -i} <NetworkBridge field>=value "
                    "{args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    EthernetBackingDevice.adapter_id\n"
     "    EthernetBackingDevice.dev_name\n"
     "    LoadGroup.pvid\n"
     "    LoadGroup.tagged_vlans\n"
     "    LoadGroup.trunk_adapters\n"
     "    LoadGroup.vnet_uri_list\n"
     "    NetworkBridge.arbitrary_pvids\n"
     "    NetworkBridge.control_channel_id\n"
     "    NetworkBridge.load_balance\n"
     "    NetworkBridge.load_grps\n"
     "    NetworkBridge.pvid\n"
     "    NetworkBridge.seas\n"
     "    NetworkBridge.uuid\n"
     "    NetworkBridge.vnet_uri_list\n"
     "    NetworkBridge.vswitch_id\n"
     "    SharedEthernetAdapter.addl_adpts\n"
     "    SharedEthernetAdapter.backing_device\n"
     "    SharedEthernetAdapter.configuration_state\n"
     "    SharedEthernetAdapter.control_channel\n"
     "    SharedEthernetAdapter.dev_name\n"
     "    SharedEthernetAdapter.ha_mode\n"
     "    SharedEthernetAdapter.is_primary\n"
     "    SharedEthernetAdapter.primary_adpt\n"
     "    SharedEthernetAdapter.pvid\n"
     "    SharedEthernetAdapter.vio_uri\n"
     "    TrunkAdapter.associated_vswitch_uri\n"
     "    TrunkAdapter.dev_name\n"
     "    TrunkAdapter.has_tag_support\n"
     "    TrunkAdapter.loc_code\n"
     "    TrunkAdapter.pvid\n"
     "    TrunkAdapter.tagged_vlans\n"
     "    TrunkAdapter.trunk_pri\n"
     "    TrunkAdapter.varied_on\n"
     "    TrunkAdapter.vios_id\n"
     "    TrunkAdapter.virtual_slot_number\n"
     "    TrunkAdapter.vswitch_id\n"])


class NetworkBridgeTest(twa.TestWrapper):
    ''' Test framework for NetworkBridge Object Type. '''
    mock_adapter_fx_args = dict(traits=fx.LocalPVMTraits)

    @property
    def file(self):
        return 'fake_network_bridge.txt'

    @property
    def wrapper_class_to_test(self):
        return n.NetBridge

    def test_network_adapter(self):
        nb = nb_c.NetworkBridge()
        # Test Object Type data
        field_models = [n.NetBridge, n.LoadGroup, n.TrunkAdapter,
                        n.SEA, n.EthernetBackingDevice]
        test_util.test_obj_type(self, nb, 'NetworkBridge',
                                n.NetBridge, field_models=field_models,
                                name_to_ops={'create', 'list', 'update',
                                             'delete', 'add-vlans',
                                             'remove-vlans'},
                                exp_help=help_nbr.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={nb.name_to_ops['list']: [ms_w.System],
                                    nb.name_to_ops['update']: [ms_w.System],
                                    nb.name_to_ops['delete']: [ms_w.System],
                                    nb.name_to_ops['create']:
                                    [ms_w.System],
                                    nb.name_to_ops['add-vlans']:
                                    [ms_w.System],
                                    nb.name_to_ops['remove-vlans']:
                                    [ms_w.System]},
            child_model_ops={nb.name_to_ops['list']: None,
                             nb.name_to_ops['update']: None,
                             nb.name_to_ops['delete']: None,
                             nb.name_to_ops['create']: None,
                             nb.name_to_ops['add-vlans']: None,
                             nb.name_to_ops['remove-vlans']: None},
            elem_wrp_func_ops={nb.name_to_ops['list']: None,
                               nb.name_to_ops['update']: None,
                               nb.name_to_ops['delete']: None,
                               nb.name_to_ops['create']: None,
                               nb.name_to_ops['add-vlans']: None,
                               nb.name_to_ops['remove-vlans']: None})

    def test_overrides(self):
        ''' Test the overridden Field objects for NetworkBridge ObjectType '''
        nb = nb_c.NetworkBridge()
        self.assertEqual(
            nb.name_to_ops['list'].id_fields['pvid'].field_type, int)

    @mock.patch("pypowervm.wrappers.network.VNet.wrap")
    @mock.patch("pypowervm.wrappers.virtual_io_server.VIOS.wrap")
    @mock.patch("pypowervm.wrappers.network.VSwitch.search")
    def test_list(self, vsw_search_mock, vios_wrap_mock, vnet_wrap_mock):
        dwrap2 = self.dwrap
        vsw_search_mock.side_effect = [vs_wrp for vs_wrp in [
            [n.VSwitch.bld(mock.Mock(), "t1")],
            [n.VSwitch.bld(mock.Mock(), "t2")]]]

        ret_vios = v.VIOS.bld(
            mock.Mock(), "test_vios",
            bp.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1))
        ret_vios._id(4)
        vios_wrap_mock.return_value = ret_vios

        vnet_wrap_mock.side_effect = [n.VNet.bld(mock.Mock(),
                                                 'test_vnet_%d' % i, i,
                                                 'test_vsw_uri', False)
                                      for i in range(20)]

        test_util.test_list(
            self, 'NetworkBridge',
            test_util.EXPECTED_HELP + LIST_USAGE_HELP + DISPLAY_FIELD_HELP,
            ["Network Bridges\n"
             "+------+---------+-----------+----------+------+--------+"
             "----------------+\n"
             "| PVID | VSwitch |    VIOS   | Back Dev | SEA  | Ld Grp |"
             "   Addl VLANs   |\n"
             "+------+---------+-----------+----------+------+--------+"
             "----------------+\n"
             "|  1   |    t1   | test_vios |   ent0   | ent8 |  ent4  |"
             "                |\n"
             "|      |         |           |          |      |  ent5  |"
             "  100,150,175,  |\n"
             "|      |         |           |          |      |        |"
             "  200,250,300,  |\n"
             "|      |         |           |          |      |        |"
             "  333,350,900,  |\n"
             "|      |         |           |          |      |        |"
             " 1001,2227,2228 |\n"
             "|  1   |    t1   | test_vios |   ent0   | ent8 |  ent4  |"
             "                |\n"
             "|      |         |           |          |      |  ent5  |"
             "  100,150,175,  |\n"
             "|      |         |           |          |      |        |"
             "  200,250,300,  |\n"
             "|      |         |           |          |      |        |"
             "  333,350,900,  |\n"
             "|      |         |           |          |      |        |"
             " 1001,2227,2228 |\n"
             "+------+---------+-----------+----------+------+--------+"
             "----------------+\n"],
            ["-----------------\n| NetworkBridge |\n----",
             "      virtual_slot_number=2\n      vswitch_id=2"],
            ['<HighAvailabilityMode kb="CUD" kxe="false">disabled</HighAvail',
             "<NetworkBridge:NetworkBridge xmlns:Netw"],
            ['group.pvid', 'EthernetBackingDevice.dev_name', 'bridge.pvid'],
            ["pvid=1,dev_name=ent0,pvid=1\npvid=4094,dev_name=ent0,pvid=1\n"],
            ['1,ent0,1\n'],
            ["pvid___1,dev_name___ent0,pvid___1\n"
             "pvid___4094,dev_name___ent0,pvid___1\n"],
            ["pvid=1___dev_name=ent0___pvid=1\n"
             "pvid=4094___dev_name=ent0___pvid=1\n"],
            ["pvid=1,dev_name=ent0,pvid=1___pvid=1,dev_name=ent0,pvid=1___"
             "pvid=4094,dev_name=ent0,pvid=1___"
             "pvid=4094,dev_name=ent0,pvid=1___pvid=1,dev_name=ent0,pvid=1___"
             "pvid=1,dev_name=ent0,pvid=1___pvid=4094,dev_name=ent0,pvid=1___"
             "pvid=4094,dev_name=ent0,pvid=1\n"],
            ["pvid___1___dev_name___ent0___pvid___1___"
             "pvid___1___dev_name___ent0___pvid___1___"
             "pvid___4094___dev_name___ent0___pvid___1___"
             "pvid___4094___dev_name___ent0___pvid___1___"
             "pvid___1___dev_name___ent0___pvid___1___"
             "pvid___1___dev_name___ent0___pvid___1___"
             "pvid___4094___dev_name___ent0___pvid___1___"
             "pvid___4094___dev_name___ent0___pvid___1\n"],
            {'root_type': 'ManagedSystem', 'child_type': 'NetworkBridge',
             'root_id': 'caae9209-25e5-35cd-a71a-ed55c03f294d'},
            wrps=[self.dwrap, dwrap2], p_wrps=[test_util.mock_sys(self.adpt)])

    def test_create(self):
        # TODO (clbush): Create test
        pass

    @mock.patch('pypowervm.tasks.network_bridger.ensure_vlans_on_nb')
    def test_add_vlans(self, ensure_vlans_mock):
        # pvmctl networkbridge add-vlans help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['networkbridge', 'add-vlans', 'help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] NetworkBridge add-vlans "
                      "{--object-id | -i} <NetworkBridge field>=value {args}",
                      help_out.getvalue())
        ensure_vlans_mock.assert_not_called()
        ensure_vlans_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans 1
        cp = test_util.setup_mock(self, ['networkbridge', 'add-vlans',
                                         '--vlans', '1'], self.adpt,
                                  wrps=[self.dwrap])

        cp.obj_type.execute()
        ensure_vlans_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, cp.obj_type.ec.msys.uuid,
            cp.obj_type.ec.wrappers[0].uuid, [1])

        ensure_vlans_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans 1 2 3
        cp = test_util.setup_mock(self, ['networkbridge', 'add-vlans',
                                         '--vlans', '1', '2', '3'], self.adpt,
                                  wrps=[self.dwrap])

        cp.obj_type.execute()
        ensure_vlans_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, cp.obj_type.ec.msys.uuid,
            cp.obj_type.ec.wrappers[0].uuid, [1, 2, 3])

        ensure_vlans_mock.reset_mock()

        # Invalid args
        # pvmctl networkbridge add-vlans
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'add-vlans'])
        ensure_vlans_mock.assert_not_called()
        ensure_vlans_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'add-vlans', '--vlans'])
        ensure_vlans_mock.assert_not_called()
        ensure_vlans_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans a
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'add-vlans', '--vlans', 'a'])
        ensure_vlans_mock.assert_not_called()
        ensure_vlans_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans 1 a
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'add-vlans', '--vlans', '1', 'a'])
        ensure_vlans_mock.assert_not_called()
        ensure_vlans_mock.reset_mock()

    @mock.patch('pypowervm.tasks.network_bridger.remove_vlan_from_nb')
    def test_remove_vlans(self, remove_vlan_mock):
        # pvmctl networkbridge remove-vlans help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['networkbridge', 'remove-vlans', 'help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] NetworkBridge remove-vla"
                      "ns {--object-id | -i} <NetworkBridge field>=value "
                      "{args}", help_out.getvalue())
        remove_vlan_mock.assert_not_called()
        remove_vlan_mock.reset_mock()

        # pvmctl networkbridge remove-vlans --vlans 1
        cp = test_util.setup_mock(self, ['networkbridge', 'remove-vlans',
                                         '--vlans', '1'], self.adpt,
                                  wrps=[self.dwrap])

        cp.obj_type.execute()
        remove_vlan_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, cp.obj_type.ec.msys.uuid,
            cp.obj_type.ec.wrappers[0].uuid, 1)

        remove_vlan_mock.reset_mock()

        # pvmctl networkbridge remove-vlans --vlans 1 2 3
        cp = test_util.setup_mock(self, ['networkbridge', 'remove-vlans',
                                         '--vlans', '1', '2', '3'], self.adpt,
                                  wrps=[self.dwrap])

        cp.obj_type.execute()

        # Verify in order execution by most recently called being last in list
        remove_vlan_mock.assert_called_with(
            cp.obj_type.ec.sm.adapter, cp.obj_type.ec.msys.uuid,
            cp.obj_type.ec.wrappers[0].uuid, 3)

        remove_vlan_mock.reset_mock()

        # Invalid args
        # pvmctl networkbridge remove-vlans
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'remove-vlans'])
        remove_vlan_mock.assert_not_called()
        remove_vlan_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'remove-vlans', '--vlans'])
        remove_vlan_mock.assert_not_called()
        remove_vlan_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans a
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'remove-vlans', '--vlans', 'a'])
        remove_vlan_mock.assert_not_called()
        remove_vlan_mock.reset_mock()

        # pvmctl networkbridge add-vlans --vlans 1 a
        self.assertRaises(SystemExit, test_util.setup_and_exec, self,
                          ['networkbridge', 'remove-vlans', '--vlans', '1',
                           'a'])
        remove_vlan_mock.assert_not_called()
        remove_vlan_mock.reset_mock()

    def test_update(self):
        # TODO (clbush): Update test
        pass

    def test_delete(self):
        ''' Test delete operation on NetworkBridge ObjectType. '''
        # TODO (clbush): Delete test
