# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import copy
import mock
import six

import pvmcli.command.physical_volume as pv_c
import pvmcli.help.help_physical_volume as help_pv
import pvmcli.tests.util.test_util as test_util

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.storage as s
from pypowervm.wrappers import virtual_io_server as v


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] PhysicalVolume list "
                    "{--parent-id | -p} <VirtualIOServer field>=value "
                    "{--object-id | -i} <PhysicalVolume field>=value "
                    "{args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    PhysicalVolume.avail_for_use\n"
     "    PhysicalVolume.capacity\n"
     "    PhysicalVolume.description\n"
     "    PhysicalVolume.emulate_model\n"
     "    PhysicalVolume.is_fc_backed\n"
     "    PhysicalVolume.loc_code\n"
     "    PhysicalVolume.name\n"
     "    PhysicalVolume.read_iops_limit\n"
     "    PhysicalVolume.state\n"
     "    PhysicalVolume.tag\n"
     "    PhysicalVolume.udid\n"
     "    PhysicalVolume.write_iops_limit\n"])


class PhysicalVolumeTest(twa.TestWrapper):
    ''' Test framework for PhysicalVolume Object Type. '''
    @property
    def file(self):
        return 'fake_vios_mappings.txt'

    @property
    def wrapper_class_to_test(self):
        return s.PV

    def test_physical_volume(self):
        pv = pv_c.PhysicalVolume()

        test_util.test_obj_type(self, pv, 'PhysicalVolume',
                                s.PV, field_models=[s.PV],
                                name_to_ops={'list', 'update'},
                                exp_help=help_pv.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={pv.name_to_ops['list']: [v.VIOS],
                                    pv.name_to_ops['update']: [v.VIOS]},
            child_model_ops={pv.name_to_ops['list']: None,
                             pv.name_to_ops['update']: None},
            elem_wrp_func_ops={pv.name_to_ops['list']:
                               [v.VIOS.phys_vols],
                               pv.name_to_ops['update']:
                               [v.VIOS.phys_vols]})

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'PhysicalVolume')

    def test_list(self):
        vios_wrap = v.VIOS.wrap(self.resp)
        vios_wrap_2 = copy.deepcopy(vios_wrap)
        vios_wrap_2.name = "test_vios"

        test_util.test_list(
            self, 'PhysicalVolume',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["Physical Volumes\n"
             "+-------------+--------+----------+--------+"
             "-----------------------+-----------+\n"
             "|     VIOS    |  Name  | Cap (MB) | State  |"
             "      Description      | Available |\n"
             "+-------------+--------+----------+--------+"
             "-----------------------+-----------+\n"
             "|  test_vios  | hdisk0 |  102400  | active |"
             " MPIO IBM 2076 FC Disk |   False   |\n"
             "|  test_vios  | hdisk1 |    0     | active |"
             " MPIO IBM 2076 FC Disk |   False   |\n"
             "|  test_vios  | hdisk2 |  20480   | active |"
             " MPIO IBM 2076 FC Disk |    True   |\n"
             "| vios_1_Drew | hdisk0 |  102400  | active |"
             " MPIO IBM 2076 FC Disk |   False   |\n"
             "| vios_1_Drew | hdisk1 |    0     | active |"
             " MPIO IBM 2076 FC Disk |   False   |\n"
             "| vios_1_Drew | hdisk2 |  20480   | active |"
             " MPIO IBM 2076 FC Disk |    True   |\n"
             "+-------------+--------+----------+--------+"
             "-----------------------+-----------+\n"],
            ["------------------\n| PhysicalVolume |\n------------------\n"
             "  avail_for_use=True\n"
             "  capacity=20480\n"
             "  description=MPIO IBM 2076 FC Disk\n"
             "  emulate_model=True\n"
             "  is_fc_backed=True\n"
             "  loc_code=U78AB.001.WZSH5ZY-P1-C5-T1-W500507680215B741-L10000"
             "00000000\n"
             "  name=hdisk2\n"
             "  state=active\n"
             "  udid=01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODE4OTk0QTAwMDAwMDAwMDAw"
             "MDhEMg==\n"],
            ['<PhysicalVolume xmlns="http://www.ibm.com/xmlns/'],
            ['PhysicalVolume.name', 'PhysicalVolume.is_fc_backed'],
            ["name=hdisk2,is_fc_backed=True\n"],
            ["hdisk2,True\n"],
            ["name___hdisk2,is_fc_backed___True\n"],
            ["name=hdisk2___is_fc_backed=True\n"],
            ["name=hdisk2,is_fc_backed=True\n"],
            ["name___hdisk2___is_fc_backed___True\n"],
            'VirtualIOServer',
            wrps=[vios_wrap.phys_vols[0]],
            p_wrps=[vios_wrap, vios_wrap_2])

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.update')
    def test_update(self, vios_upd_mock, vios_srch_mock):
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['pv', 'update', '--help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] PhysicalVolume update "
                      "{--parent-id | -p} <VirtualIOServer field>=value "
                      "{--object-id | -i} <PhysicalVolume field>=value {args}",
                      help_out.getvalue())

        vios_wrap = v.VIOS.wrap(self.resp)

        cp = test_util.setup_mock(self, args=['pv', 'update', '-p',
                                              'name=test_vios', '-i',
                                              'name=hdisk2', '-s',
                                              'read_iops_limit=22',
                                              'write_iops_limit=33'],
                                  adpt_fx=self.adpt,
                                  wrps=[vios_wrap.phys_vols[0]],
                                  p_wrps=[vios_wrap])
        self.assertEqual(vios_wrap.phys_vols[0].read_iops_limit, None)
        self.assertEqual(vios_wrap.phys_vols[0].write_iops_limit, None)
        cp.obj_type.execute()
        self.assertEqual(vios_wrap.phys_vols[0].read_iops_limit, 22)
        self.assertEqual(vios_wrap.phys_vols[0].write_iops_limit, 33)
        vios_upd_mock.assert_called_once_with()
        vios_srch_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, name='test_vios', parent_type=None,
            parent_uuid=None, xag=[])
        vios_upd_mock.reset_mock()
