# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
import six

import pvmcli.command.sriov_logical_port as sriov_c
import pvmcli.help.help_sriov_ethlp as help_ethlp
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as e

from pypowervm.tests.test_utils import pvmhttp
import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.iocard as iocard
import pypowervm.wrappers.logical_partition as lpar_w
import pypowervm.wrappers.virtual_io_server as vios_w


LIST_USAGE_HELP = (
    ["usage: pvmctl [help | --help] SRIOVEthernetLogicalPort list "
     "{--parent-id | -p} <LogicalPartition field>=value {--object-id | -i} "
     "<SRIOVEthernetLogicalPort field>=value {args}"])

DELETE_USAGE_HELP = (
    ["usage: pvmctl [help | --help] SRIOVEthernetLogicalPort delete "
     "{--parent-id | -p} <LogicalPartition field>=value {--object-id | -i} "
     "<SRIOVEthernetLogicalPort field>=value {args}"])

DISPLAY_FIELD_HELP = (
    ["    SRIOVEthernetLogicalPort.cfg_capacity\n"
     "    SRIOVEthernetLogicalPort.cfg_max_capacity\n"
     "    SRIOVEthernetLogicalPort.cur_mac\n"
     "    SRIOVEthernetLogicalPort.dev_name\n"
     "    SRIOVEthernetLogicalPort.is_promisc\n"
     "    SRIOVEthernetLogicalPort.loc_code\n"
     "    SRIOVEthernetLogicalPort.lport_id\n"
     "    SRIOVEthernetLogicalPort.mac\n"
     "    SRIOVEthernetLogicalPort.pport_id\n"
     "    SRIOVEthernetLogicalPort.pvid\n"
     "    SRIOVEthernetLogicalPort.sriov_adap_id\n"
     "    SRIOVEthernetLogicalPort.uuid\n"])


class SRIOVLogicalPortTest(twa.TestWrapper):
    ''' Test framework for SRIOVLogicalPort Object Type. '''
    @property
    def file(self):
        return 'sriov_lp_feed.txt'

    @property
    def wrapper_class_to_test(self):
        return iocard.SRIOVEthLPort

    def test_sriov_adapter(self):
        swrp = sriov_c.SRIOVLogicalPort()

        test_util.test_obj_type(self, swrp, 'SRIOVEthernetLogicalPort',
                                iocard.SRIOVEthLPort,
                                field_models=[iocard.SRIOVEthLPort],
                                name_to_ops={'list', 'create', 'delete'},
                                exp_help=help_ethlp.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={swrp.name_to_ops['list']: [vios_w.VIOS,
                                                               lpar_w.LPAR]},
            child_model_ops={swrp.name_to_ops['list']: None},
            elem_wrp_func_ops={swrp.name_to_ops['list']: None})

        test_util.test_ops(
            self, parent_model_ops={swrp.name_to_ops['create']: [vios_w.VIOS,
                                                                 lpar_w.LPAR]},
            child_model_ops={swrp.name_to_ops['create']: None},
            elem_wrp_func_ops={swrp.name_to_ops['create']: None})
        test_util.test_ops(
            self, parent_model_ops={swrp.name_to_ops['delete']: [vios_w.VIOS,
                                                                 lpar_w.LPAR]},
            child_model_ops={swrp.name_to_ops['delete']: None},
            elem_wrp_func_ops={swrp.name_to_ops['delete']: None})

    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.get')
    @mock.patch('pvmcli.util.util.vios_by_uri')
    @mock.patch('pypowervm.wrappers.iocard.VNIC.get')
    def test_list(self, vnic_mock, vios_by_uri_mock, lpar_mock):
        log_ports = iocard.SRIOVEthLPort.wrap(self.resp)
        samp_vnic = test_util.mock_vnic(self.adpt)
        b_devs = samp_vnic.back_devs
        b_devs.append(b_devs[1])
        samp_vnic.back_devs = b_devs
        samp_vnic.back_devs[0].set_href(iocard._VNICBD_LPORT,
                                        log_ports[1].href)
        samp_vnic.back_devs[1].set_href(iocard._VNICBD_LPORT,
                                        log_ports[2].href)
        samp_vnic.back_devs[2].set_href(iocard._VNICBD_LPORT,
                                        "misconfigured_href")
        vnic_mock.return_value = [samp_vnic]
        vios_by_uri_mock.return_value = test_util.mock_vios(self.adpt)
        lpar_mock.return_value = [test_util.mock_lpar(self.adpt)]

        test_util.test_list(
            self, 'SRIOVEthernetLogicalPort',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["SR-IOV Ethernet Logical Ports (VNIC)\n"
             "+-------------------------+-------------+"
             "--------------------------------+--------+------------+"
             "--------+\n"
             "|           LPAR          |     VIOS    |"
             "           Loc. Code            | Active | Fail. Pri. |"
             "  Cap.  |\n"
             "+-------------------------+-------------+"
             "--------------------------------+--------+------------+"
             "--------+\n"
             "| z3-9-5-126-127-00000001 | vios_1_Drew |"
             " U78CB.001.WZS0JYF-P1-C10-T1-S1 |        |    None    |"
             " 2.00%  |\n"
             "|                         | vios_1_Drew |"
             " U78CB.001.WZS0JYF-P1-C10-T1-S2 |   *    |    None    |"
             " 18.00% |\n"
             "+-------------------------+-------------+"
             "--------------------------------+--------+------------+"
             "--------+\n"
             "SR-IOV Ethernet Logical Ports (Non-VNIC)\n"
             "+-------------------------------+------+"
             "-------------------------+-------+--------------+\n"
             "|           Loc. Code           | PVID |"
             "        Partition        |  Cap. |     MAC      |\n"
             "+-------------------------------+------+"
             "-------------------------+-------+--------------+\n"
             "| U78CB.001.WZS0485-P1-C5-T3-S2 |  0   |"
             " z3-9-5-126-127-00000001 | 2.00% | 000000000000 |\n"
             "+-------------------------------+------+"
             "-------------------------+-------+--------------+\n"
             "Misconfigured Ethernet Logical Ports\n"
             "+--------------------------------+\n"
             "|           Loc. Code            |\n"
             "+--------------------------------+\n"
             "| U78CB.001.WZS0JYF-P1-C10-T1-S3 |\n"
             "+--------------------------------+\n"],
            ["----------------------------\n"
             "| SRIOVEthernetLogicalPort |\n"
             "----------------------------\n"
             "  allowed_macs=NONE\n"
             "  allowed_vlans=ALL\n"
             "  cfg_capacity=2.00%\n"
             "  cfg_max_capacity=100.00%\n"
             "  cur_mac=000000000000\n"
             "  dev_name=PHB 4098\n"
             "  is_promisc=False\n"
             "  loc_code=U78CB.001.WZS0485-P1-C5-T3-S2\n"
             "  lport_id=27004002\n"
             "  mac=8ACA227C6E00\n"
             "  pport_id=2\n"
             "  pvid=0\n"
             "  sriov_adap_id=1\n"
             "  uuid=080c7f1f-29e9-333a-8034-a5c8a358786e\n"
             "  vnic_port_usage=NOT_VNIC\n"
             "----------------------------\n"
             "| SRIOVEthernetLogicalPort |\n"
             "----------------------------\n"
             "  allowed_macs=NONE\n"
             "  allowed_vlans=NONE\n"
             "  cfg_capacity=2.00%\n"
             "  cfg_max_capacity=80.00%\n"
             "  cur_mac=A65BDDBBAD66\n"
             "  dev_name=unavailable\n"
             "  is_promisc=False\n"
             "  loc_code=U78CB.001.WZS0JYF-P1-C10-T1-S1\n"
             "  lport_id=27004001\n"
             "  mac=A65BDDBBAD66\n"
             "  pport_id=0\n"
             "  pvid=1095\n"
             "  sriov_adap_id=1\n"
             "  uuid=6d9cfdda-ae8b-3c58-bc0c-379a0c5ab154\n"
             "  vnic_port_usage=DEDICATED_VNIC\n"
             "----------------------------\n"
             "| SRIOVEthernetLogicalPort |\n"
             "----------------------------\n"
             "  allowed_macs=ALL\n"
             "  allowed_vlans=ALL\n"
             "  cfg_capacity=18.00%\n"
             "  cfg_max_capacity=10.00%\n"
             "  cur_mac=A65BDDBBAD67\n"
             "  dev_name=unavailable\n"
             "  is_promisc=True\n"
             "  loc_code=U78CB.001.WZS0JYF-P1-C10-T1-S2\n"
             "  lport_id=27004002\n"
             "  mac=A65BDDBBAD67\n"
             "  pport_id=1\n"
             "  pvid=42\n"
             "  sriov_adap_id=2\n"
             "  uuid=6d9cfdda-ae8b-3c58-bc0c-379a0c5ab155\n"
             "  vnic_port_usage=DEDICATED_VNIC\n"
             "----------------------------\n"
             "| SRIOVEthernetLogicalPort |\n"
             "----------------------------\n"
             "  allowed_macs=NONE\n"
             "  allowed_vlans=ALL\n"
             "  cfg_capacity=2.00%\n"
             "  cfg_max_capacity=100.00%\n"
             "  cur_mac=A65BDDBBAD68\n"
             "  dev_name=unavailable\n"
             "  is_promisc=False\n"
             "  loc_code=U78CB.001.WZS0JYF-P1-C10-T1-S3\n"
             "  lport_id=27004003\n"
             "  mac=A65BDDBBAD68\n"
             "  pport_id=1\n"
             "  pvid=42\n"
             "  sriov_adap_id=2\n"
             "  uuid=6d9cfdda-ae8b-3c58-bc0c-379a0c5ab156\n"
             "  vnic_port_usage=DEDICATED_VNIC\n"],
            ['<SRIOVEthernetLogicalPort:SRIOVEthernetLogicalPort xmlns:SRIOV'],
            ['cfg_capacity', 'pvid'],
            ["cfg_capacity=2.00%,pvid=0\ncfg_capacity=2.00%,pvid=1095\n"],
            ["2.00%,0\n2.00%,1095\n"],
            ["cfg_capacity___2.00%,pvid___0\n"
             "cfg_capacity___2.00%,pvid___1095\n"
             "cfg_capacity___18.00%,pvid___42\n"
             "cfg_capacity___2.00%,pvid___42\n"],
            ["cfg_capacity=2.00%___pvid=0\ncfg_capacity=2.00%___pvid=1095\n"
             "cfg_capacity=18.00%___pvid=42\ncfg_capacity=2.00%___pvid=42\n"],
            ["cfg_capacity=2.00%,pvid=0___"
             "cfg_capacity=2.00%,pvid=1095___"
             "cfg_capacity=18.00%,pvid=42___"
             "cfg_capacity=2.00%,pvid=42\n"],
            ["cfg_capacity___2.00%___pvid___0___"
             "cfg_capacity___2.00%___pvid___1095___"
             "cfg_capacity___18.00%___pvid___42___"
             "cfg_capacity___2.00%___pvid___42\n"],
            {'child_type': 'SRIOVEthernetLogicalPort',
             'root_id': '089FFB20-5D19-4A8C-BB80-13650627D985',
             'root_type': 'LogicalPartition'},
            wrps=log_ports,
            p_wrps=[test_util.mock_lpar(self.adpt)])

        # We run multiple tests with the above, will have multiple calls
        # for each of these.
        self.assertTrue(lpar_mock.called)
        self.assertTrue(vios_by_uri_mock.called)
        lpar_mock.reset_mock()
        vios_by_uri_mock.reset_mock()

        # Test --vnic arg and * char for Active
        samp_vnic.back_devs[0].set_parm_value(iocard._VNICBD_ACTIVE,
                                              True)
        samp_vnic.back_devs[0].failover_pri = 10
        s_lp = test_util.mock_lpar(self.adpt)
        s_lp2 = lpar_w.LPAR.bld(
            self.adpt, 'test', s_lp.mem_config, s_lp.proc_config)
        pwtc = {s_lp: [log_ports[0], log_ports[1]],
                s_lp2: [log_ports[2], log_ports[3]]}

        lpar_mock.return_value = [s_lp, s_lp2]
        # Set max width manually so terminal-less test machine prints
        # the entire output.
        cp = test_util.setup_mock(
            self, args=['ethlp', 'list', '--vnic', '--max-width', '200'],
            adpt_fx=self.adpt, wrps=log_ports, p_wrps=[])

        # Bypass any wrapper queries, building up the parent wrapper
        # structure to test different LPAR names + sorting...
        samp_vnic.back_devs[2].set_href(iocard._VNICBD_LPORT,
                                        log_ports[3].href)
        cp.obj_type.cli_out = six.StringIO()
        cp.obj_type.ec.parent_wrp_to_children = pwtc
        cp.obj_type.ec.exec_op._query_wrps = False
        cp.obj_type.execute()

        self.assertEqual(
            "SR-IOV Ethernet Logical Ports (VNIC)\n"
            "+-------------------------+-------------+"
            "--------------------------------+--------+"
            "------------+--------+\n"
            "|           LPAR          |     VIOS    |"
            "           Loc. Code            | Active |"
            " Fail. Pri. |  Cap.  |\n"
            "+-------------------------+-------------+"
            "--------------------------------+--------+"
            "------------+--------+\n"
            "|           test          | vios_1_Drew |"
            " U78CB.001.WZS0JYF-P1-C10-T1-S2 |   *    |"
            "    None    | 18.00% |\n"
            "|                         | vios_1_Drew |"
            " U78CB.001.WZS0JYF-P1-C10-T1-S3 |   *    |"
            "    None    | 2.00%  |\n"
            "|                         | vios_1_Drew |"
            " U78CB.001.WZS0JYF-P1-C10-T1-S1 |   *    |"
            "     10     | 2.00%  |\n"
            "| z3-9-5-126-127-00000001 | vios_1_Drew |"
            " U78CB.001.WZS0JYF-P1-C10-T1-S2 |   *    |"
            "    None    | 18.00% |\n"
            "|                         | vios_1_Drew |"
            " U78CB.001.WZS0JYF-P1-C10-T1-S3 |   *    |"
            "    None    | 2.00%  |\n"
            "|                         | vios_1_Drew |"
            " U78CB.001.WZS0JYF-P1-C10-T1-S1 |   *    |"
            "     10     | 2.00%  |\n"
            "+-------------------------+-------------+"
            "--------------------------------+--------+"
            "------------+--------+\n\n", cp.obj_type.cli_out.getvalue())

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.delete')
    def test_delete(self, ew_mock_delete):
        ''' Test delete operation. '''
        test_util.test_delete(self, ew_mock_delete, 'SRIOVEthernetLogicalPort',
                              DELETE_USAGE_HELP, wrps=[self.dwrap],
                              p_wrps=[test_util.mock_lpar(self.adpt)])

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search',
                new=mock.Mock())
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.search')
    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.create')
    def test_create(self, create_mock, lpar_mock):
        # Minimum args
        phys_loc_str = 'U78C7.001.RCH0004-P1-C8-T1'
        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')

        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))

        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        # All args
        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name', '--promisc',
                                              '--pvid', '2',
                                              '--capacity', '40',
                                              '--max-capacity', '83'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')

        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))

        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        # --parent-id required
        args = ['ethlp', 'create', '--loc', phys_loc_str, '--promisc',
                '--pvid', '2', '--capacity', '40.40']
        self.assertRaises(SystemExit, test_util.setup_mock, self, args,
                          self.adpt)
        create_mock.assert_not_called()
        create_mock.reset_mock()

        # --loc required
        args = ['ethlp', 'create', '-p', 'name=a_lpar', '--promisc',
                '--pvid', '2', '--capacity', '0.01']
        self.assertRaises(SystemExit, test_util.setup_mock, self, args,
                          self.adpt)
        create_mock.assert_not_called()
        create_mock.reset_mock()

        # --capacity out of bounds error
        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name', '--promisc',
                                              '--pvid', '2',
                                              '--capacity', '400'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))
        self.assertRaises(e.CapacityOutOfBoundsError, cp.obj_type.execute)
        create_mock.assert_not_called()
        create_mock.reset_mock()

        # --max-capacity out of bounds error
        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name', '--promisc',
                                              '--pvid', '2',
                                              '--max-capacity', '400'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))
        self.assertRaises(e.CapacityOutOfBoundsError, cp.obj_type.execute)
        create_mock.assert_not_called()
        create_mock.reset_mock()

    @mock.patch('pypowervm.wrappers.iocard.SRIOVEthLPort.bld')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search',
                new=mock.Mock())
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.search')
    def test_create_bld(self, lpar_mock, bld_mock):
        # Minimum args
        phys_loc_str = 'U78C7.001.RCH0004-P1-C8-T1'
        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')

        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))

        cp.obj_type.execute()
        bld_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, cfg_capacity=None, is_promisc=False,
            pvid=None, pport_id=0, max_capacity=None, sriov_adap_id=1)
        bld_mock.reset_mock()

        # All args
        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name', '--promisc',
                                              '--pvid', '2',
                                              '--capacity', '40',
                                              '--max-capacity', '83'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')

        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))

        cp.obj_type.execute()
        bld_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, cfg_capacity=0.40, is_promisc=True,
            pvid=2, pport_id=0, max_capacity=0.83, sriov_adap_id=1)
        bld_mock.reset_mock()

        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name', '--promisc',
                                              '--pvid', '2',
                                              '--capacity', '40%',
                                              '--max-capacity', '83%'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')

        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))

        cp.obj_type.execute()
        bld_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, cfg_capacity=0.40, is_promisc=True,
            pvid=2, pport_id=0, max_capacity=0.83, sriov_adap_id=1)
        bld_mock.reset_mock()

        cp = test_util.setup_mock(self, args=['ethlp', 'create',
                                              '--loc', phys_loc_str, '-p',
                                              'name=lpar_name', '--promisc',
                                              '--pvid', '2',
                                              '--capacity', '40.10%',
                                              '--max-capacity', '83.10%'],
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt')

        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, lpar_mock))

        cp.obj_type.execute()
        bld_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, cfg_capacity=0.4010, is_promisc=True,
            pvid=2, pport_id=0, max_capacity=0.8310, sriov_adap_id=1)

    def _setup_mock_lpar(self):
        resp = pvmhttp.PVMResp(
            pvmfile=pvmhttp.PVMFile('lpar.txt'),
            adapter=self.adpt).get_response()
        return lpar_w.LPAR.wrap(resp)

    def _mock_p_wrps(self, cp, lpar_wrp):
        cp.obj_type.ec.elem_p_wrappers = [lpar_wrp]
