#
# Licensed Materials - Property of IBM
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

import mock
import six

import pvmcli.command.operation as op
import pvmcli.command.virtual_fc_mapping as fc_c
import pvmcli.help.help_virtual_fc_mapping as help_vfc
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as e

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.logical_partition as lpar_w
import pypowervm.wrappers.storage as s
import pypowervm.wrappers.virtual_io_server as vios_w


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualFibreChannelMapping "
                    "list {--parent-id | -p} <VirtualIOServer field>=value"
                    " {--object-id | -i} <VirtualFibreChannelMapping field>="
                    "value {args}"])


DISPLAY_FIELD_HELP = (
    ["    ClientAdapter.is_varied_on\n"
     "    ClientAdapter.loc_code\n"
     "    ClientAdapter.lpar_id\n"
     "    ClientAdapter.lpar_slot_num\n"
     "    ClientAdapter.side\n"
     "    ClientAdapter.vios_id\n"
     "    ClientAdapter.vios_slot_num\n"
     "    ClientAdapter.wwpns\n"
     "    PhysicalFibreChannelPort.loc_code\n"
     "    PhysicalFibreChannelPort.name\n"
     "    PhysicalFibreChannelPort.npiv_available_ports\n"
     "    PhysicalFibreChannelPort.npiv_total_ports\n"
     "    PhysicalFibreChannelPort.udid\n"
     "    PhysicalFibreChannelPort.wwpn\n"
     "    ServerAdapter.is_varied_on\n"
     "    ServerAdapter.loc_code\n"
     "    ServerAdapter.lpar_id\n"
     "    ServerAdapter.lpar_slot_num\n"
     "    ServerAdapter.map_port\n"
     "    ServerAdapter.name\n"
     "    ServerAdapter.side\n"
     "    ServerAdapter.udid\n"
     "    ServerAdapter.vios_id\n"
     "    ServerAdapter.vios_slot_num\n"
     "    VirtualFibreChannelMapping.backing_port\n"
     "    VirtualFibreChannelMapping.client_adapter\n"
     "    VirtualFibreChannelMapping.client_lpar_href\n"
     "    VirtualFibreChannelMapping.server_adapter\n"])


class FCMappingTest(twa.TestWrapper):
    ''' Test framework for FCMapping Object Type. '''
    @property
    def file(self):
        return 'fake_vios_mappings.txt'

    @property
    def wrapper_class_to_test(self):
        return vios_w.VFCMapping

    def test_vfc_mapping(self):
        fc = fc_c.VFCStorageMapping()

        test_util.test_obj_type(self, fc, 'VirtualFibreChannelMapping',
                                vios_w.VFCMapping,
                                field_models=[vios_w.VFCMapping,
                                              s.VFCClientAdapterElement,
                                              s.VFCServerAdapterElement,
                                              bp.PhysFCPort],
                                name_to_ops={op.LIST_NAME, op.CREATE_NAME,
                                             op.DELETE_NAME, op.REMAP_NAME},
                                exp_help=help_vfc.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={fc.name_to_ops[op.LIST_NAME]:
                                    [vios_w.VIOS],
                                    fc.name_to_ops[op.CREATE_NAME]:
                                    [vios_w.VIOS],
                                    fc.name_to_ops[op.DELETE_NAME]:
                                    [vios_w.VIOS],
                                    fc.name_to_ops[op.REMAP_NAME]:
                                    [vios_w.VIOS]},
            child_model_ops={fc.name_to_ops[op.LIST_NAME]: None,
                             fc.name_to_ops[op.CREATE_NAME]: None,
                             fc.name_to_ops[op.DELETE_NAME]: None,
                             fc.name_to_ops[op.REMAP_NAME]: None},
            elem_wrp_func_ops={fc.name_to_ops[op.LIST_NAME]:
                               [vios_w.VIOS.vfc_mappings],
                               fc.name_to_ops[op.CREATE_NAME]:
                               [vios_w.VIOS.vfc_mappings],
                               fc.name_to_ops[op.DELETE_NAME]:
                               [vios_w.VIOS.vfc_mappings],
                               fc.name_to_ops[op.REMAP_NAME]:
                               [vios_w.VIOS.vfc_mappings]})

    @mock.patch("pypowervm.wrappers.virtual_io_server.VIOS.get")
    @mock.patch("pypowervm.wrappers.logical_partition.LPAR.get")
    def test_list(self, lpar_mock, vios_mock):
        vios_wrap = vios_w.VIOS.wrap(self.resp)
        vios_wrap.set_parm_value('PartitionID', 1)
        vios_mock.return_value = [vios_wrap]
        test_lp_wrp = lpar_w.LPAR.bld(
            mock.Mock(), "test_lpar",
            bp.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1))
        test_lp_wrp.set_parm_value('PartitionID', 24)
        lpar_mock.return_value = [test_lp_wrp]
        test_util.test_list(
            self, 'VirtualFibreChannelMapping',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["Virtual FC Mappings\n"
             "+-----------+-----------+-------------+-----------+-----------+"
             "------------+\n"
             "|    LPAR   | LPAR Slot |     VIOS    | VIOS Slot | Port Name |"
             " Adpt. Name |\n"
             "+-----------+-----------+-------------+-----------+-----------+"
             "------------+\n"
             "| test_lpar |     3     | vios_1_Drew |     43    |    fcs2   |"
             " vfchost26  |\n"
             "+-----------+-----------+-------------+-----------+-----------+"
             "------------+\n"],
            ["------------------------------\n"
             "| VirtualFibreChannelMapping |\n"
             "------------------------------\n"
             "  ----------------------------\n"
             "  | PhysicalFibreChannelPort |\n"
             "  ----------------------------\n"
             "    loc_code=U78AB.001.WZSH5ZY-P1-C5-T1\n"
             "    name=fcs2\n"
             "    npiv_available_ports=59\n"
             "    npiv_total_ports=64\n"
             "    udid=1aU78AB.001.WZSH5ZY-P1-C5-T1\n"
             "    wwpn=10000090FA1B6898\n"
             "  -----------------\n"
             "  | ServerAdapter |\n"
             "  -----------------\n"
             "    is_varied_on=true\n"
             "    loc_code=U8246.L2C.0604C7A-V1-C43\n"
             "    lpar_id=24\n"
             "    lpar_slot_num=3\n"
             "    map_port=fcs2\n"
             "    name=vfchost26\n"
             "    side=Server\n"
             "    udid=1dU8246.L2C.0604C7A-V1-C43\n"
             "    vios_id=1\n"
             "    vios_slot_num=43\n"],
            ['<VirtualFibreChannelMapping xmlns="http://www.ibm.com/xmlns/'],
            ['ServerAdapter.lpar_slot_num', 'ServerAdapter.loc_code'],
            ["lpar_slot_num=3,loc_code=U8246.L2C.0604C7A-V1-C43\n"],
            ["3,U8246.L2C.0604C7A-V1-C43"],
            ["lpar_slot_num___3,loc_code___U8246.L2C.0604C7A-V1-C43\n"],
            ["lpar_slot_num=3___loc_code=U8246.L2C.0604C7A-V1-C43\n"],
            ["lpar_slot_num=3,loc_code=U8246.L2C.0604C7A-V1-C43\n"],
            ["lpar_slot_num___3___loc_code___U8246.L2C.0604C7A-V1-C43\n"],
            'VirtualIOServer',
            wrps=[vios_wrap.vfc_mappings[0]])

    def _mock_p_wrps(self, cp, vios_wrap):
        cp.obj_type.ec.elem_p_wrappers = [vios_wrap]

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.update')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.search')
    @mock.patch('pypowervm.tasks.vfc_mapper.add_map')
    def test_create(self, add_vfc_mock, lpar_search_mock, vios_upd_mock):
        # pvmctl vfc create help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['vfc', 'create', 'help'], help_out=help_out)
        self.assertEqual(str(help_out.getvalue()),
                         str(help_vfc.help_create) + "\n")
        add_vfc_mock.assert_not_called()
        add_vfc_mock.reset_mock()

        vios_wrap = vios_w.VIOS.wrap(self.resp)

        # Valid command
        # pvmctl vfc create --lpar lpar_name --pfcport \
        # fcs2
        args = ['vfc', 'create', '--lpar', 'name=lpar_name', '--pfcport',
                'fcs2']
        cp = test_util.setup_mock(self, args, self.adpt,
                                  wrps=[vios_wrap.pfc_ports[0]])

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))
        cp.obj_type.execute()

        add_vfc_mock.assert_called_once_with(
            vios_wrap, None, lpar_search_mock().__getitem__().uuid,
            (vios_wrap.pfc_ports[0].wwpn, '-1 -1'))
        add_vfc_mock.reset_mock()

        # Invalid command
        # pvmctl vfc create --lpar lpar_name --pfcport \
        # fcs0
        args = ['vfc', 'create', '--lpar', 'name=lpar_name', '--pfcport',
                'fcs0']
        cp = test_util.setup_mock(self, args, self.adpt,
                                  wrps=[vios_wrap.pfc_ports[0]])
        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        self.assertRaises(e.NoPFCPortFoundError, cp.obj_type.execute)
        vios_upd_mock.assert_called_once_with()
        vios_upd_mock.reset_mock()
        add_vfc_mock.assert_not_called()
        add_vfc_mock.reset_mock()

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.update')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.search')
    def test_delete(self, lpar_search_mock, vios_upd_mock):
        # pvmctl vfc delete help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['vfc', 'delete', 'help'], help_out=help_out)
        self.assertEqual(str(help_out.getvalue()),
                         str(help_vfc.help_delete) + "\n")

        vios_wrap = vios_w.VIOS.wrap(self.resp)

        # Valid command
        # pvmctl vfc delete --adpt name=vfchost26
        args = ['vfc', 'delete', '--adpt', 'name=vfchost26']
        cp = test_util.setup_mock(self, args, self.adpt)

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[0]
        self.assertEqual(vfc_map.server_adapter.name, 'vfchost26')
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 31)

        cp.obj_type.execute()

        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 30)
        vios_upd_mock.assert_called_once_with()
        vios_upd_mock.reset_mock()

        # pvmctl vfc delete --vios-slot 41
        args = ['vfc', 'delete', '--vios-slot', '41']
        cp = test_util.setup_mock(self, args, self.adpt)

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[0]
        self.assertEqual(vfc_map.server_adapter.vios_slot_num, 41)
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 30)

        cp.obj_type.execute()

        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 29)
        vios_upd_mock.assert_called_once_with()
        vios_upd_mock.reset_mock()

        # Invalid command - there is more than 1 mapping to fcs2
        # pvmctl vfc delete --port name=fcs2
        args = ['vfc', 'delete', '--port', 'name=fcs2']
        cp = test_util.setup_mock(self, args, self.adpt)

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        self.assertRaises(e.SingleMappingEntryDeleteError, cp.obj_type.execute)

        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 29)
        vios_upd_mock.assert_not_called()

        # Invalid command - there are no mappings to fcs3
        # pvmctl vfc delete --port name=fcs3
        args = ['vfc', 'delete', '--port', 'name=fcs3']
        cp = test_util.setup_mock(self, args, self.adpt)

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        self.assertRaises(e.SingleMappingEntryDeleteError, cp.obj_type.execute)

        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 29)
        vios_upd_mock.assert_not_called()

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.update')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.search')
    def test_remap(self, lpar_search_mock, vios_upd_mock):
        # pvmctl vfc remap help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['vfc', 'remap', 'help'], help_out=help_out)
        self.assertEqual(str(help_out.getvalue()),
                         str(help_vfc.help_remap) + "\n")

        vios_wrap = vios_w.VIOS.wrap(self.resp)

        # Valid command
        # pvmctl vfc remap --adpt name=vfchost6 --new-port name=fcs3
        args = ['vfc', 'remap', '--adpt', 'name=vfchost6', '--new-port',
                'name=fcs3']
        cp = test_util.setup_mock(self, args, self.adpt)

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[2]
        self.assertEqual(vfc_map.server_adapter.name, 'vfchost6')
        self.assertEqual(vfc_map.backing_port.name, 'fcs2')
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 31)

        cp.obj_type.execute()
        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[2]
        self.assertEqual(vfc_map.backing_port.name, 'fcs3')
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 31)
        vios_upd_mock.assert_called_once_with()
        vios_upd_mock.reset_mock()

        # pvmctl vfc remap --adpt name=vfchost6 --new-port \
        # wwpn=10000090FA1B6898
        # Remap back to fcs2 using wwpn identifier
        args = ['vfc', 'remap', '--adpt', 'name=vfchost6', '--new-port',
                'wwpn=10000090FA1B6898']
        cp = test_util.setup_mock(self, args, self.adpt)

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[2]
        self.assertEqual(vfc_map.server_adapter.name, 'vfchost6')
        self.assertEqual(vfc_map.backing_port.name, 'fcs3')
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 31)

        cp.obj_type.execute()
        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[2]
        self.assertEqual(vfc_map.server_adapter.name, 'vfchost6')
        self.assertEqual(vfc_map.backing_port.name, 'fcs2')
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 31)
        vios_upd_mock.assert_called_once_with()
        vios_upd_mock.reset_mock()

        # Invalid command, new port is nonexistent
        # pvmctl vfc remap --adpt name=vfchost6 --new-port \
        # name=badport
        args = ['vfc', 'remap', '--adpt', 'name=vfchost6', '--new-port',
                'name=badport']
        cp = test_util.setup_mock(self, args, self.adpt)

        self._mock_p_wrps(cp, vios_wrap)
        cp.obj_type.ec._add_parent_wrappers = mock.Mock(
            side_effect=self._mock_p_wrps(cp, vios_wrap))

        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[2]
        self.assertEqual(vfc_map.server_adapter.name, 'vfchost6')
        self.assertEqual(vfc_map.backing_port.name, 'fcs2')
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 31)

        self.assertRaises(e.NoPFCPortFoundError, cp.obj_type.execute)
        vfc_map = cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings[2]
        self.assertEqual(vfc_map.server_adapter.name, 'vfchost6')
        self.assertEqual(vfc_map.backing_port.name, 'fcs2')
        self.assertEqual(
            len(cp.obj_type.ec.elem_p_wrappers[0].vfc_mappings), 31)
        vios_upd_mock.assert_not_called()
        vios_upd_mock.reset_mock()
