# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock

import pvmcli.cli_parser as cli_parser
import pvmcli.command.operation as op
import pvmcli.command.virtual_switch as vs_c
import pvmcli.help.help_virtual_switch as help_vswitch
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.managed_system as ms_w
import pypowervm.wrappers.network as n


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualSwitch list {--"
                    "object-id | -i} <VirtualSwitch field>=value {args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    VirtualSwitch.mode\n"
     "    VirtualSwitch.name\n"
     "    VirtualSwitch.switch_id\n"
     "    VirtualSwitch.uuid\n"
     "    VirtualSwitch.vnet_uri_list\n"])

DELETE_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualSwitch delete "
                      "{--object-id | -i} <VirtualSwitch field>=value "
                      "{args}"])

UPDATE_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualSwitch update "
                      "{--object-id | -i} <VirtualSwitch field>=value "
                      "{args}"])

SET_FIELD_HELP = ([
    "    VirtualSwitch.mode\n"
    "    VirtualSwitch.name\n"
    "    VirtualSwitch.vnet_uri_list\n"])


class VirtualSwitchTest(twa.TestWrapper):
    ''' Test framework for VirtualSwitch Object Type. '''
    @property
    def file(self):
        return 'fake_vswitch_feed.txt'

    @property
    def wrapper_class_to_test(self):
        return n.VSwitch

    def test_virtual_switch(self):
        vs = vs_c.VirtualSwitch()

        test_util.test_obj_type(self, vs, 'VirtualSwitch', n.VSwitch,
                                field_models=[n.VSwitch],
                                name_to_ops={op.CREATE_NAME, op.LIST_NAME,
                                             op.UPDATE_NAME, op.DELETE_NAME},
                                exp_help=help_vswitch.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={vs.name_to_ops['list']: [ms_w.System],
                                    vs.name_to_ops['update']: [ms_w.System],
                                    vs.name_to_ops['delete']: [ms_w.System],
                                    vs.name_to_ops['create']:
                                    [ms_w.System]},
            child_model_ops={vs.name_to_ops['list']: None,
                             vs.name_to_ops['update']: None,
                             vs.name_to_ops['delete']: None,
                             vs.name_to_ops['create']: None},
            elem_wrp_func_ops={vs.name_to_ops['list']: None,
                               vs.name_to_ops['update']: None,
                               vs.name_to_ops['delete']: None,
                               vs.name_to_ops['create']: None})

    @mock.patch('pypowervm.wrappers.network.VSwitch.bld')
    def test_create(self, bld_mock):
        cp = test_util.setup_mock(self, args=['VirtualSwitch', 'create',
                                              '--name', 'vswitch_name'],
                                  adpt_fx=self.adpt)
        cp.obj_type.execute()
        bld_mock.assert_called_once_with(adapter=cp.obj_type.ec.sm.adapter,
                                         name='vswitch_name',
                                         switch_mode='Veb')
        bld_mock.reset_mock()

        cp = test_util.setup_mock(self, args=['VirtualSwitch', 'create',
                                              '--name', 'vswitch_name2',
                                              '--mode', 'vEb'],
                                  adpt_fx=self.adpt)
        cp.obj_type.execute()
        bld_mock.assert_called_once_with(adapter=cp.obj_type.ec.sm.adapter,
                                         name='vswitch_name2',
                                         switch_mode='Veb')
        bld_mock.reset_mock()

        cp = test_util.setup_mock(self, args=['VirtualSwitch', 'create',
                                              '--name', 'vswitch_name3',
                                              '--mode', 'VEPa'],
                                  adpt_fx=self.adpt)
        cp.obj_type.execute()
        bld_mock.assert_called_once_with(adapter=cp.obj_type.ec.sm.adapter,
                                         name='vswitch_name3',
                                         switch_mode='Vepa')
        bld_mock.reset_mock()

        # No name argument
        cp = cli_parser.CLIParser(args=['VirtualSwitch', 'create',
                                        '--mode', 'vepa'])
        se_exc = self.assertRaises(SystemExit, cp.process)
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

        # Invalid mode arg argument
        cp = cli_parser.CLIParser(args=['VirtualSwitch', 'create', '--name',
                                        'vswitch_name', '--mode', 'wrong'])
        se_exc = self.assertRaises(SystemExit, cp.process)
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

    @mock.patch("pypowervm.wrappers.network.VNet.wrap")
    def test_list(self, vnet_wrap_mock):
        test_vnet = n.VNet.bld(self.adpt, "test_vnet", 1, "abcd", False)
        vnet_wrap_mock.return_value = test_vnet
        test_util.test_list(
            self, 'VirtualSwitch',
            test_util.EXPECTED_HELP + LIST_USAGE_HELP + DISPLAY_FIELD_HELP,
            ["Virtual Switches\n+-----------+----+------+-----------+\n|    N"
             "ame   | ID | Mode |   VNets   |\n+-----------+----+------+-----"
             "------+\n| ETHERNET0 | 0  | Veb  | test_vnet |\n|           |  "
             "  |      | test_vnet |\n|           |    |      | test_vnet |\n"
             "|           |    |      | test_vnet |\n|           |    |      "
             "| test_vnet |\n|           |    |      | test_vnet |\n+--------"
             "---+----+------+-----------+\n"],
            ["mode=Veb\n  name=ETHERNET0\n  switch_id=0\n  uuid=4d9735ae-feaf"
             "-32c2-a1bc-102026df9168\n  vnet_uri_list=['https://9.1.2.3:1244"
             "3/rest/api/uom/ManagedSystem/4abca7ff-3710-3160-b9e4-cb4456c33f"
             "43/VirtualNetwork/f3259c07-1db9-3aac-a2b5-eeb72a0388f8', 'https"
             "://9.1.2.3:12443/rest/api/uom/ManagedSystem/4abca7ff-3710-3160-"
             "b9e4-cb4456c33f43/VirtualNetwork/c4e38a65-221d-3e4e-8151-601dae"
             "8f5b66', "],
            ['    <SwitchMode kxe="false" kb="CUD">Veb</SwitchMode>'],
            ['name', 'mode'],
            ['name=ETHERNET0,mode=Veb\n'],
            ['ETHERNET0,Veb\n'],
            ["name___ETHERNET0,mode___Veb\n"],
            ["name=ETHERNET0___mode=Veb\n"],
            ["name=ETHERNET0,mode=Veb\n"],
            ["name___ETHERNET0___mode___Veb\n"],
            {'child_type': 'VirtualSwitch',
             'root_id': 'caae9209-25e5-35cd-a71a-ed55c03f294d',
             'root_type': 'ManagedSystem'},
            p_wrps=[test_util.mock_sys(self.adpt)])

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.update')
    def test_update(self, ew_update_mock):
        help_strs = UPDATE_USAGE_HELP + SET_FIELD_HELP
        test_util.test_update(
            self, ew_update_mock, 'VirtualSwitch',
            ["mode=VePa", "name=foo"], "VirtualSwitch.fake=123", help_strs,
            "Set field VirtualSwitch.fake=123 has an improper format. "
            "Please specify the set field as <Field>=<Value>. Run the command "
            "'pvmctl VirtualSwitch update help' for a list of valid "
            "set fields.")

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.delete')
    def test_delete(self, ew_mock_delete):
        ''' Test delete operation on VirtualSwitch ObjectType. '''
        test_util.test_delete(self, ew_mock_delete, 'VirtualSwitch',
                              DELETE_USAGE_HELP, wrps=[self.dwrap],
                              p_wrps=[test_util.mock_sys(self.adpt)])
