# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
import six

import pvmcli.command.vnic as vnic_c
import pvmcli.help.help_vnic as help_vnic
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc

import pypowervm.wrappers.iocard as iocard
import pypowervm.wrappers.logical_partition as lp_w
import pypowervm.tests.test_utils.test_wrapper_abc as twa


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualNICDedicated list "
                    "{--parent-id | -p} <LogicalPartition field>=value "
                    "{--object-id | -i} <VirtualNICDedicated field>=value "
                    "{args}"])


UPDATE_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualNICDedicated "
                      "update {--parent-id | -p} <LogicalPartition field>="
                      "value {--object-id | -i} <VirtualNICDedicated field>="
                      "value {args}"])

DELETE_USAGE_HELP = (
    ["usage: pvmctl [help | --help] VirtualNICDedicated delete "
     "{--parent-id | -p} <LogicalPartition field>=value {--object-id | -i} "
     "<VirtualNICDedicated field>=value {args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    VirtualNICDedicated.allowed_macs\n"
     "    VirtualNICDedicated.allowed_vlans\n"
     "    VirtualNICDedicated.auto_pri_failover\n"
     "    VirtualNICDedicated.drc_name\n"
     "    VirtualNICDedicated.gateway\n"
     "    VirtualNICDedicated.ip_address\n"
     "    VirtualNICDedicated.lpar_id\n"
     "    VirtualNICDedicated.mac\n"
     "    VirtualNICDedicated.pvid\n"
     "    VirtualNICDedicated.slot\n"
     "    VirtualNICDedicated.subnet_mask\n"
     "    VirtualNICDedicated.uuid\n"
     "    VirtualNICSRIOVBackingDevice.capacity\n"
     "    VirtualNICSRIOVBackingDevice.desired_max_capacity\n"
     "    VirtualNICSRIOVBackingDevice.failover_pri\n"
     "    VirtualNICSRIOVBackingDevice.is_active\n"
     "    VirtualNICSRIOVBackingDevice.lport_href\n"
     "    VirtualNICSRIOVBackingDevice.max_capacity\n"
     "    VirtualNICSRIOVBackingDevice.pport_id\n"
     "    VirtualNICSRIOVBackingDevice.sriov_adap_id\n"
     "    VirtualNICSRIOVBackingDevice.status\n"
     "    VirtualNICSRIOVBackingDevice.vios_href\n\n"])

SET_FIELD_HELP = (
    ["    VirtualNICDedicated.allowed_macs\n"
     "    VirtualNICDedicated.allowed_vlans\n"
     "    VirtualNICDedicated.auto_pri_failover\n"
     "    VirtualNICDedicated.pvid\n"
     "    VirtualNICSRIOVBackingDevice.failover_pri\n"])


class VNICTest(twa.TestWrapper):
    ''' Test framework for VNIC Object Type. '''
    @property
    def file(self):
        return 'vnic_feed.txt'

    @property
    def wrapper_class_to_test(self):
        return iocard.VNIC

    def test_vnic(self):
        vn = vnic_c.VNIC()

        test_util.test_obj_type(self, vn, 'VirtualNICDedicated', iocard.VNIC,
                                field_models=[iocard.VNIC, iocard.VNICBackDev],
                                name_to_ops={'list', 'delete', 'update',
                                             'create', 'add-back-devs',
                                             'remove-back-devs'},
                                exp_help=help_vnic.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={vn.name_to_ops['list']: [lp_w.LPAR]},
            child_model_ops={vn.name_to_ops['list']: None},
            elem_wrp_func_ops={vn.name_to_ops['list']: None})
        test_util.test_ops(
            self, parent_model_ops={vn.name_to_ops['delete']: [lp_w.LPAR]},
            child_model_ops={vn.name_to_ops['delete']: None},
            elem_wrp_func_ops={vn.name_to_ops['delete']: None})
        test_util.test_ops(
            self, parent_model_ops={vn.name_to_ops['create']: [lp_w.LPAR]},
            child_model_ops={vn.name_to_ops['create']: None},
            elem_wrp_func_ops={vn.name_to_ops['create']: None})
        test_util.test_ops(
            self, parent_model_ops={vn.name_to_ops['update']: [lp_w.LPAR]},
            child_model_ops={vn.name_to_ops['update']: None},
            elem_wrp_func_ops={vn.name_to_ops['update']: None})
        test_util.test_ops(
            self,
            parent_model_ops={vn.name_to_ops['add-back-devs']: [lp_w.LPAR]},
            child_model_ops={vn.name_to_ops['add-back-devs']: None},
            elem_wrp_func_ops={vn.name_to_ops['add-back-devs']: None})
        test_util.test_ops(
            self,
            parent_model_ops={vn.name_to_ops['remove-back-devs']: [lp_w.LPAR]},
            child_model_ops={vn.name_to_ops['remove-back-devs']: None},
            elem_wrp_func_ops={vn.name_to_ops['remove-back-devs']: None})

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'VNIC')

    @mock.patch('pvmcli.command.vnic.VNIC._vios_name_by_href',
                return_value='VIOS_NAME')
    def test_list(self, mock_rbp):
        lp_mock = test_util.mock_lpar(self.adpt)
        lp_mock._id(10)
        test_util.test_list(
            self, 'VNIC',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["VNICs\n"
             "+-------------------------+--------------+------+------+"
             "----------------------------+-----------+\n"
             "|           LPAR          |     MAC      | Slot | PVID |"
             "       BDev PPort Loc       |    VIOS   |\n"
             "+-------------------------+--------------+------+------+"
             "----------------------------+-----------+\n"
             "| z3-9-5-126-127-00000001 | AE7A25E59A07 |  7   |  0   |"
             " U78C7.001.RCH0004-P1-C8-T1 | VIOS_NAME |\n"
             "|                         |              |      |      |"
             " U78C7.001.RCH0004-P1-C9-T1 | VIOS_NAME |\n"
             "|                         | AE7A25E59A08 |  8   |  0   |"
             " U78C7.001.RCH0004-P1-C9-T1 | VIOS_NAME |\n"
             "+-------------------------+--------------+------+------+"
             "----------------------------+-----------+\n"],
            ["-----------------------\n"
             "| VirtualNICDedicated |\n"
             "-----------------------\n"
             "  allowed_macs=ALL\n"
             "  allowed_vlans=ALL\n"
             "  auto_pri_failover=False\n"
             "  capacity=2.00%\n"
             "  drc_name=U8286.42A.21C1B6V-V10-C3\n"
             "  gateway=192.168.2.0\n"
             "  ip_address=192.168.2.7\n"
             "  lpar_id=10\n"
             "  mac=AE7A25E59A07\n"
             "  pvid=0\n"
             "  slot=7\n"
             "  subnet_mask=255.255.255.0\n"
             "  uuid=c4cd2502-caf5-399d-97f8-b9644ac8c9f9\n"
             "  --------------------------------\n"
             "  | VirtualNICSRIOVBackingDevice |\n"
             "  --------------------------------\n"
             "    capacity=2.00%\n"
             "    desired_max_capacity=100.00%\n"
             "    is_active=False\n"
             "    lport_href=https://9.1.2.3:12443/rest/api/uom/VirtualIO"
             "Server/3E3F9BFC-C4EE-439E-B70A-1D369213ED83/SRIOVEthernetLog"
             "icalPort/af2a8c95-58d1-3496-9af6-8cd562f0e839\n"
             "    max_capacity=100.00%\n"
             "    pport_id=0\n"
             "    sriov_adap_id=1\n"
             "    status=LINK_DOWN\n"
             "    vios_href=https://9.1.2.3:12443/rest/api/uom/ManagedSystem/"
             "1cab7366-6b73-342c-9f43-ddfeb9f8edd3/VirtualIOServer/3E3F9BFC-C"
             "4EE-439E-B70A-1D369213ED83\n"
             "  --------------------------------\n"
             "  | VirtualNICSRIOVBackingDevice |\n"
             "  --------------------------------\n"
             "    capacity=42.42%\n"
             "    desired_max_capacity=100.00%\n"
             "    is_active=True\n"
             "    lport_href=https://9.1.2.3:12443/rest/api/uom/VirtualIOServ"
             "er/3E3F9BFC-C4EE-439E-B70A-1D369213ED83/SRIOVEthernetLogicalPo"
             "rt/af2a8c95-58d1-3496-9af6-8cd562f0e839\n"
             "    max_capacity=17.00%\n"
             "    pport_id=0\n"
             "    sriov_adap_id=2\n"
             "    status=OPERATIONAL\n"
             "    vios_href=https://9.1.2.3:12443/rest/api/uom/ManagedSystem"
             "/1cab7366-6b73-342c-9f43-ddfeb9f8edd3/VirtualIOServer/3E3F9BFC"
             "-C4EE-439E-B70A-1D369213ED83\n"],
            ['<VirtualNICDedicated:VirtualNICDedicated xmlns:Virtu'],
            ['drc_name', 'pport_id'],
            ["drc_name=U8286.42A.21C1B6V-V10-C3,pport_id=0\n"],
            ["U8286.42A.21C1B6V-V10-C3,0\n"],
            ["drc_name___U8286.42A.21C1B6V-V10-C3,pport_id___0\n"],
            ["drc_name=U8286.42A.21C1B6V-V10-C3___pport_id=0\n"],
            ["drc_name=U8286.42A.21C1B6V-V10-C3,pport_id=0\n"],
            ["drc_name___U8286.42A.21C1B6V-V10-C3___pport_id___0\n"],
            {'root_type': 'LogicalPartition',
             'root_id': '089FFB20-5D19-4A8C-BB80-13650627D985',
             'child_type': 'VirtualNICDedicated'},
            wrps=[self.dwrap], p_wrps=[lp_mock],
            msys_mock_file='sys_with_sriov.txt')
        mock_rbp.assertEqual(mock_rbp.call_count, 3)

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.delete')
    def test_delete(self, ew_mock_delete):
        ''' Test delete operation. '''
        test_util.test_delete(self, ew_mock_delete, 'VNIC',
                              DELETE_USAGE_HELP, wrps=[self.dwrap],
                              p_wrps=[test_util.mock_lpar(self.adpt)])

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.create')
    def test_create(self, create_mock, vios_mock):
        sample_vios_w = test_util.mock_vios(self.adpt)
        vios_mock.return_value = [sample_vios_w]

        # Minimum args
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        # Add capacity to backing device
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1', 'cap=42%']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])

        create_mock.reset_mock()

        # Percentage sign on capacity
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1',
                    'cap=42.0%']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1', 'cap=42.0']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1', 'pri=42']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        full_bd = ['--back-dev', 'vios=vios1',
                   'loc=U78C7.001.RCH0004-P1-C8-T1', 'cap=42%', 'pri=42']
        all_args = ['vnic', 'create', '-p', 'name=lpar_name']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'FF:FF:FF:FF:FF:FF', '11:11:11:11:11:11',
                    '--allowed-vlans', '1', '2', '42']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'none', '--allowed-vlans', 'none']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        # Test 2 backing devices
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'none', '--allowed-vlans', 'none']
        all_args.extend(full_bd)
        all_args.extend(full_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        create_mock.assert_called_once_with(
            parent=cp.obj_type.ec.elem_p_wrappers[0])
        create_mock.reset_mock()

        # Failure test cases
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'FF:FF:FF:FF:FF:FF', '11:11:11:11:11:11',
                    '--allowed-vlans', '1', '2', '42']
        invalid_bd = ['--back-dev', 'loc=U78C7.001.RCH0004-P1-C8-T1']
        all_args.extend(invalid_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        self.assertRaises(exc.InvalidBackDevError, cp.obj_type.execute)
        create_mock.assert_not_called()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'FF:FF:FF:FF:FF:FF', '11:11:11:11:11:11',
                    '--allowed-vlans', '1', '2', '42']
        invalid_bd = ['--back-dev', 'vios=vios1', 'cap=42.0%', 'pri=42']
        all_args.extend(invalid_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        self.assertRaises(exc.InvalidBackDevError, cp.obj_type.execute)
        create_mock.assert_not_called()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'FF:FF:FF:FF:FF:FF', '11:11:11:11:11:11',
                    '--allowed-vlans', '1', '2', '42']
        invalid_bd = ['--back-dev', 'cap=42.0%', 'pri=42']
        all_args.extend(invalid_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        self.assertRaises(exc.InvalidBackDevError, cp.obj_type.execute)
        create_mock.assert_not_called()

    @mock.patch('pypowervm.wrappers.iocard.VNIC.bld')
    @mock.patch('pypowervm.wrappers.iocard.VNICBackDev.bld')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    def test_create_bld(self, vios_mock, backdev_mock, vnic_mock):
        sample_vios_w = test_util.mock_vios(self.adpt)
        vios_mock.return_value = [sample_vios_w]
        sample_backdev = iocard.VNICBackDev.bld()
        backdev_mock.return_value = sample_backdev
        backdev_mock.reset_mock()

        # Minimum args
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, pport_id=0,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA', sriov_adap_id=1,
            failover_pri=None, capacity=None, max_capacity=None)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, pvid=None, allowed_vlans='ALL',
            mac_addr=None, allowed_macs='ALL', slot_num=None,
            back_devs=[sample_backdev])
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        # Add capacity to backing device
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1', 'cap=42%',
                    'maxcap=84%']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=None, pport_id=0, max_capacity=0.84)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, pvid=None, allowed_vlans='ALL',
            mac_addr=None, allowed_macs='ALL', slot_num=None,
            back_devs=[sample_backdev])
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        # Percentage sign on capacity
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1',
                    'cap=42.0%']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=None, pport_id=0, max_capacity=None)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, pvid=None, allowed_vlans='ALL',
            mac_addr=None, allowed_macs='ALL', slot_num=None,
            back_devs=[sample_backdev])
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1', 'cap=42.0',
                    'maxcap=84.0']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=None, pport_id=0, max_capacity=0.84)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, pvid=None, allowed_vlans='ALL',
            mac_addr=None, allowed_macs='ALL', slot_num=None,
            back_devs=[sample_backdev])
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1', 'pri=42']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=None, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=42, pport_id=0, max_capacity=None)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, pvid=None, allowed_vlans='ALL',
            mac_addr=None, allowed_macs='ALL', slot_num=None,
            back_devs=[sample_backdev])
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        full_bd = ['--back-dev', 'vios=vios1',
                   'loc=U78C7.001.RCH0004-P1-C8-T1', 'cap=42%', 'pri=42',
                   'maxcap=84']
        all_args = ['vnic', 'create', '-p', 'name=lpar_name']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=42, pport_id=0, max_capacity=0.84)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, pvid=None, allowed_vlans='ALL',
            mac_addr=None, allowed_macs='ALL', slot_num=None,
            back_devs=[sample_backdev])
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'FF:FF:FF:FF:FF:FF', '11:11:11:11:11:11',
                    '--allowed-vlans', '1', '2', '42']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=42, pport_id=0, max_capacity=0.84)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, mac_addr='FF:FF:FF:FF:FF:FF',
            allowed_vlans=[1, 2, 42], slot_num=42,
            allowed_macs=['FFFFFFFFFFFF', '111111111111'],
            back_devs=[sample_backdev], pvid=42)
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'FF:FF:FF:FF:FF:AA,11:11:11:11:11:22',
                    '--allowed-vlans', '3,4,53']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=42, pport_id=0, max_capacity=0.84)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, mac_addr='FF:FF:FF:FF:FF:FF',
            allowed_vlans=[3, 4, 53], slot_num=42,
            allowed_macs=['FFFFFFFFFFAA', '111111111122'],
            back_devs=[sample_backdev], pvid=42)
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'none', '--allowed-vlans', 'none']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=42, pport_id=0, max_capacity=0.84)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, allowed_macs='NONE',
            mac_addr='FF:FF:FF:FF:FF:FF', back_devs=[sample_backdev],
            allowed_vlans='NONE', slot_num=42, pvid=42)
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'aLl', '--allowed-vlans', 'AlL']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=42, pport_id=0, max_capacity=0.84)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, allowed_macs='ALL',
            mac_addr='FF:FF:FF:FF:FF:FF', back_devs=[sample_backdev],
            allowed_vlans='ALL', slot_num=42, pvid=42)
        backdev_mock.reset_mock()
        vnic_mock.reset_mock()

        # Test 2 backing devices
        all_args = ['vnic', 'create', '-p', 'name=lpar_name', '--mac',
                    'FF:FF:FF:FF:FF:FF', '--slot', '42', '--pvid', '42',
                    '--allowed-macs', 'none', '--allowed-vlans', 'none']
        full_bd2 = ['--back-dev', 'vios=vios2',
                    'loc=U78C7.001.RCH0004-P1-C8-T2', 'maxcap=11%']
        all_args.extend(full_bd)
        all_args.extend(full_bd2)
        cp = test_util.setup_mock(self, args=all_args,
                                  adpt_fx=self.adpt,
                                  msys_mock_file='sys_with_sriov.txt',
                                  p_wrps=[test_util.mock_lpar(self.adpt)])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        backdev_mock.assert_any_call(
            cp.obj_type.ec.sm.adapter, capacity=0.42, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=42, pport_id=0, max_capacity=0.84)
        backdev_mock.assert_any_call(
            cp.obj_type.ec.sm.adapter, capacity=None, sriov_adap_id=1,
            vios_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            failover_pri=None, pport_id=1, max_capacity=0.11)
        vnic_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, allowed_macs='NONE',
            mac_addr='FF:FF:FF:FF:FF:FF',
            back_devs=[sample_backdev, sample_backdev], allowed_vlans='NONE',
            slot_num=42, pvid=42)

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.update')
    def test_update(self, ew_update_mock):
        help_strs = UPDATE_USAGE_HELP + SET_FIELD_HELP
        test_util.test_update(
            self, ew_update_mock, 'vnic', ["pvid=42"],
            "VirtualNICDedicated.notathing=123", help_strs,
            "Set field VirtualNICDedicated.notathing=123 has an improper "
            "format. Please specify the set field as <Field>=<Value>. Run "
            "the command 'pvmctl VirtualNICDedicated update help' for a list "
            "of valid set fields.")

        with mock.patch('pypowervm.wrappers.iocard.VNIC.update') as v_mock:
            sample_vnic_w = test_util.mock_vnic(self.adpt)
            v_mock.return_value = sample_vnic_w
            all_args = ['vnic', 'update', '-p', 'name=lpar_name',
                        '-i', 'mac=11:11:11:11:11:11', '-s',
                        'VirtualNICDedicated.allowed_vlans=3 4']
            cp = test_util.setup_mock(
                self, args=all_args, adpt_fx=self.adpt,
                msys_mock_file='sys_with_sriov.txt',
                p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

            self.assertEqual(sample_vnic_w.allowed_vlans, 'ALL')
            cp.obj_type.execute()
            v_mock.assert_called_once_with()
            self.assertEqual(sample_vnic_w.allowed_vlans, [3, 4])
            v_mock.reset_mock()

            all_args = ['vnic', 'update', '-p', 'name=lpar_name',
                        '-i', 'mac=11:11:11:11:11:11', '-s',
                        "VirtualNICDedicated.allowed_macs=11:11:11:11:11:11 "
                        "22:22:22:22:22:22"]
            cp = test_util.setup_mock(
                self, args=all_args, adpt_fx=self.adpt,
                msys_mock_file='sys_with_sriov.txt',
                p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

            self.assertEqual(sample_vnic_w.allowed_macs, 'ALL')
            cp.obj_type.execute()
            v_mock.assert_called_once_with()
            self.assertEqual(sample_vnic_w.allowed_macs, ['111111111111',
                                                          '222222222222'])
            v_mock.reset_mock()

            all_args = ['vnic', 'update', '-p', 'name=lpar_name',
                        '-i', 'mac=11:11:11:11:11:11', '-s',
                        'VirtualNICDedicated.allowed_vlans=4,5']
            cp = test_util.setup_mock(
                self, args=all_args, adpt_fx=self.adpt,
                msys_mock_file='sys_with_sriov.txt',
                p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

            self.assertEqual(sample_vnic_w.allowed_vlans, [3, 4])
            cp.obj_type.execute()
            v_mock.assert_called_once_with()
            self.assertEqual(sample_vnic_w.allowed_vlans, [4, 5])
            v_mock.reset_mock()

            all_args = ['vnic', 'update', '-p', 'name=lpar_name',
                        '-i', 'mac=11:11:11:11:11:11', '-s',
                        "VirtualNICDedicated.allowed_macs=11:11:11:11:11:EE,"
                        "22:22:22:22:22:FF"]
            cp = test_util.setup_mock(
                self, args=all_args, adpt_fx=self.adpt,
                msys_mock_file='sys_with_sriov.txt',
                p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

            self.assertEqual(sample_vnic_w.allowed_macs, ['111111111111',
                                                          '222222222222'])
            cp.obj_type.execute()
            v_mock.assert_called_once_with()
            self.assertEqual(sample_vnic_w.allowed_macs, ['1111111111EE',
                                                          '2222222222FF'])
            v_mock.reset_mock()

        # Test back dev update
        with mock.patch('pvmcli.command.object_type.update_retry') as ur_mock:
            sample_vnic_w = test_util.mock_vnic(self.adpt)
            sample_vnic_w.back_devs[0].failover_pri = 2
            all_args = ['vnic', 'update', '-p', 'name=lpar_name',
                        '-i', 'mac=11:11:11:11:11:11', '--loc',
                        'U78C7.001.RCH0004-P1-C8-T1', '-s',
                        'VirtualNICSRIOVBackingDevice.failover_pri=20',
                        'pvid=42']
            cp = test_util.setup_mock(
                self, args=all_args, adpt_fx=self.adpt,
                msys_mock_file='sys_with_sriov.txt',
                p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

            self.assertEqual(sample_vnic_w.back_devs[0].failover_pri, 2)
            cp.obj_type.execute()
            # Should not pass in backing device setter to update_retry since
            # it is handled separately
            ur_mock.assert_called_once_with(
                sample_vnic_w, {'VirtualNICDedicated.pvid': '42'}, cp.obj_type)
            self.assertEqual(sample_vnic_w.back_devs[0].failover_pri, 20)
            ur_mock.reset_mock()

            # Test no --loc on backing device update
            all_args = ['vnic', 'update', '-p', 'name=lpar_name',
                        '-i', 'mac=11:11:11:11:11:11', '-s',
                        'VirtualNICSRIOVBackingDevice.failover_pri=20',
                        'pvid=42']
            cp = test_util.setup_mock(
                self, args=all_args, adpt_fx=self.adpt,
                msys_mock_file='sys_with_sriov.txt',
                p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])
            self.assertRaises(exc.VNICBackDevLocRequiredError,
                              cp.obj_type.execute)
            ur_mock.assert_not_called()

    @mock.patch('pypowervm.wrappers.iocard.VNIC.update')
    @mock.patch('pypowervm.wrappers.iocard.VNIC.search')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    def test_add_back_devs(self, vios_mock, vnic_search_mock, vnic_upd_mock):
        # pvmctl vnic add-back-devs help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['vnic', 'add-back-devs', 'help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] VirtualNICDedicated "
                      "add-back-devs {--parent-id | -p} <LogicalPartition "
                      "field>=value {--object-id | -i} <VirtualNICDedicated "
                      "field>=value {args}", help_out.getvalue())
        vnic_upd_mock.assert_not_called()
        vnic_upd_mock.reset_mock()

        sample_vios_w = test_util.mock_vios(self.adpt)
        vios_mock.return_value = [sample_vios_w]
        sample_vnic_w = test_util.mock_vnic(self.adpt)
        sample_vnic_w.back_devs = []
        vnic_search_mock.return_value = [sample_vnic_w]

        # Minimum args
        all_args = ['vnic', 'add-back-devs', '-p', 'name=lpar_name',
                    '-i', 'mac=11:11:11:11:11:11', '--back-dev',
                    'vios=vios1', 'loc=U78C7.001.RCH0004-P1-C8-T1']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        self.assertEqual(len(sample_vnic_w.back_devs), 1)
        self.assertEqual(sample_vnic_w.back_devs[0].vios_href,
                         sample_vios_w.href)
        vnic_upd_mock.assert_called_once_with()
        vnic_upd_mock.reset_mock()

        full_bd = ['--back-dev', 'vios=vios1',
                   'loc=U78C7.001.RCH0004-P1-C8-T1', 'cap=42%', 'pri=42']
        all_args = ['vnic', 'add-back-devs', '-p', 'name=lpar_name',
                    '-i', 'mac=11:11:11:11:11:11']
        all_args.extend(full_bd)
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        self.assertEqual(len(sample_vnic_w.back_devs), 2)
        self.assertEqual(sample_vnic_w.back_devs[1].vios_href,
                         sample_vios_w.href)
        vnic_upd_mock.assert_called_once_with()
        vnic_upd_mock.reset_mock()

        all_args = ['vnic', 'add-back-devs', '-p', 'name=lpar_name',
                    '-i', 'mac=11:11:11:11:11:11']
        all_args.extend(full_bd)
        all_args.extend(full_bd)
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        cp.obj_type.execute()
        self.assertEqual(len(sample_vnic_w.back_devs), 4)
        self.assertEqual(sample_vnic_w.back_devs[2].vios_href,
                         sample_vios_w.href)
        self.assertEqual(sample_vnic_w.back_devs[3].vios_href,
                         sample_vios_w.href)
        vnic_upd_mock.assert_called_once_with()
        vnic_upd_mock.reset_mock()

    @mock.patch('pypowervm.tasks.sriov.find_pport')
    @mock.patch('pypowervm.wrappers.iocard.VNIC.update')
    @mock.patch('pypowervm.wrappers.iocard.VNIC.search')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    def test_rem_back_devs(self, vios_mock, vnic_search_mock, vnic_upd_mock,
                           find_pport_mock):
        # pvmctl vnic remove-back-devs help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['vnic', 'remove-back-devs', 'help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] VirtualNICDedicated "
                      "remove-back-devs {--parent-id | -p} <LogicalPartition "
                      "field>=value {--object-id | -i} <VirtualNICDedicated "
                      "field>=value {args}", help_out.getvalue())
        vnic_upd_mock.assert_not_called()
        vnic_upd_mock.reset_mock()

        sample_vios_w = test_util.mock_vios(self.adpt)
        vios_mock.return_value = [sample_vios_w]
        sample_vnic_w = test_util.mock_vnic(self.adpt)
        curr_bds = sample_vnic_w.back_devs
        self.adpt.build_href.return_value = sample_vios_w.href
        curr_bds.append(iocard.VNICBackDev.bld(
            self.adpt, sample_vios_w.uuid, 2, 3))
        sample_vnic_w.back_devs = curr_bds
        vnic_search_mock.return_value = [sample_vnic_w]
        find_pport_mock.return_value = test_util.mock_sys_with_sriov(
            self.adpt).asio_config.sriov_adapters[0].phys_ports[0]

        all_args = ['vnic', 'remove-back-devs', '-p', 'name=lpar_name',
                    '-i', 'mac=11:11:11:11:11:11', '--loc',
                    'U78C7.001.RCH0004-P1-C8-T1']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        self.assertEqual(len(sample_vnic_w.back_devs), 3)
        self.assertEqual(sample_vnic_w.back_devs[0].sriov_adap_id, 1)
        self.assertEqual(sample_vnic_w.back_devs[0].pport_id, 0)
        self.assertEqual(sample_vnic_w.back_devs[1].sriov_adap_id, 2)
        self.assertEqual(sample_vnic_w.back_devs[1].pport_id, 0)
        cp.obj_type.execute()
        self.assertEqual(len(sample_vnic_w.back_devs), 2)
        find_pport_mock.assert_called_once_with(cp.obj_type.ec.msys,
                                                'U78C7.001.RCH0004-P1-C8-T1')
        find_pport_mock.reset_mock()
        vnic_upd_mock.assert_called_once_with()
        vnic_upd_mock.reset_mock()

        find_pport_mock.return_value = None

        all_args = ['vnic', 'remove-back-devs', '-p', 'name=lpar_name',
                    '-i', 'mac=11:11:11:11:11:11', '--loc',
                    'U78C7.001.RCH0004-P1-C8-T1']
        cp = test_util.setup_mock(
            self, args=all_args, adpt_fx=self.adpt,
            msys_mock_file='sys_with_sriov.txt',
            p_wrps=[test_util.mock_lpar(self.adpt)], wrps=[sample_vnic_w])

        cp.obj_type.ec.sm.adapter.build_href.return_value = sample_vios_w.href
        self.assertRaises(exc.NoPPortFoundFromLocError, cp.obj_type.execute)
        find_pport_mock.assert_called_once_with(cp.obj_type.ec.msys,
                                                'U78C7.001.RCH0004-P1-C8-T1')
        find_pport_mock.reset_mock()
        vnic_upd_mock.assert_not_called()
        vnic_upd_mock.reset_mock()
