# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import copy
import mock

import pvmcli.command.volume_group as vg_c
import pvmcli.help.help_volume_group as help_vg
import pvmcli.tests.util.test_util as test_util

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.storage as stor
from pypowervm.wrappers import virtual_io_server as v


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] VolumeGroup list "
                    "{--parent-id | -p} <VirtualIOServer field>=value "
                    "{--object-id | -i} <VolumeGroup field>=value "
                    "{args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    LogicalVolume.backstore_type\n"
     "    LogicalVolume.capacity\n"
     "    LogicalVolume.emulate_model\n"
     "    LogicalVolume.file_format\n"
     "    LogicalVolume.label\n"
     "    LogicalVolume.name\n"
     "    LogicalVolume.read_iops_limit\n"
     "    LogicalVolume.tag\n"
     "    LogicalVolume.udid\n"
     "    LogicalVolume.vdtype\n"
     "    LogicalVolume.vg_uri\n"
     "    LogicalVolume.write_iops_limit\n"
     "    PhysicalVolume.avail_for_use\n"
     "    PhysicalVolume.capacity\n"
     "    PhysicalVolume.description\n"
     "    PhysicalVolume.emulate_model\n"
     "    PhysicalVolume.is_fc_backed\n"
     "    PhysicalVolume.loc_code\n"
     "    PhysicalVolume.name\n"
     "    PhysicalVolume.read_iops_limit\n"
     "    PhysicalVolume.state\n"
     "    PhysicalVolume.tag\n"
     "    PhysicalVolume.udid\n"
     "    PhysicalVolume.write_iops_limit\n"
     "    VirtualMediaRepository.name\n"
     "    VirtualMediaRepository.optical_media\n"
     "    VirtualMediaRepository.size\n"
     "    VirtualOpticalMedia.media_name\n"
     "    VirtualOpticalMedia.mount_type\n"
     "    VirtualOpticalMedia.name\n"
     "    VirtualOpticalMedia.size\n"
     "    VirtualOpticalMedia.udid\n"
     "    VolumeGroup.available_size\n"
     "    VolumeGroup.capacity\n"
     "    VolumeGroup.free_space\n"
     "    VolumeGroup.name\n"
     "    VolumeGroup.phys_vols\n"
     "    VolumeGroup.serial_id\n"
     "    VolumeGroup.uuid\n"
     "    VolumeGroup.virtual_disks\n"
     "    VolumeGroup.vmedia_repos\n"])


class VolumeGroupTest(twa.TestWrapper):
    ''' Test framework for VolumeGroup Object Type. '''
    @property
    def file(self):
        return 'fake_volume_group.txt'

    @property
    def wrapper_class_to_test(self):
        return stor.VG

    def test_volume_group(self):
        vg = vg_c.VolumeGroup()

        test_util.test_obj_type(self, vg, 'VolumeGroup', stor.VG,
                                field_models=[stor.VG, stor.VMediaRepos,
                                              stor.PV, stor.VDisk,
                                              stor.VOptMedia],
                                name_to_ops={'list'},
                                exp_help=help_vg.help_text)
        self.assertEqual(vg.get_schema(stor.VDisk.schema_type),
                         vg_c.LV_SCHEMA)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={vg.name_to_ops['list']: [v.VIOS]},
            child_model_ops={vg.name_to_ops['list']: None},
            elem_wrp_func_ops={vg.name_to_ops['list']: None})

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'VolumeGroup')

    @mock.patch("pypowervm.wrappers.virtual_io_server.VIOS.search")
    def test_list(self, vios_search_mock):
        sample_vios_w = test_util.mock_vios(self.adpt)
        vios_search_mock.return_value = [sample_vios_w]

        vg_wrap2 = copy.deepcopy(self.dwrap)
        vg_wrap2.name = 'longer_vg_name'
        vg_wrap2.virtual_disks = []

        test_util.test_list(
            self, 'VolumeGroup',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["Volume Groups\n"
             "+-------------+----------------+--------------+-----------+"
             "---------+--------------+\n"
             "|     VIOS    |    VG Name     | VG Size (GB) | Free (GB) |"
             " LV Name | LV Size (GB) |\n"
             "+-------------+----------------+--------------+-----------+"
             "---------+--------------+\n"
             "| vios_1_Drew |   image_pool   |    1063.3    |   1051.2  |"
             "   asdf  |     1.0      |\n"
             "| vios_1_Drew | longer_vg_name |    1063.3    |   1051.2  |"
             "         |              |\n"
             "+-------------+----------------+--------------+-----------+"
             "---------+--------------+\n"],
            ["---------------\n"
             "| VolumeGroup |\n"
             "---------------\n"
             "  available_size=1051.1\n"
             "  capacity=1063.3\n"
             "  free_space=1051.2\n"
             "  name=image_pool\n"
             "  serial_id=00f8d6de00004b000000014a54555cd9\n"
             "  uuid=b6bdbf1f-eddf-3c81-8801-9859eb6fedcb\n"
             "  ------------------\n"
             "  | PhysicalVolume |\n"
             "  ------------------\n"
             "    avail_for_use=True\n"
             "    capacity=1089592\n"
             "    description=SAS RAID 0 Disk Array\n"
             "    emulate_model=True\n"
             "    is_fc_backed=False\n"
             "    loc_code=U78C9.001.WZS0095-P1-C14-R1-L405D828300-L0\n"
             "    name=hdisk1\n"
             "    read_iops_limit=22\n"
             "    state=active\n"
             "    udid=01MUlCTSAgICAgSVBSLTAgICA1RDgyODMwMDAwMDAwMDQw\n"
             "    write_iops_limit=33\n"
             "  -----------------\n"
             "  | LogicalVolume |\n"
             "  -----------------\n"
             "    capacity=1.0\n"
             "    emulate_model=True\n"
             "    label=None\n"
             "    name=asdf\n"
             "    read_iops_limit=44\n"
             "    udid=0300f8d6de00004b000000014a54555cd9.1\n"
             "    vg_uri=https://9.1.2.3:12443/rest/api/uom/VirtualIOServer/"
             "14B854F7-42CE-4FF0-BD57-1D117054E701/VolumeGroup/b6bdbf1f-eddf-"
             "3c81-8801-9859eb6fedcb\n"
             "    write_iops_limit=55\n"
             "  --------------------------\n"
             "  | VirtualMediaRepository |\n"
             "  --------------------------\n"
             "    name=VMLibrary\n"
             "    size=11.0\n"
             "    -----------------------\n"
             "    | VirtualOpticalMedia |\n"
             "    -----------------------\n"
             "      media_name=blank_media1\n"
             "      mount_type=rw\n"
             "      name=blank_media1\n"
             "      size=0.0977\n"
             "      udid=0eblank_media1\n"
             "    -----------------------\n"
             "    | VirtualOpticalMedia |\n"
             "    -----------------------\n"
             "      media_name=blank_media_2\n"
             "      mount_type=rw\n"
             "      name=blank_media_2\n"
             "      size=0.0488\n"
             "      udid=0eblank_media_2\n"
             "---------------\n"
             "| VolumeGroup |\n"
             "---------------\n"
             "  available_size=1051.1\n"
             "  capacity=1063.3\n"
             "  free_space=1051.2\n"
             "  name=longer_vg_name\n"
             "  serial_id=00f8d6de00004b000000014a54555cd9\n"
             "  uuid=b6bdbf1f-eddf-3c81-8801-9859eb6fedcb\n"
             "  ------------------\n"
             "  | PhysicalVolume |\n"
             "  ------------------\n"
             "    avail_for_use=True\n"
             "    capacity=1089592\n"
             "    description=SAS RAID 0 Disk Array\n"
             "    emulate_model=True\n"
             "    is_fc_backed=False\n"
             "    loc_code=U78C9.001.WZS0095-P1-C14-R1-L405D828300-L0\n"
             "    name=hdisk1\n"
             "    read_iops_limit=22\n"
             "    state=active\n"
             "    udid=01MUlCTSAgICAgSVBSLTAgICA1RDgyODMwMDAwMDAwMDQw\n"
             "    write_iops_limit=33\n"
             "  --------------------------\n"
             "  | VirtualMediaRepository |\n"
             "  --------------------------\n"
             "    name=VMLibrary\n"
             "    size=11.0\n"
             "    -----------------------\n"
             "    | VirtualOpticalMedia |\n"
             "    -----------------------\n"
             "      media_name=blank_media1\n"
             "      mount_type=rw\n"
             "      name=blank_media1\n"
             "      size=0.0977\n"
             "      udid=0eblank_media1\n"
             "    -----------------------\n"
             "    | VirtualOpticalMedia |\n"
             "    -----------------------\n"
             "      media_name=blank_media_2\n"
             "      mount_type=rw\n"
             "      name=blank_media_2\n"
             "      size=0.0488\n"
             "      udid=0eblank_media_2\n"],
            ['<VolumeGroup:VolumeGroup xmlns:VolumeGroup="http://www.ibm.com'],
            ['VolumeGroup.name', 'VirtualOpticalMedia.media_name'],
            ["name=image_pool,media_name=blank_media1\n"
             "name=image_pool,media_name=blank_media_2\n"
             "name=longer_vg_name,media_name=blank_media1\n"
             "name=longer_vg_name,media_name=blank_media_2\n"],
            ["image_pool,blank_media1\n"
             "image_pool,blank_media_2\n"
             "longer_vg_name,blank_media1\n"
             "longer_vg_name,blank_media_2\n"],
            ["name___image_pool,media_name___blank_media1\n"
             "name___image_pool,media_name___blank_media_2\n"
             "name___longer_vg_name,media_name___blank_media1\n"
             "name___longer_vg_name,media_name___blank_media_2\n"],
            ["name=image_pool___media_name=blank_media1\n"
             "name=image_pool___media_name=blank_media_2\n"
             "name=longer_vg_name___media_name=blank_media1\n"
             "name=longer_vg_name___media_name=blank_media_2\n"],
            ["name=image_pool,media_name=blank_media1___"
             "name=image_pool,media_name=blank_media_2___"
             "name=longer_vg_name,media_name=blank_media1___"
             "name=longer_vg_name,media_name=blank_media_2\n"],
            ["name___image_pool___media_name___blank_media1___"
             "name___image_pool___media_name___blank_media_2___"
             "name___longer_vg_name___media_name___blank_media1___"
             "name___longer_vg_name___media_name___blank_media_2\n"],
            {'root_id': '014C1F78-B210-40B2-A12E-C82316B878CA',
             'root_type': 'VirtualIOServer',
             'child_type': 'VolumeGroup'},
            wrps=[self.dwrap, vg_wrap2], p_wrps=[sample_vios_w])
