# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

import mock
import six
import testtools

import pvmcli.command.logical_partition as lp_c
import pvmcli.command.managed_system as msys
import pvmcli.command.object_type as object_type
import pvmcli.util.exceptions as exc
import pvmcli.util.output_processor as out_proc
import pvmcli.util.util as util

import pypowervm.exceptions as pvm_exc
import pypowervm.wrappers.base_partition as bp_w
import pypowervm.wrappers.logical_partition as lp_w
from pypowervm.tests.test_utils import pvmhttp

HTTPRESP_FILE = "fake_httperror.txt"
LPAR_FILE = "lpar.txt"


class OutputProcessorTest(testtools.TestCase):
    ''' Test framework for output processor functions. '''
    def test_process_http_error(self):
        http_err_resp = pvmhttp.load_pvm_resp(HTTPRESP_FILE).response
        # Cast str response status to int to not break pypowervm but function
        # like expected by exceptions HttpError
        http_err_resp.status = int(http_err_resp.status)
        http_error = pvm_exc.HttpError(http_err_resp)

        self.assertIsNotNone(http_error,
                             "Could not load %s " %
                             HTTPRESP_FILE)
        self.assertIsNotNone(http_error.her_wrap)
        self.assertEqual(
            http_error.her_wrap.message,
            "Unexpected error occurred while fetching Cluster/SSP information "
            ": 9999-99Z*2125D4A/1 : Unable to send command to VIOS at this mom"
            "ent. VIOS 1*9999-99Z*2125D4A is busy processing some other reques"
            "t. Please retry the operation after sometime.")
        self.assertEqual(http_error.response.status, 500)

        msys_ot = msys.ManagedSystem()
        msys_ot.cli_err = six.StringIO()
        out_proc.process_http_error(msys_ot, http_error)
        self.assertEqual(
            "HTTP error 500 for method None on path None: Internal Server Erro"
            "r -- Unexpected error occurred while fetching Cluster/SSP informa"
            "tion : 9999-99Z*2125D4A/1 : Unable to send command to VIOS at thi"
            "s moment. VIOS 1*9999-99Z*2125D4A is busy processing some other r"
            "equest. Please retry the operation after sometime.\n",
            msys_ot.cli_err.getvalue())

    def test_four_char_hex(self):
        self.assertEqual('0200', out_proc.four_char_hex(512))
        self.assertEqual('00DF', out_proc.four_char_hex(223))

    def test_two_char_hex(self):
        self.assertEqual('00', out_proc.two_char_hex(0))
        self.assertEqual('FF', out_proc.two_char_hex(255))
        self.assertEqual('DF', out_proc.two_char_hex(223))

    def _get_lpar_obj_type(self, wrp):
        lpar_cmd = lp_c.LogicalPartition()
        lpar_cmd.ec = object_type.ExecutionContext(None, None, None)
        lpar_cmd.ec.parsed_args = mock.Mock()
        lpar_cmd.ec.parsed_args.max_width = None
        lpar_cmd.ec.parsed_args.where = None
        lpar_cmd.ec.wrappers = [wrp]
        return lpar_cmd

    def test_pri_output(self):
        lpar_resp = pvmhttp.load_pvm_resp(LPAR_FILE).response
        lpar_wrps = lp_w.LPAR.wrap(lpar_resp)

        lpar_cmd = self._get_lpar_obj_type(lpar_wrps[0])

        pri_out = lpar_cmd.pri_output()
        # Check lengths of table, ignoring table title
        for pri_out_split in pri_out.split('\n'):
            self.assertTrue(
                len(pri_out_split) < util.get_term_width())

        with mock.patch('pvmcli.util.util.get_term_width') as gtw_mock:
            gtw_mock.return_value = 80
            pri_out = lpar_cmd.pri_output()
            for out_row in pri_out.split('\n'):
                self.assertTrue(len(out_row) <= 80)
            gtw_mock.assert_called_once_with()
            gtw_mock.reset_mock()

            lpar_wrps[0].name = "A" * 200
            lpar_cmd = self._get_lpar_obj_type(lpar_wrps[0])

            gtw_mock.return_value = 200
            pri_out = lpar_cmd.pri_output()
            self.assertIn('A>', pri_out)
            for out_row in pri_out.split('\n')[1:]:
                self.assertTrue(len(out_row) <= 200)
            gtw_mock.assert_called_once_with()
            gtw_mock.reset_mock()

            gtw_mock.return_value = 400
            pri_out = lpar_cmd.pri_output()
            self.assertNotIn('A>', pri_out)
            for out_row in pri_out.split('\n')[1:]:
                self.assertTrue(len(out_row) <= 400)
                self.assertTrue(len(out_row) > 200)
            gtw_mock.assert_called_once_with()
            gtw_mock.reset_mock()

    def test_get_term_width(self):
        with mock.patch('pvmcli.util.config.value') as cfg_mock:
            cfg_mock.return_value = 70
            self.assertEqual(util.get_term_width(), 80)
            cfg_mock.assert_called_once_with('term', 'default_width')
            cfg_mock.reset_mock()

            cfg_mock.return_value = 90
            self.assertEqual(util.get_term_width(), 90)
            cfg_mock.assert_called_once_with('term', 'default_width')
            cfg_mock.reset_mock()

        with mock.patch('subprocess.check_output') as sp_mock:
            sp_mock.side_effect = Exception()
            self.assertEqual(util.get_term_width(), 80)
            sp_mock.assert_called_once_with(['tput', 'cols'])
            sp_mock.reset_mock()

    def test_is_field_displayable(self):
        lpar_resp = pvmhttp.load_pvm_resp(LPAR_FILE).response
        lpar_wrps = lp_w.LPAR.wrap(lpar_resp)

        self.assertFalse(out_proc.is_field_displayable(lpar_wrps[0],
                                                       'adapter'))
        self.assertTrue(out_proc.is_field_displayable(
            bp_w.IOSlot.bld(lpar_wrps[0].adapter, True, 1), 'adapter'))

    def test_where_maps(self):
        test_clauses = ["LogicalPartition.name=ABCD"]
        actual_eq, actual_ne = out_proc.where_maps(test_clauses)
        self.assertEqual({"LogicalPartition.name": "ABCD"}, actual_eq)
        self.assertEqual({}, actual_ne)

        test_clauses = ["LogicalPartition.name!=ABCD"]
        actual_eq, actual_ne = out_proc.where_maps(test_clauses)
        self.assertEqual({}, actual_eq)
        self.assertEqual({"LogicalPartition.name": "ABCD"}, actual_ne)

        test_clauses = ["LogicalPartition.name    =ABCD",
                        "LogicalPartition.name    !=ABCDE"]
        actual_eq, actual_ne = out_proc.where_maps(test_clauses)
        self.assertEqual({"LogicalPartition.name": "ABCD"}, actual_eq)
        self.assertEqual({"LogicalPartition.name": "ABCDE"}, actual_ne)

        test_clauses = ["LogicalPartition.name=    ABCD",
                        "LogicalPartition.name!=    ABCDE"]
        actual_eq, actual_ne = out_proc.where_maps(test_clauses)
        self.assertEqual({"LogicalPartition.name": "ABCD"}, actual_eq)
        self.assertEqual({"LogicalPartition.name": "ABCDE"}, actual_ne)

        test_clauses = ["LogicalPartition.name    =    ABCD",
                        "LogicalPartition.name    !=    ABCDE"]
        actual_eq, actual_ne = out_proc.where_maps(test_clauses)
        self.assertEqual({"LogicalPartition.name": "ABCD"}, actual_eq)
        self.assertEqual({"LogicalPartition.name": "ABCDE"}, actual_ne)

    def test_graph_conflict(self):
        lpar_resp = pvmhttp.load_pvm_resp(LPAR_FILE).response
        lpar_wrps = lp_w.LPAR.wrap(lpar_resp)

        lpar_cmd = lp_c.LogicalPartition()
        lpar_cmd.ec = object_type.ExecutionContext(None, None, None)
        lpar_cmd.ec.parsed_args = mock.Mock()
        lpar_cmd.ec.parsed_args.max_width = None
        lpar_cmd.ec.parsed_args.where = None
        lpar_cmd.ec.wrappers = [lpar_wrps[0]]
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])

        self.assertFalse(out_proc.graph_conflict(out_graph, lpar_cmd))

        where_clauses = ["LogicalPartition.name=z3-9-5-126-127-00000001"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])

        self.assertFalse(out_proc.graph_conflict(out_graph, lpar_cmd))

        where_clauses = ["LogicalPartition.name!=z3-9-5-126-127-00000001"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])

        self.assertTrue(out_proc.graph_conflict(out_graph, lpar_cmd))

        where_clauses = ["LogicalPartition.name =  z3-9-5-126-127-00000001"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])

        self.assertFalse(out_proc.graph_conflict(out_graph, lpar_cmd))

        where_clauses = ["LogicalPartition.name   != z3-9-5-126-127-00000001"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])

        self.assertTrue(out_proc.graph_conflict(out_graph, lpar_cmd))

        where_clauses = ["PartitionMemoryConfiguration.min=512"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])
        check_graph = out_graph.sch_to_out_graph[
            "PartitionMemoryConfiguration"][0]

        self.assertFalse(out_proc.graph_conflict(check_graph, lpar_cmd))

        where_clauses = ["PartitionMemoryConfiguration.min!=512"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])
        check_graph = out_graph.sch_to_out_graph[
            "PartitionMemoryConfiguration"][0]

        self.assertTrue(out_proc.graph_conflict(check_graph, lpar_cmd))

        where_clauses = ["LogicalPartition.name=z3-9-5-126-127-00000001",
                         "PartitionMemoryConfiguration.min=512"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])
        check_graph = out_graph.sch_to_out_graph[
            "PartitionMemoryConfiguration"][0]

        self.assertFalse(out_proc.graph_conflict(check_graph, lpar_cmd))
        self.assertFalse(out_proc.graph_conflict(out_graph, lpar_cmd))

        where_clauses = ["LogicalPartition.name=z3-9-5-126-127-00000001",
                         "PartitionMemoryConfiguration.min!=512"]
        lpar_cmd.ec.parsed_args.where = where_clauses
        out_graph = out_proc.OutputGraph(lpar_cmd, lpar_wrps[0])
        check_graph = out_graph.sch_to_out_graph[
            "PartitionMemoryConfiguration"][0]

        self.assertTrue(out_proc.graph_conflict(check_graph, lpar_cmd))
        self.assertFalse(out_proc.graph_conflict(out_graph, lpar_cmd))

        where_clauses = ["LogicalPartition.notafield=notavalue"]
        lpar_cmd.ec.parsed_args.where = where_clauses

        self.assertRaises(exc.InvalidIDArgumentError, out_proc.OutputGraph,
                          lpar_cmd, lpar_wrps[0])
