# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

import os
import six

import pvmcli.command.field as fd
import pvmcli.command.operation as op
import pvmcli.util.exceptions as e
from pvmcli.util import help
from pvmcli.util.i18n import _
import pvmcli.util.util as util

import pypowervm.wrappers.entry_wrapper as ew
import pypowervm.wrappers.base_partition as bp_w


FIELD_DELIM_DEFAULT = ','
RECORD_DELIM_DEFAULT = os.linesep
LABEL_CHAR_DEFAULT = '='


HELP_GROUP = util.utf8_str(_('help arguments'))
OUTPUT_GROUP = util.utf8_str(_('output arguments'))
ID_GROUP = util.utf8_str(_('identifier arguments'))

NON_USER_WRAPPER_FIELD_NAMES = ('schema_type',
                                'schema_ns',
                                'default_attrib',
                                'has_metadata',
                                'child_order',
                                'traits',
                                'adapter',
                                'href',
                                'related_href',
                                'element',
                                'etag',
                                'pg83',
                                'parent_entry')

NON_USER_SET_FIELDS = ('uuid')

FIELD_OVERRIDES = {bp_w.IOSlot.schema_type: ["adapter"]}


atomic_f_types = (property, str, bool, float) + six.integer_types
f_types = atomic_f_types + (ew.ElementWrapper, list, tuple)


FGET_FUNC = 'fget'
FSET_FUNC = 'fset'


def _f_name(obj_type, wrp, attr):
    ''' Generate the field name for a given wrapper's attribute.

    :param obj_type: Object type object driving the command execution.
    :param wrp: The wrapper, or wrapper class, for this field.
    :param attr: The field/function name in the given wrapper for this field.
    :return: The field name for the given wrapper and attribute.
    '''
    return "%(sch)s.%(attr)s" % {
        'sch': obj_type.get_schema(wrp.schema_type),
        'attr': attr}


def _valid_attr(wrp_cls, attr):
    ''' Determine if the given attribute is to be included as a field.

    :param wrp_cls: The wrapper class for this field.
    :param attr: The field/function name in the given wrapper for this field.
    :return: If the given attribute of a wrapper is a valid field.
    '''
    # Exclude private fields
    if attr.startswith('_'):
        return False

    # Exclude invalid types
    if not isinstance(getattr(wrp_cls, attr), f_types):
        return False

    # Exclude non-user fields
    if attr in NON_USER_WRAPPER_FIELD_NAMES:
        # If we don't override anything for this wrp_cls
        if wrp_cls.schema_type not in FIELD_OVERRIDES:
            return False
        # Or we do and this attr isn't overridden...
        elif attr not in FIELD_OVERRIDES[wrp_cls.schema_type]:
            return False

    # ...phew. We're valid.
    return True


def _valid_prop_fn(prop_fn, obj_type, wrp_cls, attr):
    ''' Determine if the given attribute is a valid getter or setter.

    :param prop_fn: The property function type (either FGET_FUNC or FSET_FUNC)
    :param obj_type: Object type object driving the command execution.
    :param wrp_cls: The wrapper class for this field.
    :param attr: The field/function name in the given wrapper for this field.
    :return: If the given attribute is valid as the specified field.
    '''
    if prop_fn == FGET_FUNC:
        fn_e = _f_name(obj_type, wrp_cls, attr) not in obj_type.exc_disp_fields
        v_uuid = not(issubclass(wrp_cls, ew.ElementWrapper) and attr == 'uuid')
        return fn_e and v_uuid
    elif prop_fn == FSET_FUNC:
        fn_e = _f_name(obj_type, wrp_cls, attr) not in obj_type.exc_set_fields
        attr_exc = attr not in NON_USER_SET_FIELDS
        return fn_e and attr_exc

    return False


def _field_iterator(prop_fn, obj_type, wrp_cls):
    ''' Generate the display or set fields for a wrapper class.

    :param prop_fn: The property function type (either FGET_FUNC or FSET_FUNC)
    :param obj_type: Object type object driving the command execution.
    :param wrp_cls: The wrapper class for this field.
    :return: List of Field objects for this function type and wrapper.
    '''
    iter_struct = []

    for attr in (a for a in dir(wrp_cls) if _valid_attr(wrp_cls, a)):
        if (getattr(getattr(wrp_cls, attr), prop_fn) and
                _valid_prop_fn(prop_fn, obj_type, wrp_cls, attr)):
            iter_struct.append(attr)

    return iter_struct


def _match_by_substring(substring, list_of_strings):
    ret = []
    # Special case: '' is technically 'in' any string, but that's not useful.
    if substring == '':
        return ret
    return [string for string in list_of_strings if substring in string]


def _split_set_field(set_field, obj_type):
    sf_split = set_field.split('=', 1)
    if len(sf_split) < 2 or len(sf_split[1]) <= 0:
        raise e.InvalidSetFieldFormatError(field=set_field,
                                           obj_type=obj_type.name)
    return sf_split[0], sf_split[1]


def _shorthand_set_field_func(fields, obj_type):
    def expand_field(infield):
        f_label, f_value = _split_set_field(infield, obj_type)
        exp_f_label = _shorthand_field_func(fields, obj_type)(f_label)
        m_l_list = ['%(m)s.%(l)s' % {
            'm': obj_type.get_schema(f.model.schema_type),
            'l': f.label} for f in fields]
        if exp_f_label not in m_l_list:
            raise e.InvalidSetFieldFormatError(field=infield,
                                               obj_type=obj_type.name)
        return "%(label)s=%(val)s" % {'label': exp_f_label, 'val': f_value}
    return expand_field


def _shorthand_field_func(fields, obj_type):
    def expand_field(infield):
        # If input field name is qualified (contains '.'), it's model.label.
        # Otherwise, assume it's just the label.
        inmodel, inlabel = (infield.split('.', 1) if '.' in infield
                            else (None, infield))

        # If model specified, narrow the possibilities accordingly.
        if inmodel is not None:
            upmodel = inmodel.upper()
            # all_models is a set of unique model names, upcased.
            all_models = {obj_type.get_schema(f.model.schema_type).upper()
                          for f in fields}
            # Exact match wins; else substring match.  Say we have models "Foo"
            # and "Foobar": If "Foo" is given, we match only "Foo"; but if "Fo"
            # or "oo" is given, we will match both.
            model_matches = ([upmodel] if upmodel in all_models else
                             _match_by_substring(upmodel, all_models))
            # Possibilities include any whose model name is in the list of
            # model name matches
            possible_fields = [
                df for df in fields if
                obj_type.get_schema(df.model.schema_type).upper()
                in model_matches]
        else:
            # No model - index all display fields
            possible_fields = fields[:]

        # Having filtered by model name, index all the remaining possibilities.
        # label_index = {LABEL: [Model.label, ...]}
        label_index = {}
        for pf in possible_fields:
            labelup = pf.label.upper()
            if labelup not in label_index:
                label_index[labelup] = []
            label_index[labelup].append(pf.field_name)

        uplabel = inlabel.upper()
        # Try for exact label match first.  (This may still yield multiple
        # Model.label matches if the Model was unspecified or ambiguous.)
        matches = label_index.get(uplabel)
        if not matches:
            # No exact match - try to find a substring match.
            matches = []
            for string in _match_by_substring(uplabel, label_index.keys()):
                matches.extend(label_index.get(string))

        # There is a possibility to have duplicates of a display field in the
        # case of shared schema types. For instance, VirtualIOServer has a
        # ClientAdapter and ServerAdapter type for both Fibre Channel Mappings
        # and SCSI Mappings. That will cause ServerAdapter.* and
        # ClientAdapter.* to show up twice. Use set/list type casting to
        # remove redundant types.
        # TODO (clbush): Remove this by revisiting the logic in _field_iterator
        matches = list(set(matches))

        if not matches:
            # BAD PATH: No matches.
            # Let argparse throw default "no such field" error.
            return infield
        if len(matches) == 1:
            # GOOD PATH: Unambiguous match.  Return the hit.
            return matches[0]
        # Check if match parent type had a match.
        for match in matches:
            match_schema = obj_type.get_schema(
                obj_type.get_disp_fields()[match].schema_name)
            if match_schema == obj_type.model.schema_type:
                return match
        # BAD PATH: Ambiguous.
        raise e.AmbiguousFieldNameError(ret_code=e.ARGPARSE_ERR_RET_CODE,
                                        given_field=infield, matches=matches)

    return expand_field


def display_argument(obj_type):
    ''' Returns the Argument object for --display-fields argument.

    :param obj_type: The ObjectType object to build the argument against.
    '''
    dfs = [fd.DisplayField(wrp_func_name=f, model=m,
                           schema_name=obj_type.get_schema(m.schema_type))
           for m in obj_type.field_models
           for f in _field_iterator(FGET_FUNC, obj_type, m)]

    df_choices = []
    for f in dfs:
        # The schema-qualified field name
        df_choices.append(f.field_name)

    return op.Argument(op.DF_NAME, short_name='-d',
                       arg_help=util.utf8_str(help.DISPLAY_FIELDS),
                       metavar=util.utf8_str(_('option')), choices=df_choices,
                       arg_type=_shorthand_field_func(dfs, obj_type),
                       nargs='+', required=False, fields=dfs)


def set_argument(obj_type):
    ''' Returns the Argument object for --set-fields argument.

    :param obj_type: The ObjectType object to build the argument against.
    '''
    set_fields = [
        fd.ModelField(wrp_func_name=f, model=m,
                      schema_name=obj_type.get_schema(m.schema_type))
        for m in obj_type.field_models for f in _field_iterator(FSET_FUNC,
                                                                obj_type, m)]
    return op.Argument(op.SF_NAME, short_name='-s',
                       arg_help=util.utf8_str(help.SET_FIELDS),
                       arg_type=_shorthand_set_field_func(set_fields,
                                                          obj_type),
                       metavar=util.utf8_str(_('option')), nargs='+',
                       required=True, fields=set_fields)


def help_arguments():
    return [help_argument(), dash_help_argument()]


def help_argument():
    return op.Argument('help', arg_help=util.utf8_str(help.HELP_DESC),
                       action="store_true", group=HELP_GROUP)


def dash_help_argument():
    return op.Argument('--help', arg_help=util.utf8_str(help.HELP_DESC),
                       action="help", group=HELP_GROUP)


def generate_id_fields(obj_type, model):
    return [fd.Field(field_name=f_func)
            for f_func in _field_iterator(FGET_FUNC, obj_type, model)]


def object_id_argument(fields, oid_required=False):
    return op.Argument('--object-id', short_name='-i',
                       arg_help=util.utf8_str(help.HELP_ID_ARG),
                       arg_type=str, metavar=util.utf8_str('id=value'),
                       group=ID_GROUP, required=oid_required, fields=fields)


def parent_id_argument(fields, id_required=False):
    return op.Argument('--parent-id', short_name='-p',
                       arg_help=util.utf8_str(help.HELP_PARENT_ID_ARG),
                       arg_type=str, metavar=util.utf8_str('id=value'),
                       group=ID_GROUP, required=id_required,
                       fields=fields)


def child_id_argument(fields, id_required=False):
    return op.Argument('--child-id', short_name='-c',
                       arg_help=util.utf8_str(help.HELP_CHILD_ID_ARG),
                       arg_type=str, metavar=util.utf8_str('id=value'),
                       group=ID_GROUP, fields=fields, required=id_required)


def output_arguments():
    return [delim_argument(), display_label_argument(),
            raw_xml_display_argument(), raw_all_display_argument(),
            where_argument(), separator_argument(), label_char_argument(),
            max_width_argument()]


def delim_argument():
    return op.Argument('--field-delim', short_name='-f',
                       arg_help=util.utf8_str(help.FIELD_DELIM),
                       arg_type=delim_arg_processor, required=False,
                       default=FIELD_DELIM_DEFAULT,
                       group=OUTPUT_GROUP)


def separator_argument():
    return op.Argument('--record-delim', short_name='-r',
                       arg_help=util.utf8_str(help.RECORD_DELIM),
                       arg_type=delim_arg_processor, required=False,
                       default=RECORD_DELIM_DEFAULT,
                       group=OUTPUT_GROUP)


def label_char_argument():
    return op.Argument('--label-char', short_name='-l',
                       arg_help=util.utf8_str(help.LABEL_CHAR),
                       arg_type=delim_arg_processor, required=False,
                       default=LABEL_CHAR_DEFAULT,
                       group=OUTPUT_GROUP)


def max_width_argument():
    return op.Argument('--max-width', short_name='-m',
                       arg_help=util.utf8_str(help.MAX_WIDTH),
                       arg_type=int, required=False, group=OUTPUT_GROUP)


def display_label_argument():
    return op.Argument('--hide-label',
                       arg_help=util.utf8_str(help.HIDE_LABEL),
                       action='store_true', group=OUTPUT_GROUP,
                       default=False)


def where_argument():
    ''' Filter wrappers from output based on specified field equality.

        name: --where
        short name: -w
        nargs: 1 - N
        required: False
        Example:
            Assume a user needs to get all physical IO adapter names off of a
            managed system such that the IO slot description on that object is
            not "Empty slot" and the pci_class is 65535. That information
            exists as part of AssociatedSystemIOConfig.IOSlot under the System
            wrapper. This argument will allow a user to only see the output
            they want with the following command:
            pvmctl ManagedSystem list -d IOSlot.dyn_reconfig_conn_name \
                --delim , --hide-label \
                --where "IOSlot.description!=Empty slot" \
                        "IOSlot.pci_class=65535"
    '''
    # TODO (clbush): regex support
    # TODO (clbush): Look into not requiring double quotes
    return op.Argument('--where', short_name='-w',
                       arg_help=util.utf8_str(help.WHERE_DISP),
                       nargs='+', required=False, group=OUTPUT_GROUP)


def raw_xml_display_argument():
    return op.Argument('--xml',
                       arg_help=util.utf8_str(help.XML_DISP),
                       action='store_true', group=OUTPUT_GROUP,
                       default=False)


def raw_all_display_argument():
    return op.Argument('--all-out',
                       arg_help=util.utf8_str(help.ALL_DISP),
                       action='store_true', group=OUTPUT_GROUP,
                       default=False)


def default_args(obj_type, parent_models=None, child_models=None,
                 include_output=True, object_id_required=False,
                 parent_id_required=False, child_id_required=False):
    ''' Generate a list of default arguments for most operations.

    :param obj_type: The object type that owns the operation.
    :param parent_models: The parent models on the object type to generate
                          information for the -p/--parent-id argument
    :param child_models: The child models on the object type to generate
                         information for the -c/--child-id argument
    :param include_output: Flag to include output and display field arguments
    :param object_id_required: Flag to force --object-id/-i to be required.
    :param parent_id_required: Flag to force --parent-id/-p to be required.
    :param child_id_required: Flag to force --child-id/-c to be required.
    :returns: A list of Argument objects to be added to an operation.
    '''
    args = []
    if include_output:
        args.append(display_argument(obj_type))
        args.extend(output_arguments())

    args.extend(help_arguments())

    id_fields = generate_id_fields(obj_type, obj_type.model)
    args.append(object_id_argument(id_fields, oid_required=object_id_required))

    if parent_models:
        for parent_model in parent_models:
            args.append(parent_id_argument(generate_id_fields(obj_type,
                                                              parent_model),
                                           id_required=parent_id_required))
    if child_models:
        for child_model in child_models:
            args.append(child_id_argument(generate_id_fields(obj_type,
                                                             child_model),
                                          id_required=child_id_required))

    return args


def delim_arg_processor(arg_str):
    ''' Used for delimiter and label-char arguments type to process \\n

    :param arg_str: The argument string to parse for \\n.
    '''
    return str.replace(arg_str, r'\n', os.linesep)
