# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import os

import pvmcli.util.exceptions as exc
from pvmcli.util import env_var

from pypowervm import adapter


DEFAULT_HOST = 'localhost'
DEFAULT_HTTP_PORT = 12080
DEFAULT_HTTPS_PORT = 12443
PROTOCOL_HTTP = 'http'
PROTOCOL_HTTPS = 'https'


class SessionManager(object):
    ''' pypowervm session adapter wrapper used by pvmctl. '''
    def __init__(self, user_arg=None, pass_arg=None,
                 host_arg=None, port_arg=None):
        self._session_adapter = None
        self.user_arg = user_arg
        self.pass_arg = pass_arg
        self.host_arg = host_arg
        self.port_arg = port_arg

    def split_arg_search(self, arg, wrp_cls, parent_type=None,
                         parent_uuid=None):
        '''
        This method provides a CLI input friendly interface to pypowervm's
        session adapter search functionality. For instance, to search for a
        LPAR with a name of foo, a consumer would pass
        pypowervm.wrappers.logical_partition.LPAR as wrp_cls, and "name=foo" as
        arg. "name=foo" is parsed and then passed as kwargs to the
        wrp_cls.search method.

        :param arg: CLI argument of the form foo<delim>bar (e.g. foo=bar)
                    to pass as arguments to the search function kwargs
        :param wrp_cls: The wrapper class we are searching for.
        :param parent_type: Parent wrapper class of the wrp_cls, if any.
        :param parent_uuid: UUID of the parent wrapper owning the object we are
                            searching for.
        :return: A list of wrp_cls wrapper objects matching the provided arg
                 search criteria.
        '''
        split_arg = arg.split('=', 1)
        if len(split_arg) != 2 or len(split_arg[1]) == 0:
            raise exc.InvalidSplitArgError(object_type=wrp_cls.schema_type,
                                           arg=arg)
        return wrp_cls.search(
            self.adapter, parent_type=parent_type, parent_uuid=parent_uuid,
            **{split_arg[0]: split_arg[1]})

    @property
    def adapter(self):
        if not self._session_adapter:
            self.create_session_adapter()
        return self._session_adapter

    def create_session_adapter(self):
        # TODO do we kill? throw exception?
        # error handling (see story 120327)
        if self._session_adapter is not None:
            pass

        # User is optional when running locally
        # If left as None, pypowervm will get current user
        user = None
        # If env variable is set, use that instead
        if os.environ.get(env_var.ENV_USER):
            user = os.environ.get(env_var.ENV_USER)
        # Passed in user arg overrides everything
        if self.user_arg:
            user = self.user_arg

        # Password is optional when running locally
        password = None
        if os.environ.get(env_var.ENV_PASS):
            password = os.environ.get(env_var.ENV_PASS)
        if self.pass_arg:
            password = self.pass_arg

        # Defaults to localhost, for running locally
        host = DEFAULT_HOST
        if os.environ.get(env_var.ENV_HOST):
            host = os.environ.get(env_var.ENV_HOST)
        if self.host_arg:
            host = self.host_arg

        # Port is optional when running locally,
        # defaults to default HTTP port.
        port = (DEFAULT_HTTP_PORT if host == DEFAULT_HOST
                else DEFAULT_HTTPS_PORT)
        if os.environ.get(env_var.ENV_PORT):
            port = os.environ.get(env_var.ENV_PORT)
        if self.port_arg:
            port = self.port_arg

        # All exceptions thrown to caller.
        protocol = (PROTOCOL_HTTP if host == DEFAULT_HOST
                    else PROTOCOL_HTTPS)

        session = adapter.Session(
            host, user, password, protocol=protocol, port=int(port),
            certpath=None)

        self._session_adapter = adapter.Adapter(session, use_cache=False)
        return self.adapter
