# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017.
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import os
import re
import six
import subprocess
import sys

import pvmcli.util.config as config
import pvmcli.util.exceptions as exc
from pvmcli.util.i18n import _

import pypowervm.tasks.sriov as sriov_task
import pypowervm.util as py_util
from pypowervm.wrappers import base_partition as bp_w
import pypowervm.wrappers.iocard as iocard
from pypowervm.wrappers import logical_partition as lpar_w
import pypowervm.wrappers.managed_system as m_sys
from pypowervm.wrappers import network as net_w
from pypowervm.wrappers import storage as stor
from pypowervm.wrappers import virtual_io_server as vios_w


MIN_TERM_WIDTH = 80

lpar_cache = {}
vios_cache = {}
vswitch_cache = {}
name_cache = {}


def bool_str(val):
    return _("True") if val else _("False")


def utf8_str(val):
    if not val:
        return val
    try:
        if six.PY3:
            return val.encode('utf-8').decode('utf-8')
        return val.encode('utf-8')
    except UnicodeError:
        return val


def disable_print_attr(val):
    '''
    Replace the attributes from --help print output which are not implemented,
    but only defined in class property
    '''
    if "pvmctl [help | --help] Cluster list" in val:
        if "Cluster.nodes\n" in val:
            val = val.replace("    Cluster.nodes\n", "")
        if "PhysicalVolume.loc_code\n" in val:
            val = val.replace("    PhysicalVolume.loc_code\n", "")
        if "PhysicalVolume.read_iops_limit\n" in val:
            val = val.replace("    PhysicalVolume.read_iops_limit\n", "")
        if "PhysicalVolume.tag\n" in val:
            val = val.replace("    PhysicalVolume.tag\n", "")
        if "PhysicalVolume.write_iops_limit\n" in val:
            val = val.replace("    PhysicalVolume.write_iops_limit\n", "")
    if "pvmctl [help | --help] VirtualNICDedicated list" in val:
        if "VirtualNICDedicated.capacity\n" in val:
            val = val.replace("    VirtualNICDedicated.capacity\n", "")
        if "VirtualNICDedicated.back_devs\n" in val:
            val = val.replace("    VirtualNICDedicated.back_devs\n", "")
    if "Valid Fields: assoc_partition_id, auto_assign, uuid, vsn" in val:
            val = val.replace(" assoc_partition_id, auto_assign, uuid,", "")
    if "pvmctl [help | --help] LogicalPartition update" in val:
        val = val.replace("    LogicalPartition.nvram\n", "")
        val = val.replace("    LogicalPartition.proc_config\n", "")
        val = val.replace("    LogicalPartition.mem_config\n", "")
        val = val.replace("    LogicalPartition.io_config\n", "")
        val = val.replace("    PartitionIOConfiguration.io_slots\n", "")
    if "pvmctl [help | --help] VirtualEthernetAdapter update" in val:
        val = val.replace("    VirtualEthernetAdapter.ovs_bridge\n", "")
        val = val.replace("    VirtualEthernetAdapter.configured_mtu\n", "")
        val = val.replace("    VirtualEthernetAdapter.ovs_ext_ids\n", "")
        val = val.replace(
            "    VirtualEthernetAdapter.is_tagged_vlan_supported\n", "")
    if "pvmctl [help | --help] NetworkBridge update" in val:
        val = val.replace("    NetworkBridge.seas\n", "")
        val = val.replace("    NetworkBridge.load_grps\n", "")
        val = val.replace("    LoadGroup.trunk_adapters\n", "")
        val = val.replace("    LoadGroup.vnet_uri_list\n", "")
        val = val.replace("    SharedEthernetAdapter.addl_adpts\n", "")
        val = val.replace("    TrunkAdapter.has_tag_support\n", "")
        val = val.replace("    TrunkAdapter.tagged_vlans\n", "")
        val = val.replace("    TrunkAdapter.pvid\n", "")

    if "pvmctl [help | --help] ManagedSystem list" in val:
        val = val.replace("    IBMiVirtualSoftwareTiers."
                          "virtual_software_tier\n", "")
    return val


def print_(val, outfile=sys.stdout):
    '''
    Take in a translated string, decode it with utf-8, then print to file.
    :param outfile: File (e.g. sys.stdout, sys.stderr) to print to.
    '''
    ''' Function for disabling unused attr from --help
    '''
    val = disable_print_attr(val)
    six.print_(utf8_str(val), file=outfile)


def input(prompt):
    '''
    A python version independent user prompted input function

    :param prompt: The text input prompt
    :return: The user provided value
    '''
    return six.moves.input(prompt)


def split_set_fields(set_fields):
    field_map = {}
    for sf in set_fields:
        key, value = sf.split('=', 1)
        field_map[key] = value

    return field_map


def group_join_list(join_list, num_cols=3):
    """Return a list of comma separated strings for output.

    For example, join_list of [1, 2, 3, 4, 5, 6, 7] with num_cols=3 will
    return ["1,2,3,"4,5,6","7"]. Used for priority output.

    :param join_list: A list of elements to comma separate.
    :param num_cols: The number of elements per string.
    :return: A list of comma separated strings
    """
    def group_strs(elems):
        if len(elems) <= num_cols:
            return [','.join(elems)]
        ret = [','.join(elems[:num_cols]) + ',']
        ret.extend(group_strs(elems[num_cols:]))
        return ret
    return group_strs([six.text_type(elem) for elem in join_list])


def get_term_width():
    tput_cmd = ['tput', 'cols']
    try:
        width = int(config.value('term', 'default_width'))
    except:
        try:
            width = int(subprocess.check_output(tput_cmd))
        except:
            width = MIN_TERM_WIDTH

    return width if width >= MIN_TERM_WIDTH else MIN_TERM_WIDTH


def _set_vios_wrps(obj_type):
    if obj_type.ec.parsed_args.vios:
        return [find_vios_by_name(obj_type.ec.sm.adapter,
                                  obj_type.ec.parsed_args.vios)]
    return vios_w.VIOS.get(obj_type.ec.sm.adapter)


def _vios_for_upload(obj_type):
    vios_wrps = _set_vios_wrps(obj_type)
    if not vios_wrps or None in vios_wrps:
        raise exc.NoEntriesFoundError(object_type=vios_w.VIOS.schema_type)
    for vios_wrp in vios_wrps:
        v_state = vios_wrp.state == bp_w.LPARState.RUNNING
        m_part = vios_wrp.is_mgmt_partition
        rmc_active = vios_wrp.rmc_state == bp_w.RMCState.ACTIVE
        if v_state and (m_part or rmc_active):
            return vios_wrp

    raise exc.NoRunningVIOSFoundError()


def upload_setup(obj_type):
    """Setup function for a storage upload operation.

    :param obj_type: The object type that owns the upload operation. The
                     argparse parsed_args in the object type must contain a
                     data_size, data, and vios argument.
    :returns: A tuple of the data, data_size arguments and VIOS UUID
              for the upload operation.
    """
    data_arg = sys.stdin
    data_size_arg = obj_type.ec.parsed_args.data_size
    if obj_type.ec.parsed_args.data:
        data_arg = obj_type.ec.parsed_args.data
        if not data_size_arg:
            data_size_arg = os.path.getsize(data_arg)
    elif not data_size_arg:
        raise exc.UploadStdInRequiresSizeError()

    return data_arg, data_size_arg, _vios_for_upload(obj_type).uuid


def vios_by_uri(adpt, vio_uri, xag=None):
    ''' Build a VIOS wrapper from a VIOS URI

    :param adpt: The adapter facilitating the search request
    :param vio_uri: A VIOS URI from which to build the VIOS wrapper.
    :param xag: List of extended attribute group enum value.
    :return: The VIOS wrapper associated with that vio_uri
    '''
    if vio_uri not in vios_cache:
        vios_cache[vio_uri] = vios_w.VIOS.wrap(
            adpt.read_by_href(vio_uri, xag=xag))

    return vios_cache[vio_uri]


def vswitch_search(adpt, switch_id, ms_uuid):
    ''' Build a VSwitch wrapper from a given VSwitch ID

    :param adpt: The adapter facilitating the search request
    :param switch_id: A VSwitch ID
    :return: The VSwitch wrapper associated with that vswitch_id
    '''
    if switch_id not in vswitch_cache:
        vsw_list = vswitch_cache[switch_id] = net_w.VSwitch.search(
            adpt, parent_type=m_sys.System.schema_type,
            parent_uuid=ms_uuid, switch_id=switch_id)
        vswitch_cache[switch_id] = (vsw_list[0] if vsw_list else None)

    return vswitch_cache[switch_id]


def vswitch_by_uri(adpt, vswitch_uri):
    ''' Build a VSwitch wrapper from a given VSwitch uri

    :param adpt: The adapter facilitating the search request
    :param vswitch_uri: A VSwitch URI
    :return: The VSwitch wrapper associated with that vswitch_uri
    '''
    if vswitch_uri not in vswitch_cache:
        vswitch_cache[vswitch_uri] = net_w.VSwitch.wrap(
            adpt.read_by_href(vswitch_uri))
    return vswitch_cache[vswitch_uri]


def vios_for_vg(adpt, vg_uri):
    vios_uuid = py_util.get_req_path_uuid(vg_uri, preserve_case=True,
                                          root=True)

    if vios_uuid not in vios_cache:
        vios_cache[vios_uuid] = vios_w.VIOS.search(adpt, uuid=vios_uuid)[0]
    return vios_cache[vios_uuid]


def find_entry_by_name(adpt, entry_name, entry_type,
                       parent_type=None, parent_uuid=None):
    if entry_type not in name_cache:
        name_cache[entry_type.schema_type] = {}

    if entry_name not in name_cache[entry_type.schema_type]:
        found_entry = entry_type.search(
            adpt, name=entry_name, parent_type=parent_type,
            parent_uuid=parent_uuid)
        name_cache[entry_type.schema_type][entry_name] = (
            found_entry[0] if found_entry else None)

    return name_cache[entry_type.schema_type][entry_name]


def find_vios_by_name(adpt, vios_name):
    return find_entry_by_name(adpt, vios_name, vios_w.VIOS)


def find_ssp_by_name(adpt, ssp_name):
    return find_entry_by_name(adpt, ssp_name, stor.SSP)


def find_vg_by_name(adpt, vg_name, vios_uuid):
    return find_entry_by_name(adpt, vg_name, stor.VG,
                              parent_type=vios_w.VIOS, parent_uuid=vios_uuid)


def _name_by_id(adpt, part_id, cache, wrp_cls):
    if not cache:
        for part_wrp in wrp_cls.get(adpt):
            cache[six.text_type(part_wrp.id)] = part_wrp

    return (cache[part_id].name if part_id in cache else "")


def vios_name_by_id(adpt, vios_id):
    return _name_by_id(adpt, vios_id, vios_cache, vios_w.VIOS)


def lpar_name_by_id(adpt, lpar_id):
    return _name_by_id(adpt, lpar_id, lpar_cache, lpar_w.LPAR)


def part_name_by_id(adpt, part_id):
    part_name = lpar_name_by_id(adpt, part_id)
    if not part_name:
        return vios_name_by_id(adpt, part_id)
    return part_name


def split_eq_arg(arg, id_arg_fields):
    id_kwargs = {}
    split_id = arg.split('=', 1)
    if split_id[0] not in id_arg_fields:
        raise exc.InvalidIDArgumentError(
            id_arg=arg, valid_fields=six.text_type(", ".join(id_arg_fields)))
    f_type = id_arg_fields[split_id[0]].field_type

    if len(split_id) != 2 or len(split_id[1]) <= 0:
        raise exc.MissingValueOfKey(field=split_id[0])
    id_kwargs[split_id[0]] = f_type(split_id[1])
    return id_kwargs


def parse_float_as_pct(str_with_float):
    ''' Parse a float value out of a string containing a float, divide by 100.

    :param str_with_float: A string containing a float value, e.g. 42.42%.
    :return: A float representing that string, e.g. 0.4242. Will raise
             TypeError if provided string is not parseable as a float.
    '''
    parsed_val = re.findall(r"^\d*\.?\d+%?$", str_with_float)
    if parsed_val:
        return float(re.sub('%', '', parsed_val[0])) / 100
    raise TypeError(_("Provided string %s is not parseable as a "
                      "float type." % str_with_float))


def _parse_back_dev(back_dev):
    back_dev_str = ' '.join(back_dev)
    bd_keys = ['vios', 'loc', 'cap', 'pri', 'maxcap']
    bd_key_str = ', '.join(bd_keys)
    parsed_vios, parsed_loc, parsed_cap = None, None, None
    parsed_pri, parsed_max = None, None
    for bd_token in back_dev:
        try:
            bd_key, bd_val = bd_token.split('=', 1)
            bd_key = bd_key.lower()
            if bd_key not in bd_keys:
                spec_err = ("Unrecognized --back-dev key '%(bd_key)s'. Valid "
                            "keys are: %(bd_keys)s." % {'bd_key': bd_key,
                                                        'bd_keys': bd_key_str})
                raise exc.InvalidBackDevError(
                    spec_err=spec_err, back_dev_str=back_dev)
            if bd_key == 'vios':
                parsed_vios = bd_val
            elif bd_key == 'loc':
                parsed_loc = bd_val
            elif bd_key == 'cap':
                parsed_cap = parse_float_as_pct(bd_val)
            elif bd_key == 'pri':
                parsed_pri = int(bd_val)
            elif bd_key == 'maxcap':
                parsed_max = parse_float_as_pct(bd_val)
        except ValueError:
            raise exc.InvalidBackDevError(
                spec_err="Unable to parse token %s." % bd_token,
                back_dev_str=back_dev_str)
    return parsed_vios, parsed_loc, parsed_cap, parsed_pri, parsed_max


def build_back_dev(adpt, back_dev, msys):
    '''
    Parse a list representing a VNIC backing device to be constructed by
    pypowervm. Example: ['vios=vios1', 'loc=U78CB.001.WZS0485-P1-C5-T3',
                         'cap=42.42%', 'pri=5']
    :param adpt: pypowervm Adapter object.
    :param back_dev: A list of strings representing the backing device. Each
                     string is in a "key=value" form, where key is one of the
                     following: vios, loc, cap, and pri.
    :param msys: ManagedSystem wrapper for this NovaLink instance.
    :return: A pypowervm.iocard.VNICBackDev object
    '''

    # Parse and validate the tokens passed in for back_dev.
    p_vios, p_loc, p_cap, p_pri, p_max_cap = _parse_back_dev(back_dev)

    back_dev_str = ' '.join(back_dev)
    if not p_vios:
        raise exc.InvalidBackDevError(
            spec_err="No VIOS specified in CLI.", back_dev_str=back_dev_str)

    if not p_loc:
        raise exc.InvalidBackDevError(
            spec_err="No loc specified in CLI.", back_dev_str=back_dev_str)

    # Use pypowervm to find a physical port object based on location code
    pport = sriov_task.find_pport(msys, p_loc)

    if not pport:
        spec_err = ("An SR-IOV physical port with location code %s "
                    "could not be found." % p_loc)
        raise exc.InvalidBackDevError(
            spec_err=spec_err, back_dev_str=back_dev_str)

    fnd_vios = find_vios_by_name(adpt, p_vios)
    if not fnd_vios:
        raise exc.NoEntriesFoundError(object_type=vios_w.VIOS.schema_type)

    return iocard.VNICBackDev.bld(
        adpt, vios_uuid=fnd_vios.uuid, sriov_adap_id=pport.sriov_adap_id,
        pport_id=pport.port_id, capacity=p_cap, failover_pri=p_pri,
        max_capacity=p_max_cap)
